/*****************************************************************************/
/*                                                                           */
/*                              SOFTING AG                                   */
/*                        Richard-Reitzner-Allee 6                           */
/*                              D-85540 Haar                                 */
/*                      Phone: (++49)-(0)89-45656-0                          */
/*                      Fax:   (++49)-(0)89-45656-399                        */
/*                                                                           */
/*                  Copyright (C) SOFTING AG 1995-2004                       */
/*                                                                           */
/* See file CREDITS for list of people who contributed to this               */
/* project.                                                                  */
/*                                                                           */
/* This program is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU General Public License as            */
/* published by the Free Software Foundation; either version 2 of            */
/* the License, or (at your option) any later version.                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston,                    */
/* MA 02111-1307, USA.                                                       */
/*                                                                           */
/* email: support.automation@softing.com                                     */
/* Softing AG, Richard-Reitzner-Allee 6, 85540 Haar, Germany                 */
/*                                                                           */
/*****************************************************************************/
/*****************************************************************************/
/*                                                                           */
/*  Author      : SOFTING                                                    */
/*                                                                           */
/*  CHANGE_NOTES                                                             */
/*                                                                           */
/*  date      name      change                                               */
/*  -----------------------------------------------------------------------  */
/*****************************************************************************/
/*****************************************************************************/

#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <linux/version.h>
#include <linux/sched.h>

#include "kernel_version.h"

#if (VERSION_CODE(2,2,0) <= LINUX_VERSION_CODE)
  #include <linux/poll.h>
#endif

#if ((VERSION_CODE(2,0,0) <= LINUX_VERSION_CODE) && \
     (LINUX_VERSION_CODE < VERSION_CODE(2,3,0)) )
#include <linux/malloc.h>
#elif (VERSION_CODE(2,4,0) <= LINUX_VERSION_CODE)
#include <linux/slab.h>
#endif

#include "basetypes.h" 
#include "fw.h"
#include "cntrl.h"
#include "cmi.h"
#include "xx_if.h"
#include "xx_err.h"
#include "xx_conf.h"
#include "pb_xxx.h"
#include "pb_debug.h"

#define CONTROLLER_EXT struct CntrlExt

#define SYNC_BY_LOCK 0

typedef T_PROFI_SERVICE_DESCR T_SDB;
typedef unsigned char uchar;

extern void /*inline*/ StartTimeout( int *Timeout, int value);
extern int SleepAWhile( int *Timeout );
extern unsigned short reset_cntrl_irq( struct CntrlExt *CtrlExt);
extern int PbInitBoard( struct CntrlExt *Board, char Reset );
extern void DumpMem(void *pMem, unsigned long size, char *pText);




int cmi_init( struct CntrlExt *CtrlExt );
static int16 cmi_get_data_image_idx( CONTROLLER_EXT *CtrlExt,
  T_DATA_DESCR *h_data_descr_list, uint8 image_id, uint8 *image_idx );
//static int16 cmi_set_data_descr( struct CntrlExt *CtrlExt,
//    T_DATA_DESCR *h_data_descr_list, uint8 image_idx, uint16 offset,
//    uint16 data_descr_size, void *data_descr_ptr, int mode );
static int16 cmi_lock_data_image( struct CntrlExt *CtrlExt,
    T_DATA_DESCR *h_data_descr_list, int SyncBySema, uint8 image_idx);
//static uint8 cmi_unlock_data_image( struct CntrlExt *CtrlExt,
//    T_DATA_DESCR *h_data_descr_list, int SyncBySema, uint8 image_idx);
static int16 cmi_wait_for_unlock_data_image( struct CntrlExt *CtrlExt,
  T_DATA_DESCR *h_data_descr_list, uint8 image_idx );

static unsigned long dummy;

uint32 swap_32_intel_motorola( uint32 ToSwap)
{
  uint8 *inval = (uint8*)&ToSwap;
  return (inval[0]<<24) | (inval[1]<<16) | (inval[2]<<8) | (inval[3]);
}

PB_ADDRESS cmi_correct_address( T_CMI_DESCRIPTOR *cmi_descriptor, USIGN32 addr)
{
  PB_ADDRESS address = addr;
  
  if (address == 0) return(0);

  switch( cmi_descriptor->h_address_swap_mode ) 
  {
  case SEGMENT_OFFSET_INTEL:
    switch( read_io_byte(&cmi_descriptor->c_address_swap_mode) ) 
    {
    case ABS_16BIT_INTEL:
    case SEGMENT_OFFSET_INTEL:
      address = address - read_io_long(&cmi_descriptor->c_base_address) +
      read_io_long(&cmi_descriptor->h_base_address);
      break;
    case ABS_32BIT_INTEL:  // base address MUST be linear
      address = address - read_io_long(&cmi_descriptor->c_base_address);
      address = ((address & 0xf0000L) << 12) + (address & 0xffffL);
      address = address + read_io_long(&cmi_descriptor->h_base_address);
      break;
    case REL_32BIT_INTEL:
      address = ((address & 0xf0000L) << 12) + (address & 0xffffL);
    case REL_16BIT_INTEL:
      //address = address + read_io_long(&cmi_descriptor->h_base_address);
      address = address + (PB_ADDRESS)((char*)cmi_descriptor);
      break;
    case ABS_16BIT_MOTOROLA:
      address = swap_32_intel_motorola( address );
      address = address - swap_32_intel_motorola(
        read_io_long(&cmi_descriptor->c_base_address)) +
        read_io_long(&cmi_descriptor->h_base_address);
      break;
    case ABS_32BIT_MOTOROLA:
      address = swap_32_intel_motorola( address );
      address = address - swap_32_intel_motorola(
        read_io_long(&cmi_descriptor->c_base_address));
      address = (((address & 0xf0000L) << 12) + (address & 0xffffL) ) +
        read_io_long(&cmi_descriptor->h_base_address);
      break;
    case REL_32BIT_MOTOROLA:
      address  = swap_32_intel_motorola( address );
      address  = (((address & 0xf0000L) << 12) + (address & 0xffffL)) +
        read_io_long(&cmi_descriptor->h_base_address);
      break;
    case REL_16BIT_MOTOROLA:
      address = swap_32_intel_motorola( address );
      address = address + read_io_long(&cmi_descriptor->h_base_address);
      break;
    default:
      break;
    }
    break;
  default:
    break;
  }
  return(address);
}

/******************************************************************/
/* The DP stack returns the offset of input and output areas in a */
/* specific structure (T_DP_SLAVE_IO_CONFIG_IMAGE).               */
/* This information togeether with the base address of the image  */
/* is needed for the copying functions of the interrupt routine.  */
/* Together with the input and output data the status field (one  */
/* byte for every station) has to be copied. Because the length   */
/* is fixed, a static field in the Board->dp structure is used.   */
/******************************************************************/

int cmi_init_dp( struct CntrlExt *Board )
{
  T_CMI_DESCRIPTOR  *cmi_descriptor = (T_CMI_DESCRIPTOR *) Board->DprAddr;
  T_DATA_DESCR *h_data_descr_list; // list of supported data types
  uint8  *h_data_descr_ptr;
  uint8 data_image_idx = 0, data_status_idx=0;
  int16 ret_val = E_OK;
  T_DP_SLAVE_IO_CONFIG_IMAGE IoCnf;
  uint16 BufSize = sizeof(IoCnf);

  PR_DEBUG(D_CMI_L1,"cmi_init_dp\n");
  h_data_descr_list = (T_DATA_DESCR *)Board->h_data_descr_addr;

  // check the address read in the cmi
  if( (*(USIGN32*)h_data_descr_list == 0xFFFFFFFF) ||
      (*(USIGN32*)h_data_descr_list == 0x00000000)
    ) 
  {
    // the addresses can be 0xFFFFFFFF if DPRAM is disabled or
    // if a PROFIcard is yet removed
    printk("profiboard: wrong image list 0x%x\n",
      read_io_long((long*) h_data_descr_list) );
    return E_IF_FATAL_ERROR;
  }

  if( (ret_val = cmi_get_data_image_idx( Board, h_data_descr_list,
      ID_DP_SLAVE_IO_CONFIG_IMAGE,&data_image_idx))!= E_OK ) 
  {
    //printk("profiboard: get_data_image_idx %d\n", ret_val );
    return( ret_val );
  }

  PR_DEBUG(D_CMI_L1,"data_image_idx = %d\n", data_image_idx );
  h_data_descr_list += data_image_idx;

  /* compute number of bytes to copy */
  if (BufSize > read_io_word( &h_data_descr_list->d_data_size) )
    BufSize = read_io_word( &h_data_descr_list->d_data_size );

  // get address of data image
  h_data_descr_ptr = (uint8 *) cmi_correct_address(cmi_descriptor, h_data_descr_list->d_data_addr);
  // check for CMI error
  if( (h_data_descr_ptr < (unsigned char*)cmi_descriptor) ||
    (h_data_descr_ptr+BufSize >=
    (unsigned char*)cmi_descriptor+DPR_PAGE_SIZE)) 
  {
    printk("profiboard: E_IF_CMI_ERROR\n");
    return( E_IF_CMI_ERROR );
  }

  // THIS IS ONLY FOR INTEL BASED SYSTEMS!!!
  // because data below 1M has to be read by read[bwl]
  copy_from_io ((void*)&IoCnf, h_data_descr_ptr, (unsigned int) BufSize);

  /********************************************************************/
  /* Now, after having the offset and length information, we need the */
  /* base address.                                                    */
  /********************************************************************/
  h_data_descr_list=(T_DATA_DESCR *) Board->h_data_descr_addr; 
  data_image_idx = 0;
  data_status_idx = 0;
  if( (ret_val = cmi_get_data_image_idx( Board, h_data_descr_list,
      ID_DP_SLAVE_IO_IMAGE,&data_image_idx))!= E_OK ) 
  {
    //printk("profiboard: io_image %d\n", ret_val );
    return( ret_val );
  }
  if( (ret_val = cmi_get_data_image_idx( Board, h_data_descr_list,
      ID_DP_STATUS_IMAGE,&data_status_idx))!= E_OK ) 
  {
    //printk("profiboard: status %d\n", ret_val );
    return( ret_val );
  }
  PR_DEBUG(D_CMI_L1,"list=0x%p - idx=%d - ",h_data_descr_list,data_image_idx);
  h_data_descr_list += data_image_idx;
  PR_DEBUG(D_CMI_L1,"list=0x%p, in-offset=%d, out-offset=%d\n",
    h_data_descr_list,IoCnf.offset_inputs, IoCnf.offset_outputs );
  h_data_descr_ptr = (uint8 *) cmi_correct_address(cmi_descriptor, h_data_descr_list->d_data_addr); 
  Board->Dp.InLength = IoCnf.number_inputs;
  Board->Dp.OutLength = IoCnf.number_outputs;
  Board->Dp.InOffset = IoCnf.offset_inputs;
  Board->Dp.OutOffset = IoCnf.offset_outputs;
  Board->Dp.InAddr  = (PB_ADDRESS)(h_data_descr_ptr + IoCnf.offset_inputs);
  Board->Dp.OutAddr = (PB_ADDRESS)(h_data_descr_ptr + IoCnf.offset_outputs);
  Board->Dp.StateLength = DP_MAX_NUMBER_STATIONS;
  // compute address of status field
  h_data_descr_list -= data_image_idx;
  h_data_descr_list += data_status_idx;
  h_data_descr_ptr = (uint8 *) cmi_correct_address(cmi_descriptor, h_data_descr_list->d_data_addr);
  
  Board->Dp.StateAddr=(PB_ADDRESS)h_data_descr_ptr;

  if( Board->Dp.InBuf ) kfree( Board->Dp.InBuf );
  if( Board->Dp.OutBuf ) kfree( Board->Dp.OutBuf );
  if( Board->Dp.StateBuf ) kfree( Board->Dp.StateBuf );
  if( Board->Dp.InBufCopy ) kfree( Board->Dp.InBufCopy );
  if( Board->Dp.OutBufCopy ) kfree( Board->Dp.OutBufCopy );
  if( Board->Dp.StateBufCopy ) kfree( Board->Dp.StateBufCopy );
  
  PR_DEBUG(D_CMI_L1,"Allocating memory %d %d %d...\n",
    IoCnf.number_inputs, IoCnf.number_outputs,Board->Dp.StateLength);
  
  Board->Dp.InBuf = kmalloc( IoCnf.number_inputs, GFP_ATOMIC );
  Board->Dp.InBufCopy = kmalloc( IoCnf.number_inputs, GFP_ATOMIC );
  Board->Dp.OutBuf = kmalloc( IoCnf.number_outputs, GFP_ATOMIC );
  Board->Dp.OutBufCopy = kmalloc( IoCnf.number_outputs, GFP_ATOMIC );
  Board->Dp.StateBuf = kmalloc( Board->Dp.StateLength, GFP_ATOMIC );
  Board->Dp.StateBufCopy = kmalloc( Board->Dp.StateLength, GFP_ATOMIC );
  
  memset( Board->Dp.OutBuf, 0, IoCnf.number_outputs );
  memset( Board->Dp.InBuf, 0, IoCnf.number_inputs );
  memset( Board->Dp.StateBuf, 0, Board->Dp.StateLength );
  
  memset( Board->Dp.OutBufCopy, 0, IoCnf.number_outputs );
  memset( Board->Dp.InBufCopy, 0, IoCnf.number_inputs );
  memset( Board->Dp.StateBufCopy, 0, Board->Dp.StateLength );
  
  PR_DEBUG(D_CMI_L1,"cmi_init_dp end\n");
  return E_OK;
}

void cmi_exit_dp( struct CntrlExt *Board )
{
  PR_DEBUG(D_CMI_L1,"cmi_exit_dp %p %p %p ...\n",
      Board->Dp.InBuf, Board->Dp.OutBuf, Board->Dp.StateBuf );
  Board->Dp.InLength = 0;
  Board->Dp.OutLength = 0;
  Board->Dp.InOffset = 0;
  Board->Dp.OutOffset = 0;
  Board->Dp.InAddr  = 0L;
  Board->Dp.OutAddr = 0L;
  Board->Data.OptimizedMode = 0;
  if( Board->Dp.InBuf ) 
  {
    kfree( Board->Dp.InBuf );
    Board->Dp.InBuf = NULL;
  }
  if( Board->Dp.OutBuf ) 
  {
    kfree( Board->Dp.OutBuf );
    Board->Dp.OutBuf = NULL;
  }
  if( Board->Dp.StateBuf ) 
  {
    kfree( Board->Dp.StateBuf );
    Board->Dp.StateBuf = NULL;
  }
  
  if( Board->Dp.InBufCopy ) 
  {
    kfree( Board->Dp.InBufCopy );
    Board->Dp.InBufCopy = NULL;
  }
  if( Board->Dp.OutBufCopy ) 
  {
    kfree( Board->Dp.OutBufCopy );
    Board->Dp.OutBufCopy = NULL;
  }
  if( Board->Dp.StateBufCopy ) 
  {
    kfree( Board->Dp.StateBufCopy );
    Board->Dp.StateBufCopy = NULL;
  }
  
}

int16 cmi_get_data_descr( struct CntrlExt *CtrlExt,
  T_DATA_DESCR *h_data_descr_list, uint8 image_idx, uint16 offset,
  uint16 *data_descr_size, void *UserDataPtr, int mode )
{
  T_CMI_DESCRIPTOR  *cmi_descriptor = (T_CMI_DESCRIPTOR *) CtrlExt->DprAddr;
  uint8  *h_data_descr_ptr;
  unsigned long dummy;

  h_data_descr_list += image_idx;
  if( offset >= read_io_word(&h_data_descr_list->d_data_size) ) 
  {
    printk( "offset to large %X >= %X",
      offset, read_io_word( &h_data_descr_list->d_data_size ) );
    return E_IF_INVALID_DATA_SIZE;  // offset to large
  }

  /* compute number of bytes to copy */
  if (*data_descr_size > read_io_word( &h_data_descr_list->d_data_size) )
    *data_descr_size = read_io_word( &h_data_descr_list->d_data_size );

  if( offset >(read_io_word(&h_data_descr_list->d_data_size)-*data_descr_size) )
    *data_descr_size = read_io_word(&h_data_descr_list->d_data_size) - offset;

  // get address of data image
  h_data_descr_ptr = (uint8 *) cmi_correct_address(cmi_descriptor,
    read_io_long( &h_data_descr_list->d_data_addr) );

  // check for CMI error
  if( (&h_data_descr_ptr[offset] < (unsigned char*)cmi_descriptor) ||
    (&h_data_descr_ptr[offset]+*data_descr_size >=
    (unsigned char*)cmi_descriptor+DPR_PAGE_SIZE)) {
    printk("profiboard: E_IF_CMI_ERROR\n");
    return( E_IF_CMI_ERROR );
  }

  // get data
  // THIS IS ONLY FOR INTEL BASED SYSTEMS!!!
  // because data below 1M has to be read by read[bwl]
  if( mode == PB_USER_SPACE ) 
  {
    dummy = copy_to_user (UserDataPtr, &h_data_descr_ptr[offset], *data_descr_size);
   // DumpMem(&h_data_descr_ptr[offset], *data_descr_size, "DPRAM input data");
  } 
  else 
  {
    copy_from_io (UserDataPtr, &h_data_descr_ptr[offset], *data_descr_size);
  }

  return E_OK;
}

int16 cmi_wait_for_controller_irq( struct CntrlExt *Board, T_CMI_DESCRIPTOR *cmi_descriptor)
{
  unsigned short volatile *Irq;
  long          timeout;
  int16 ret_val = E_OK;

  PR_DEBUG(D_CMI_L1,"cmi_wait_for_controller_irq()\n");
 // printk("cmi_wait_for_controller: \n");
 
 if ((timeout = pb_sleep_on_timeout (CMI_TIMEOUT2, &Board->Service.wait_for_write)) < 0) 
 {
   if(timeout == -EIO) 
   {
     printk("profiboard(wait): channel %lu controller does not respond!\n", Board->Channel);
     Irq = Board->c_to_h_irq_val; //Board->DprAddr + C_TO_H_IRQ_VAL_OFFSET;
     printk("Host: sema %d - ret_val %d - enable 0x%x c2h-irq 0x%x\n",
            read_io_byte( &cmi_descriptor->h_sema ),
            read_io_byte( &cmi_descriptor->h_ret_val ),
            read_io_byte( &cmi_descriptor->h_int_enable ),
            read_io_byte( Irq ) );
            
            
            Irq = Board->h_to_c_irq_val; //(unsigned char *)Board->DprAddr + H_TO_C_IRQ_VAL_OFFSET;
            
            printk(  "Cont: sema %d - ret_val %d - enable 0x%x h2c-irq 0x%x\n",
            read_io_byte( &cmi_descriptor->c_sema ),
            read_io_byte( &cmi_descriptor->c_ret_val ),
            read_io_byte( &cmi_descriptor->c_int_enable ),
            read_io_byte( Irq ) );
            
            ret_val = E_IF_NO_CNTRL_RES;
   }
   if (timeout == -ERESTARTSYS) 
   {
     if (Board->Service.Exception) 
     {
       printk("got exception!\n");
       ret_val = E_IF_FATAL_ERROR;
     } 
     else 
     {
       printk("profiboard(wait_for_cntrl): ERESTARTSYS\n");
       ret_val = E_IF_NO_CNTRL_RES;
     }
   }
 }
 
 if( read_io_byte( &cmi_descriptor->c_ret_val ) != E_OK   ) 
 {
   printk("cmi_wait_for_controller: c_ret_val=%u line %u done! \n",cmi_descriptor->c_ret_val,__LINE__);
   return( (int16)read_io_byte( &cmi_descriptor->c_ret_val ) );
 }

 // printk("cmi_wait_for_controller: line %u done! \n",__LINE__);
  return (E_OK);
}

// int16 cmi_wait_for_controller( struct CntrlExt *Board, T_CMI_DESCRIPTOR *cmi_descriptor)
// {
//   unsigned short volatile *Irq;
//   long          timeout;
//   int16 ret_val = E_OK;
// 
//   PR_DEBUG(D_CMI_L1,"cmi_wait_for_controller()\n");
//  // printk("cmi_wait_for_controller: \n");
//   
//   while( read_io_byte( &cmi_descriptor->h_sema ) != IDLE) 
//   {
//     if( read_io_byte( &cmi_descriptor->c_ret_val ) != E_OK   ) 
//     {
//        printk("cmi_wait_for_controller: c_ret_val=%u line %u done! \n",cmi_descriptor->c_ret_val,__LINE__);
//       return( (int16)read_io_byte( &cmi_descriptor->c_ret_val ) );
//     }
//     printk("%s line %u\n",__func__, __LINE__);
//     if ((timeout = pb_sleep_on_timeout (CMI_TIMEOUT2, &Board->Service.wait_for_write)) < 0) 
//     {
//       if(timeout == -EIO) 
//       {
//         printk("profiboard(wait): controller does not respond!\n");
//         Irq = Board->c_to_h_irq_val; //Board->DprAddr + C_TO_H_IRQ_VAL_OFFSET;
//         printk("Host: sema %d - ret_val %d - enable 0x%x c2h-irq 0x%x\n",
//                  read_io_byte( &cmi_descriptor->h_sema ),
//                  read_io_byte( &cmi_descriptor->h_ret_val ),
//                  read_io_byte( &cmi_descriptor->h_int_enable ),
//                  read_io_byte( Irq ) );
//         
//         
//         Irq = Board->h_to_c_irq_val; //(unsigned char *)Board->DprAddr + H_TO_C_IRQ_VAL_OFFSET;
//         
//         printk(  "Cont: sema %d - ret_val %d - enable 0x%x h2c-irq 0x%x\n",
//                  read_io_byte( &cmi_descriptor->c_sema ),
//                  read_io_byte( &cmi_descriptor->c_ret_val ),
//                  read_io_byte( &cmi_descriptor->c_int_enable ),
//                  read_io_byte( Irq ) );
//         
//         ret_val = E_IF_NO_CNTRL_RES;
//       }
//       if (timeout == -ERESTARTSYS) 
//       {
//         if (Board->Service.Exception) 
//         {
//           printk("got exception!\n");
//           ret_val = E_IF_FATAL_ERROR;
//         } 
//         else 
//         {
//           printk("profiboard(wait_for_cntrl): ERESTARTSYS\n");
//           ret_val = E_IF_NO_CNTRL_RES;
//         }
//       }
//     }
//     //cli(); dieses cli in der Schleife wird ersatzlos gestrichen kno 21.02.08
//   }
// 
//  // printk("cmi_wait_for_controller: line %u done! \n",__LINE__);
//   return (E_OK);
// }

static int16 cmi_get_data_image_idx( CONTROLLER_EXT *CtrlExt,
                                     T_DATA_DESCR *h_data_descr_list, uint8 image_id, uint8 *image_idx )
{
  uint8 ImageId;

  // loop through the list until end of list found
  PR_DEBUG(D_CMI_L1,"h_data_descr: 0x%p, idx: 0x%x id: 0x%x",
           (void *)h_data_descr_list,*image_idx, image_id);
  h_data_descr_list += *image_idx;
  PR_DEBUG(D_CMI_L1,"--> h_data_descr: 0x%p\n", h_data_descr_list );
  if (h_data_descr_list) 
  {
    while( ((ImageId=read_io_byte( &h_data_descr_list->d_id ))!= 0) ) 
    {
      PR_DEBUG(D_CMI_L1,"ImageId= 0x%x - adr= 0x%lx\n", ImageId,
               read_io_long (&h_data_descr_list->d_data_addr));
      if( ImageId == image_id ) 
      {
      // image found
        PR_DEBUG(D_CMI_L1,"profiboard: found image index %d\n", *image_idx);
        return(E_OK);
      } 
      else 
      {
        h_data_descr_list++;
        (*image_idx)++;
            
      }
    }
  }

  printk(KERN_ERR "profiboard: no image found (id %d)\n", image_id );
  return(E_IF_SERVICE_NOT_SUPPORTED);
}

int16 cmi_get_data( CONTROLLER_EXT *CtrlExt, uint8 data_id, USIGN16 offset, USIGN16 *data_size, VOID *UserDataPtr, int mode) 
{
//   T_CMI_DESCRIPTOR  *cmi_descriptor = (T_CMI_DESCRIPTOR *) CtrlExt->DprAddr;
  T_DATA_DESCR *h_data_descr_list; // list of supported data types
  uint8 data_image_idx = 0;
  int16 ret_val = E_OK;

  // printk("cmi_get_data: offset==%u\n",offset);
  
  h_data_descr_list  = (T_DATA_DESCR *) CtrlExt->h_data_descr_addr;
  //printk(/*D_CMI_L1,*/"cmi_get_data(%d) - 0x%p\n",*data_size,h_data_descr_list);

  // get DATA_IMAGE index
  if( (ret_val = cmi_get_data_image_idx( CtrlExt,
    h_data_descr_list, data_id,&data_image_idx))!= E_OK ) 
  {
    //printk("profiboard: get_data_image_idx %d\n", ret_val );
    return( ret_val );
  }

  // read DATA from DATA_IMAGE
  ret_val = cmi_get_data_descr( CtrlExt,
    h_data_descr_list, data_image_idx, offset, data_size, UserDataPtr,mode);

  return(ret_val);
}

/**********************/
/* Check for telegram */
/**********************/

int16 CmiCheckForTelegram( struct CntrlExt *CtrlExt )
{
  T_CMI_DESCRIPTOR *cmi_descriptor;

  // init pointers
  cmi_descriptor = (T_CMI_DESCRIPTOR  *) CtrlExt->DprAddr;

  // test if telegram arrived
  if( read_io_byte( &cmi_descriptor->c_sema ) == IDLE) {
    PR_DEBUG(D_CMI_L1,"CmiCheckForTelegram: no con ind received\n");
    return( NO_CON_IND_RECEIVED ); // controller is idle -> no telegram
  }
  return( CON_IND_RECEIVED );
}

/************/
/* cmi_read */
/************/

int16 cmi_read( struct CntrlExt *CtrlExt, void *UserDataPtr, USIGN16 *data_len)
{
  T_CMI_DESCRIPTOR *cmi_descriptor;
  int16 ret_val = E_OK;
  T_SDB *c_serv_descr;
  unsigned long flags;

  // init pointers
  cmi_descriptor = (T_CMI_DESCRIPTOR  *) CtrlExt->DprAddr;
  
  GetSpinlock(CtrlExt, &flags, __LINE__);
  
  c_serv_descr = (T_SDB *) CtrlExt->CSdbStart;

  PR_DEBUG(D_CMI_L1,"read %d\n", *data_len );
  if( UserDataPtr == NULL || *data_len < sizeof(T_SDB) ) 
  {
    PR_DEBUG(D_CMI_L1,"profiboard: parameter error %p - %d\n",
        UserDataPtr, *data_len);
    *data_len=0;
    ReleaseSpinlock(CtrlExt, flags, __LINE__);
   // printk("cmi_read: invalid call\n");
    return(E_IF_INVALID_CMI_CALL);
  }
  // test if telegram arrived
  if( read_io_byte( &cmi_descriptor->c_sema ) == IDLE) 
  {
    PR_DEBUG(D_CMI_L3,"profiboard(read): no con ind received\n");
        
    *data_len = 0;
    ReleaseSpinlock(CtrlExt, flags, __LINE__);
   // printk("cmi_read: no event\n");
            
    return(NO_CON_IND_RECEIVED);  // controller is idle -> no telegram
  }
  // get return value form controller
  ret_val = (int16) read_io_byte( &cmi_descriptor->c_ret_val );

  if( *data_len < read_io_word( &cmi_descriptor->c_data_size ) ) 
  {
    PR_DEBUG(D_CMI_L1,"profiboard(read): buf=%d too small, need=%d\n",
      *data_len, read_io_word( &cmi_descriptor->c_data_size ) );
    
   // printk("cmi_read: invalid data size\n");
    ret_val = E_IF_INVALID_DATA_SIZE;
  } 
  else 
  {
    ret_val = (ret_val==E_OK) ? CON_IND_RECEIVED : E_IF_FATAL_ERROR;
  }
  if( *data_len > CtrlExt->CSdbDataLength ) 
  {
    *data_len = CtrlExt->CSdbDataLength;
  }

  //-- copy controller service description block --
  PR_DEBUG(D_CMI_L1,"read: cr: %d l: %d srv: %d id: %d res: %d (%d)\n",
           read_io_word (&c_serv_descr->comm_ref),
           read_io_byte (&c_serv_descr->layer),
           read_io_byte (&c_serv_descr->service),
           read_io_byte (&c_serv_descr->invoke_id),
           read_io_word (&c_serv_descr->result),
           sizeof (T_SDB));
  copy_from_io_to_user (UserDataPtr, c_serv_descr, *data_len);

  PR_DEBUG(D_CMI_L1,"read %d bytes\n", *data_len );

  // reinit board on:  - FMB_EXIT.con
  //                   - FMB_RESET.con
  //                   - FMB_EXCEPTION.ind
  if (read_io_byte (&c_serv_descr->layer) == FMB_USR              &&
      ((read_io_byte (&c_serv_descr->primitive) == CON            &&
       (read_io_byte (&c_serv_descr->service) == FMB_EXIT         ||
        read_io_byte (&c_serv_descr->service) == FMB_RESET     )) ||
      (read_io_byte (&c_serv_descr->primitive) == IND             &&
       read_io_byte (&c_serv_descr->service) == FMB_EXCEPTION)) ) 
  {
    PR_DEBUG(D_CMI_L1,"FMB_EXIT/RESET/EXCEPTION --> reinitboard\n");
  
    printk("FMB_EXIT/RESET/EXCEPTION --> reinitboard\n");

    //-- signal the controller that the information was read --
    write_io_byte( &cmi_descriptor->c_ret_val, E_OK);
    write_io_byte( &cmi_descriptor->c_sema, IDLE);

    PbInitBoard( CtrlExt, 0 );
    PbInitBoard( CtrlExt, C165_RESET );
    cmi_exit_dp( CtrlExt );
    ReleaseSpinlock(CtrlExt, flags, __LINE__);
    cmi_init( CtrlExt );
    GetSpinlock(CtrlExt, &flags, __LINE__);
  } 
  else
  {
    if ((read_io_byte (&c_serv_descr->layer) == DPS_USR          ) &&
      ((read_io_byte (&c_serv_descr->primitive) == CON          &&
      read_io_byte (&c_serv_descr->service) == DPS_EXIT_SLAVE) ||
      (read_io_byte (&c_serv_descr->service) == DPS_EXCEPTION  )) ) 
    {
      PR_DEBUG(D_CMI_L1,"EXIT_SLAVE or DPS_EXCEPTION\n");
      printk("EXIT_SLAVE or DPS_EXCEPTION\n");
      if (read_io_byte (&c_serv_descr->service) == DPS_EXCEPTION) 
      {
	T_EXCEPTION *exc = (T_EXCEPTION *) (c_serv_descr + 1);
	printk(KERN_ERR "DPS_EXCEPTION.ind Task 0x%x: 0x%x 0x%x 0%x\n",
	       read_io_byte (&exc->task_id), read_io_byte (&exc->par1),
	       read_io_word (&exc->par2), read_io_word (&exc->par3));
	
	//-- signal the controller that the information was read --
	write_io_byte( &cmi_descriptor->c_ret_val, E_OK);
	write_io_byte( &cmi_descriptor->c_sema, IDLE);
	
      } 
      else 
      {
	if( c_serv_descr->service==DPS_EXIT_SLAVE ) 
	{
	  PbInitBoard( CtrlExt, C165_RESET );
	} 
	else 
	{
	  printk("INVALID SERVICE: %d(0x%x)/%d(0x%x) (prim=%d(0x%x)\n",
	  read_io_byte (&c_serv_descr->layer), read_io_byte (&c_serv_descr->layer),
	  read_io_byte (&c_serv_descr->service), read_io_byte (&c_serv_descr->service),
	  read_io_byte (&c_serv_descr->primitive), read_io_byte (&c_serv_descr->primitive));
	  PbInitBoard( CtrlExt, C165_RESET );
	}
      }
      PR_DEBUG(D_CMI_L1,"DPS_EXIT_SLAVE/DPS_EXCEPTION --> reinitboard\n");
      ReleaseSpinlock(CtrlExt, flags, __LINE__);
      cmi_init( CtrlExt );
      GetSpinlock(CtrlExt, &flags, __LINE__);
    } 
    else 
    {
      //-- signal the controller that the information was read --
      write_io_byte( &cmi_descriptor->c_ret_val, E_OK);
      write_io_byte( &cmi_descriptor->c_sema, IDLE);
      
      // activate controller interrupt
      if( read_io_byte( &cmi_descriptor->c_int_enable ) & 0x02 ) 
      {
	irq_to_cntrl(CtrlExt, ACK_IRQ);
      }
    }
  }
  ReleaseSpinlock(CtrlExt, flags, __LINE__);
 // printk("cmi_read: ok; datalen=%u\n",*data_len);
  return(ret_val);
}

/*************/
/* cmi_write */
/*************/

int cmi_write( struct CntrlExt *CtrlExt, void *UserDataPtr, size_t *data_len, int MemFlag)
{
  T_CMI_DESCRIPTOR *cmi_descriptor;
  int16 ret_val;
  T_SDB *service;
  unsigned long flags;

  // printk("cmi_write: \n");
  PR_DEBUG(D_CMI_L1,"profiboard: cmi_write( %p, %d )\n",
      UserDataPtr,*data_len);
  
  
  
  cmi_descriptor = (T_CMI_DESCRIPTOR *)CtrlExt->DprAddr;
  
  if( UserDataPtr == NULL || *data_len < sizeof(T_SDB) ) 
  {
    PR_DEBUG(D_CMI_L1,"profiboard: parameter error %p - %d\n",
        UserDataPtr, *data_len);
    *data_len=0;
    return(E_IF_INVALID_CMI_CALL);
  }
  
  
  if( *data_len > CtrlExt->HSdbDataLength ) 
  {
    *data_len = CtrlExt->HSdbDataLength;
  }

  // XXX - THIS IS VALID ONLY FOR X86!
  PR_DEBUG(D_CMI_L1,"profiboard: memcpy_fromfs( 0x%p, 0x%p )\n",
    CtrlExt->HSdbStart, UserDataPtr );
  
  
  service = (T_SDB *)CtrlExt->HSdbStart;
  GetSpinlock(CtrlExt, &flags, __LINE__);
  
  if( MemFlag == PB_USER_SPACE ) 
  {
    if(CtrlExt->HSdbBuffer)
    {
      ReleaseSpinlock(CtrlExt, flags, __LINE__);
      dummy = copy_from_user (CtrlExt->HSdbBuffer,(char*)UserDataPtr, *data_len);
      GetSpinlock(CtrlExt, &flags, __LINE__);
      memcpy (CtrlExt->HSdbStart, CtrlExt->HSdbBuffer, *data_len);
    }
    else
    {
      printk("ERROR: cmi_write: Sdb buffer is a null pointer!");
    }
    
  } 
  else 
  {
    copy_to_io (CtrlExt->HSdbBuffer, (char*)UserDataPtr, *data_len);
  }
  
  write_io_word( &cmi_descriptor->h_param_size, ((__u16) sizeof(T_SDB) ) );
  if (read_io_byte (&service->layer) == DP || read_io_byte (&service->layer) == DP_USR) 
  {
    if( *data_len > 264 )
      *data_len = 264;
  }
  write_io_word( &cmi_descriptor->h_data_size, (__u16)(*data_len-sizeof(T_SDB)));

  // take care, that only DP_AAM_IO_BLOCKS mode is used!
  if ((read_io_byte (&service->layer) == DP              ) &&
      (read_io_byte (&service->service) == DP_INIT_MASTER) &&
      (read_io_byte (&service->primitive) == REQ         ) &&
      (*data_len >= (sizeof(T_SDB) + sizeof(T_DP_INIT_MASTER_REQ)))) 
  {
    T_DP_INIT_MASTER_REQ *Req =(T_DP_INIT_MASTER_REQ *) CtrlExt->HSdbStart
      +sizeof(T_SDB);
    if(read_io_byte (&Req->slave_io_address_mode) != DP_AAM_IO_BLOCKS ) 
    {
      PR_DEBUG(D_CMI_L1,
      "profiboard: slave_io_address_mode (%d) != DP_AAM_IO_BLOCKS\n",
          read_io_byte (&Req->slave_io_address_mode));
      write_io_byte (&Req->slave_io_address_mode, DP_AAM_IO_BLOCKS);
    }
  }

  
  
  //-- set host semaphore to BUSY --
  write_io_byte( &cmi_descriptor->h_sema, BUSY);

  //-- activate controller interrupt --
  if( read_io_byte(&cmi_descriptor->c_int_enable) & 0x01 ) 
  {
    irq_to_cntrl( CtrlExt, REQ_IRQ );
  }

  //-- wait until the semaphore in host cmi common block is set to IDLE --
  ReleaseSpinlock(CtrlExt, flags, __LINE__);
//   printk("%s line %u, cmi_wait_for_controller\n",__func__, __LINE__);
  if ((ret_val = cmi_wait_for_controller_irq(CtrlExt, cmi_descriptor)) != E_OK) 
  {
    GetSpinlock(CtrlExt, &flags, __LINE__);
    write_io_byte( &cmi_descriptor->h_sema, IDLE);
    ReleaseSpinlock(CtrlExt, flags, __LINE__);
    return( ret_val );
  }
  GetSpinlock(CtrlExt, &flags, __LINE__);

  // set semaphore in host cmi common block again to IDLE
  write_io_byte( &cmi_descriptor->h_sema, IDLE );

  // return the result of controller
  ret_val = (int16) read_io_byte( &cmi_descriptor->h_ret_val );
  write_io_byte( &cmi_descriptor->h_ret_val, E_OK);

  if (ret_val == 0xFF) 
  {
    ret_val = E_IF_FATAL_ERROR;
    //printk("profiboard(cmi_write): bad cntrl ret_val\n");
  }

  ReleaseSpinlock(CtrlExt, flags, __LINE__);
  PR_DEBUG(D_CMI_L1,"profiboard: wrote %d bytes, ret=%d\n",
      *data_len, ret_val );
  
  return(ret_val);
}

int cmi_init( struct CntrlExt *CtrlExt )
{
  volatile T_CMI_DESCRIPTOR *cmi_descriptor;  // CMI descriptor
  int timeout;        // timeout time
  PB_ADDRESS address;

 // printk("cmi_init: Port: 0x%x Irq: %d Mem: 0x%lx\n",CtrlExt->port, CtrlExt->interrupt, CtrlExt->DprAddr );
  
  PR_DEBUG(D_CMI_L1,"Port: 0x%x Irq: %d Mem: 0x%lx\n", CtrlExt->port, CtrlExt->interrupt, CtrlExt->DprAddr ); 
  CtrlExt->Service.Exception = 0; // set CMI host base address
  cmi_descriptor = (T_CMI_DESCRIPTOR  *) CtrlExt->DprAddr;

//   // intialize CMI host and controller ready mask
//   write_io_long(&cmi_descriptor->h_ready_mask, 0);
//   write_io_long(&cmi_descriptor->c_ready_mask, 0);
// 
//   // intialize CMI host and controller return value
//   write_io_byte(&cmi_descriptor->h_ret_val, 0);
//   write_io_byte(&cmi_descriptor->c_ret_val, 0);
// 
//   // set the cmi host descriptor block parameters
//   write_io_long(&cmi_descriptor->h_base_address, ((USIGN32) CtrlExt->DprAddr));
  cmi_descriptor->h_ready_mask = 0;
  cmi_descriptor->c_ready_mask = 0;
  cmi_descriptor->h_ret_val = 0;
  cmi_descriptor->c_ret_val = 0;
  cmi_descriptor->h_base_address = (USIGN32) CtrlExt->DprAddr;
  
  if( CtrlExt->TypeOfFW == PB_DP_SLAVE_FW )
  {
    write_io_byte(&cmi_descriptor->h_id, CMI_V410_DP_SLAVE );
  }
  else
  {
    write_io_byte(&cmi_descriptor->h_id, CMI_V410 );
  }

  switch( CtrlExt->interrupt ) 
  {
    // the PROFIcard FW does not check the interrupt set in the CMI
    // so all interrupts can be used
    case  5:
      write_io_byte(&cmi_descriptor->h_int_enable, (POLL_INT_MODE | INT_5));
      break;
    case 10:
      write_io_byte(&cmi_descriptor->h_int_enable, (POLL_INT_MODE | INT_10));
      break;
    case 11:
      write_io_byte(&cmi_descriptor->h_int_enable, (POLL_INT_MODE | INT_11));
      break;
    case 12:
      write_io_byte(&cmi_descriptor->h_int_enable, (POLL_INT_MODE | INT_12));
      break;
    case 15:
      write_io_byte(&cmi_descriptor->h_int_enable, (POLL_INT_MODE | INT_15));
      break;
    default:
      write_io_byte(&cmi_descriptor->h_int_enable, POLL_INT_MODE);
      break;
    }
  write_io_byte(&cmi_descriptor->h_address_swap_mode , ADDRESS_SWAP_MODE);
  write_io_byte(&cmi_descriptor->h_state , INIT_MODE);

  reset_cntrl_irq( CtrlExt );

  write_io_byte(&cmi_descriptor->h_sema, IDLE);
//   write_io_long(&cmi_descriptor->h_ready_mask, ((uint32) H_READY_MASK));
  cmi_descriptor->h_ready_mask = (USIGN32) H_READY_MASK;
  
  printk("%s line %u start waiting!\n", __func__, __LINE__);
  StartTimeout (&timeout, CMI_TIMEOUT);
  // wait until controller ready mask is set
  while(cmi_descriptor->c_ready_mask != C_READY_MASK) 
  {
    //     write_io_long(&cmi_descriptor->h_ready_mask, ((uint32) H_READY_MASK));
    cmi_descriptor->h_ready_mask = H_READY_MASK;
    
    if( SleepAWhile(&timeout) ) 
    {
      //DpramOut(CtrlExt, 0 );
      printk("%s line %u stop waiting!\n", __func__, __LINE__);
      return(-1);
    }
  }
  printk("%s line %u stop waiting!\n", __func__, __LINE__);
  
 //printk("profiboard: controller found\n");

  // set host ready mask again
  //   write_io_long(&cmi_descriptor->h_ready_mask, ((uint32) H_READY_MASK));
  cmi_descriptor->h_ready_mask = H_READY_MASK;

  // determine host param_addr and data_addr
  CtrlExt->HSdbStart = (void *)cmi_correct_address((T_CMI_DESCRIPTOR*)  cmi_descriptor, cmi_descriptor->h_param_addr);
  //   write_io_long( &cmi_descriptor->h_param_addr, (uint32)CtrlExt->HSdbStart );
  cmi_descriptor->h_param_addr = (USIGN32)CtrlExt->HSdbStart;
  
  address = cmi_correct_address((T_CMI_DESCRIPTOR*) cmi_descriptor, cmi_descriptor->h_data_addr);
  if( (CtrlExt->HSdbStart + sizeof(T_SDB))!=(void*)address ) 
  {
    printk(KERN_ERR "CMI_ERROR: h_param_addr %p %p %d\n", (void *)address,
        CtrlExt->HSdbStart, (USIGN32)sizeof(T_SDB));
    return( -1 );
  }
//   write_io_long( &cmi_descriptor->h_data_addr, address );
  cmi_descriptor->h_data_addr = address;
  address = cmi_correct_address((T_CMI_DESCRIPTOR*) cmi_descriptor, cmi_descriptor->h_data_descr_addr);
  
//   write_io_long( &cmi_descriptor->h_data_descr_addr, address );
  cmi_descriptor->h_data_descr_addr = address;
  CtrlExt->h_data_descr_addr = address;
  
  CtrlExt->HSdbDataLength = read_io_word (&cmi_descriptor->h_data_size) + sizeof(T_SDB);
  PR_DEBUG(D_CMI_L1,"HSdbStart: %p, HSdbDataLength: 0x%x\n", CtrlExt->HSdbStart,
      CtrlExt->HSdbDataLength );

  CtrlExt->CSdbStart = (T_SDB *) cmi_correct_address((T_CMI_DESCRIPTOR*) cmi_descriptor, cmi_descriptor->c_param_addr); 
  
  CtrlExt->CSdbDataLength = read_io_word (&cmi_descriptor->c_data_size) + sizeof(T_SDB);
  if(CtrlExt->HSdbBuffer)
  {
    kfree(CtrlExt->HSdbBuffer);
    CtrlExt->HSdbBuffer = kmalloc(CtrlExt->CSdbDataLength, GFP_ATOMIC);
  }
  else
  {
    CtrlExt->HSdbBuffer = kmalloc(CtrlExt->CSdbDataLength, GFP_ATOMIC);
  }
  PR_DEBUG(D_CMI_L1,"CSdbStart: %p, CSdbDataLength: 0x%x\n", CtrlExt->CSdbStart,
      CtrlExt->CSdbDataLength );

  write_io_byte(&cmi_descriptor->h_state, CONFIG_MODE);

  printk("%s line %u start waiting!\n", __func__, __LINE__);
  StartTimeout( &timeout, CMI_TIMEOUT);
  // wait until controller state is 'CONFIG_MODE'
  while (read_io_byte(&cmi_descriptor->c_state) != CONFIG_MODE) 
  {
    if( read_io_byte(&cmi_descriptor->c_ret_val) != E_OK) 
    {
      printk("%s line %u stop waiting!\n", __func__, __LINE__);
      return(read_io_byte(&cmi_descriptor->c_ret_val));
    }
    if( SleepAWhile(&timeout) ) 
    {
      printk(KERN_ERR "profiboard: controller not responding!\n");
      DpramOut(CtrlExt, 0 );
      printk("%s line %u stop waiting!\n", __func__, __LINE__);
      return(E_IF_NO_CNTRL_RES);
    }
  }
  
  printk("%s line %u stop waiting!\n", __func__, __LINE__);

  //DpramOut(CtrlExt, 0 );
  write_io_byte(&cmi_descriptor->h_ret_val, E_OK);
//  printk("cmi_init: done!\n");
  PR_DEBUG(D_CMI_L1, "profiboard ok.\n");
  return(0);
}

/***************************/
/* DP-Slave data interface */
/***************************/

static unsigned char input_data_copy[DP_MAX_INPUT_DATA_LEN];

int16 cmi_set_dps_input_data( struct CntrlExt *Board, uint16 InputDataSize,
    uint8 *FSInputData, uint8 *FSInputState )
{
  int16 Result = E_OK;
  unsigned long dummy;
  unsigned long flags;
//   T_CMI_DESCRIPTOR * cmi_desk = (T_CMI_DESCRIPTOR *) Board->DprAddr;
  T_DATA_DESCR *pHostDataDescrList;//list of supported data types
  __u8 InputDataLength, InputState, InputStateCopy;
  DECLARE_WAITQUEUE (wait, current);

  T_DPS_INPUT_DATA_HEADER *DpramSlaveIn; 

  DpramSlaveIn = (T_DPS_INPUT_DATA_HEADER *)Board->SlaveIn.InStruct;
  pHostDataDescrList = (T_DATA_DESCR *) Board->h_data_descr_addr;
  
//   if(cmi_desk->h_data_descr_addr != Board->h_data_descr_addr)
//   {
//     printk("ERROR: 0x%lx <> 0x%lx\n", (PB_ADDRESS)cmi_desk->h_data_descr_addr, Board->h_data_descr_addr);
//   }
  
  InputDataLength = read_io_byte( &DpramSlaveIn->input_data_len );
  InputState = read_io_byte( &DpramSlaveIn->input_state );
  PR_DEBUG(D_CMI_L1,"cmi_set_dps_input_data( %p %p %d - state: 0x%x)\n",
      DpramSlaveIn, pHostDataDescrList, InputDataSize, InputState );
  if( (USIGN16) InputDataLength == InputDataSize) 
  {
    //lock DPS_SLAVE_INPUT_IMAGE (DPS ready indicator after access)
    //cli(); solved 21.2.08
    dummy = copy_from_user (input_data_copy, FSInputData, InputDataLength);
   
    GetSpinlock(Board, &flags, __LINE__);
    cmi_lock_data_image (Board,pHostDataDescrList, SYNC_BY_LOCK, Board->SlaveIn.ImageIdx); 
    memcpy (&DpramSlaveIn->input_data[0], input_data_copy, InputDataLength);
    add_wait_queue( &Board->SlaveIn.QInputAtController,&wait);
    current->state = TASK_INTERRUPTIBLE;
    irq_to_cntrl(Board, DPS_SLAVE_INP_REQ_IRQ );
    
    ReleaseSpinlock(Board, flags, __LINE__);
    // sleep, until interrupt arrived
    // sti();
    schedule(); // what a dangerous hack ???
    //cli();  solved 21.2.08 kno
    GetSpinlock(Board, &flags, __LINE__);
    
//     for(i=0; i<1000; i++)
//     {
//       if(Board->h_to_c_irq_val)
// 	break;
//       udelay(1);
//     }
//     ReleaseSpinlock(Board, flags, __LINE__);
//     GetSpinlock(Board, &flags, __LINE__);
    
    remove_wait_queue(&Board->SlaveIn.QInputAtController, &wait);
    Result = cmi_wait_for_unlock_data_image(Board, pHostDataDescrList,Board->SlaveIn.ImageIdx); 
    if(Result == E_OK) 
    {
      // get DPS_SLAVE_INPUT state
      InputState = read_io_byte( &DpramSlaveIn->input_state );
      InputStateCopy = InputState;
//       dummy = copy_to_user (FSInputState, &InputState, sizeof(InputState));
      // update DPS_SLAVE_INPUT state    
      InputState &= ~DPS_INPUT_STATE_FREEZE_COMMAND;
      write_io_byte( &DpramSlaveIn->input_state, InputState );
    }
    ReleaseSpinlock(Board, flags, __LINE__);
    dummy = copy_to_user (FSInputState, &InputStateCopy, sizeof(InputState));
  } 
  else 
  {
    //cmi_unlock_data_image( Board,pHostDataDescrList,
    //    SYNC_BY_LOCK,DpsInputDataImageIdx);
    PR_DEBUG( D_CMI_L1, "InputDataLength %d InputDataSize %d\n",
        InputDataLength, InputDataSize );
    Result = E_IF_INVALID_DATA_SIZE;
  }
    return (Result);
}

int cmi_get_dps_input_data(struct CntrlExt *Board,__u16 *pSize,__u8 *pData )
{
  int Result = E_OK;
  T_DPS_INPUT_DATA_HEADER *DpramSlaveIn;
  __u8 InputDataLength;
  __u8 InputState;
  unsigned long flags;

  // get DPS_SLAVE_INPUT data length
  DpramSlaveIn = (T_DPS_INPUT_DATA_HEADER *)Board->SlaveIn.InStruct;
  InputDataLength = read_io_byte( &DpramSlaveIn->input_data_len );

  if( InputDataLength <= (*pSize-1) ) 
  {
    // read input state from DPRAM
    InputState = read_io_byte( &DpramSlaveIn->input_state );
    // read input data from DPRAM
   // save_flags( flags );
   // cli(); solved 21.2.08 kno
    GetSpinlock(Board, &flags, __LINE__);
    copy_from_io_to_user (pData, &DpramSlaveIn->input_data[0], InputDataLength);
    copy_from_io_to_user (pData+InputDataLength, &InputState, 1);
    ReleaseSpinlock(Board, flags, __LINE__);
   // restore_flags( flags );
    //write_io_byte( &DpramSlaveIn->input_data[InputDataLength],InputState );
    *pSize = InputDataLength+1;
    PR_DEBUG( D_CMI_L1,"cmi_get_dps_input_data: *pSize=%d/(%d)\n",
              *pSize, InputDataLength );
  } 
  else 
  {
    printk("cmi_get_dps_input_data: data len = %u > %u\n",InputDataLength,(*pSize-1));
    Result = E_IF_INVALID_DATA_SIZE;
  }
  if (Result != E_OK) 
  {
    *pSize = 0;
  }
  return (Result);
}

/***********************************************************************/
/* Function to get the pointer for the input and the output structures */
/***********************************************************************/

int cmi_init_slave( struct CntrlExt *Board )
{
  T_CMI_DESCRIPTOR  *CmiD = (T_CMI_DESCRIPTOR *) Board->DprAddr;
  T_DATA_DESCR *h_data_descr_list; // list of supported data types
  uint8 OutputImageIdx = 0;
  uint8 InputImageIdx = 0;
  int16 ret_val = E_OK;

  PR_DEBUG(D_CMI_L1,"cmi_init_slave\n");
  memset(&Board->SlaveOut.Cache.output_data[0],0,sizeof(Board->SlaveOut.Cache.output_data));
  Board->SlaveOut.Cache.output_state = 0;

  h_data_descr_list=(T_DATA_DESCR *) Board->h_data_descr_addr;
  // check the address read in the cmi
  if( (*(USIGN32*)h_data_descr_list == 0xFFFFFFFF) ||
      (*(USIGN32*)h_data_descr_list == 0x00000000)
    ) 
  {
    // the addresses can be 0xFFFFFFFF if DPRAM is disabled or
    // if a PROFIcard is yet removed
    printk("profiboard: wrong image list 0x%x\n",
      read_io_long((long*) h_data_descr_list) );
    return E_IF_FATAL_ERROR;
  }

  if( (ret_val = cmi_get_data_image_idx( Board, h_data_descr_list,
      ID_DPS_SLAVE_OUTPUT_IMAGE,&OutputImageIdx))!= E_OK ) {
   // printk("profiboard: cmi_get_data_image_idx out %d\n", ret_val );
    return( ret_val );
  }
  h_data_descr_list += OutputImageIdx;

  // get address of output image
  Board->SlaveOut.OutStruct = (void *) cmi_correct_address(CmiD, h_data_descr_list->d_data_addr);
  
  if( (Board->SlaveOut.OutStruct < (void *)CmiD) ||
    (Board->SlaveOut.OutStruct >= (void *)CmiD+DPR_PAGE_SIZE)) 
  {
    printk("profiboard: E_IF_CMI_ERROR\n");
    return( E_IF_CMI_ERROR );
  }

  h_data_descr_list=(T_DATA_DESCR *) Board->h_data_descr_addr;
  if( (ret_val = cmi_get_data_image_idx( Board, h_data_descr_list,
      ID_DPS_SLAVE_INPUT_IMAGE,&InputImageIdx))!= E_OK ) 
  {
   // printk("profiboard: cmi_get_data_image_idx in %d\n", ret_val );
    return( ret_val );
  }

  PR_DEBUG(D_CMI_L1,"In-/OutputImageIdx = %d/%d\n",
      InputImageIdx, OutputImageIdx );
  Board->SlaveIn.ImageIdx = InputImageIdx;
  h_data_descr_list += InputImageIdx;

  // get address of output image
  Board->SlaveIn.InStruct = (void *) cmi_correct_address(CmiD, h_data_descr_list->d_data_addr);
  if( (Board->SlaveIn.InStruct < (void *)CmiD) ||
    (Board->SlaveIn.InStruct >= (void *)CmiD+DPR_PAGE_SIZE)) {
    printk("profiboard: E_IF_CMI_ERROR\n");
    return( E_IF_CMI_ERROR );
  }
    PR_DEBUG(D_CMI_L1,"cmi_init( out=%p, in=%p )\n", Board->SlaveOut.OutStruct,
      Board->SlaveIn.InStruct );
  return E_OK;
}

int cmi_get_dps_output_data( struct CntrlExt *Board )
{
  int Result = E_OK;
  uint8 state;
  T_DPS_OUTPUT_DATA_HEADER *DpramSlaveOut;

  DpramSlaveOut = (T_DPS_OUTPUT_DATA_HEADER *)Board->SlaveOut.OutStruct;

  if(read_io_byte(&DpramSlaveOut->output_data_len)<=DP_MAX_OUTPUT_DATA_LEN ) 
  {
    state = read_io_byte( &DpramSlaveOut->output_state );
   // printk("cmi_get_dps_output_data: state=0x%x\n",state);
    Board->SlaveOut.Cache.output_data_len =
      read_io_byte(&DpramSlaveOut->output_data_len);
    Board->SlaveOut.Cache.output_state = state;
    state &= ~(DPS_OUTPUT_STATE_NEW_DATA|DPS_OUTPUT_STATE_GLOBAL_CONTROL);
    write_io_byte( &DpramSlaveOut->output_state, state );
    copy_from_io (&Board->SlaveOut.Cache.output_data[0],
                  &DpramSlaveOut->output_data[0],
                  Board->SlaveOut.Cache.output_data_len );
  } 
  else 
  {
    printk(KERN_ERR "too much output data (%d)\n",
        DpramSlaveOut->output_data_len );
    Result = E_IF_INVALID_DATA_SIZE;
  }
    return(Result);
}

static int16 cmi_wait_for_unlock_data_image( struct CntrlExt *CtrlExt,
  T_DATA_DESCR *h_data_descr_list, uint8 image_idx )
{
  long timeout=1000; // maximal 1 millisecond

  while(read_io_byte(&h_data_descr_list[image_idx].d_lock) != D_DATA_IDLE) 
  {
    // XXX - Busy Loop? Arrrrgghhh
    udelay( 1 ); // wait 1 microsecond
    if( timeout--== 0 ) 
    {
     // printk("cmi_wait_for_unlock_data_image: timeout\n");
     // printk("%p, %p, %d\n", CtrlExt, h_data_descr_list, image_idx );
            return(E_IF_NO_CNTRL_RES);
        }
    }
  return(E_OK);
}

static int16 cmi_lock_data_image( struct CntrlExt *CtrlExt,
  T_DATA_DESCR *h_data_descr_list, int SyncBySema, uint8 image_idx )
{
  write_io_byte(&h_data_descr_list[image_idx].d_lock, D_DATA_BUSY);
  return(E_OK);
}

#if 0
static uint8 cmi_unlock_data_image( struct CntrlExt *CtrlExt,
  T_DATA_DESCR *h_data_descr_list, int SyncBySema, uint8 image_idx )
{
      write_io_byte(&h_data_descr_list[image_idx].d_lock,D_DATA_IDLE);
      return(D_DATA_IDLE);
} // cmi_unlock_data_image
#endif
