/*****************************************************************************/
/*                                                                           */
/*                              SOFTING AG                                   */
/*                        Richard-Reitzner-Allee 6                           */
/*                              D-85540 Haar                                 */
/*                      Phone: (++49)-(0)89-45656-0                          */
/*                      Fax:   (++49)-(0)89-45656-399                        */
/*                                                                           */
/*                  Copyright (C) SOFTING AG 1995-2004                       */
/*                                                                           */
/* See file CREDITS for list of people who contributed to this               */
/* project.                                                                  */
/*                                                                           */
/* This program is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU General Public License as            */
/* published by the Free Software Foundation; either version 2 of            */
/* the License, or (at your option) any later version.                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston,                    */
/* MA 02111-1307, USA.                                                       */
/*                                                                           */
/* email: support.automation@softing.com                                     */
/* Softing AG, Richard-Reitzner-Allee 6, 85540 Haar, Germany                 */
/*                                                                           */
/*****************************************************************************/
/*****************************************************************************/
/*                                                                           */
/*  Author      : SOFTING                                                    */
/*                                                                           */
/*  CHANGE_NOTES                                                             */
/*                                                                           */
/*  date      name      change                                               */
/*  -----------------------------------------------------------------------  */
/*****************************************************************************/
/*****************************************************************************/

#ifndef _CMI_H_
#define _CMI_H_

#define CMI_TIMEOUT  4000
#define CMI_TIMEOUT2 3000

#define PB_KERNEL_SPACE 0
#define PB_USER_SPACE 1


//typedef void VOID;
//typedef char INT8;
typedef char STRINGV;
typedef unsigned char PB_BOOL;
typedef USIGN32 uint32;
typedef USIGN16 uint16;
typedef USIGN8 uint8;
typedef INT16 int16;

#ifndef IN
#define IN
#define OUT
#endif

#define INT_5  0x10
#define INT_10 0x20
#define INT_11 0x30
#define INT_12 0x40
#define INT_15 0x50

#define VERSION_STRING_LENGTH        100  /* length of version string buffer */

//-----------------------------------------------------------------------------
// defines

// timeout to wait for controller response
#define CMI_TIME3000    // timeout after 3 seconds

//-----------------------------------------------------------------------------
// data structures

// cmi data description block
typedef struct T_DATA_DESCR
{
  USIGN8 volatile d_id;          // identifier to describe the data description
  USIGN8 volatile d_lock;        // semaphore for lock (old)
#ifdef PB
  USIGN8 volatile d_sema_c;      // semaphore controller
  USIGN8 volatile d_sema_h;      // semaphore host
#endif
  USIGN16 volatile d_data_size;  // data block size
  USIGN32 volatile d_data_addr;  // data block address
}  __attribute__ ((packed)) T_DATA_DESCR;


// structure of the data type DP_SLAVE_IO_CONFIG_IMAGE
typedef struct _T_DP_SLAVE_IO_CONFIG_IMAGE
{
  USIGN16 volatile offset_inputs; // offset of input area in AAM_IO_BLOCKS
  USIGN16 volatile offset_outputs;  // offset of output area in AAM_IO_BLOCKS
  USIGN16 volatile number_inputs; // number of all slave inputs
  USIGN16 volatile number_outputs;  // number of all slave outputs
} __attribute__ ((packed)) T_DP_SLAVE_IO_CONFIG_IMAGE;


// CMI descriptor
typedef struct T_CMI_DESCRIPTOR
{
  USIGN32 volatile h_ready_mask;       // ready mask
  USIGN32 volatile h_base_address;     // base address of DPR (lower 32 bits only)
  USIGN8 volatile h_id;                // type and version of host software
  USIGN8 volatile h_int_enable;        // call by interrupt
  USIGN8 volatile h_address_swap_mode; // address mode
  USIGN8 volatile h_state;             // page addressing
  USIGN32 volatile h_param_addr;       // parameter block address (SDB)
  USIGN32 volatile h_data_addr;        // data block address
  USIGN16 volatile h_param_size;       // parameter block size
  USIGN16 volatile h_data_size;        // data block size
  USIGN8 volatile h_sema;              // semaphore
  USIGN8 volatile h_ret_val;           // return value
  USIGN8 volatile h_head;              // reserved for a circular request block buffer
  USIGN8 volatile h_tail;              // reserved for a circular request block buffer
  USIGN32 volatile h_data_descr_addr;  // data descriptior list address (DATA_DESCR)
  USIGN32 volatile c_ready_mask;       // ready mask
  USIGN32 volatile c_base_address;     // base address of DPR
  USIGN8 volatile c_id;                // type and version of host software
  USIGN8 volatile c_int_enable;        // call by interrupt
  USIGN8 volatile c_address_swap_mode; // address mode
  USIGN8 volatile c_state;             // page addressing
  USIGN32 volatile c_param_addr;       // parameter block address (SDB)
  USIGN32 volatile c_data_addr;        // data block address
  USIGN16 volatile c_param_size;       // parameter block size
  USIGN16 volatile c_data_size;        // data block size
  USIGN8 volatile c_sema;              // semaphore
  USIGN8 volatile c_ret_val;           // return value
  USIGN8 volatile c_head;              // reserved for a circular request block buffer
  USIGN8 volatile c_tail;              // reserved for a circular request block buffer
  USIGN32 volatile c_data_descr_addr;  // data descriptior list address (DATA_DESCR)
} __attribute__ ((packed)) T_CMI_DESCRIPTOR;

INT16 cmi_get_data(
  struct CntrlExt *CtrlExt, // controller extension
  USIGN8 data_id,     // data image id
  USIGN16 offset,     // offset in data image
  USIGN16 *data_size, // size of data to read
  VOID *UserDataPtr,    // pointer to data
  int mode);          // reading to user or to kernel space

INT16 cmi_get_data_2_kernel(
  struct CntrlExt *CtrlExt, // controller extension
  USIGN8 data_id,     // data image id
  USIGN16 offset,     // offset in data image
  USIGN16 *data_size, // size of data to read
  VOID *UserDataPtr);     // pointer to data

extern int cmi_write( struct CntrlExt *CtrlExt, void *UserDataPtr,
  size_t *data_len, int MemFlag );
extern int16 cmi_read ( struct CntrlExt *CtrlExt, void *UserDataPtr,
  USIGN16 *data_len );
extern int cmi_init_dp( struct CntrlExt *Board );
extern void cmi_exit_dp( struct CntrlExt *Board );
extern int16 CmiCheckForTelegram( struct CntrlExt *CtrlExt );
extern int cmi_get_dps_output_data(struct CntrlExt *Board);
extern int cmi_get_dps_input_data(struct CntrlExt *,uint16 *blen, uint8 *buf);
extern int16 cmi_set_dps_input_data(struct CntrlExt *, uint16 InLen,
    uint8 *pInputData, uint8 *pInputState );
    
/*****************************************************************************/
/*** INLINE FUNCTIONS ********************************************************/
/*****************************************************************************/



long pb_sleep_on_timeout (long timeout, WAIT_STRUCT *wait_reason);


#endif
