/*****************************************************************************/
/*                                                                           */
/*                              SOFTING AG                                   */
/*                        Richard-Reitzner-Allee 6                           */
/*                              D-85540 Haar                                 */
/*                      Phone: (++49)-(0)89-45656-0                          */
/*                      Fax:   (++49)-(0)89-45656-399                        */
/*                                                                           */
/*                  Copyright (C) SOFTING AG 1995-2004                       */
/*                                                                           */
/* See file CREDITS for list of people who contributed to this               */
/* project.                                                                  */
/*                                                                           */
/* This program is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU General Public License as            */
/* published by the Free Software Foundation; either version 2 of            */
/* the License, or (at your option) any later version.                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston,                    */
/* MA 02111-1307, USA.                                                       */
/*                                                                           */
/* email: support.automation@softing.com                                     */
/* Softing AG, Richard-Reitzner-Allee 6, 85540 Haar, Germany                 */
/*                                                                           */
/*****************************************************************************/
/*****************************************************************************/
/*                                                                           */
/*  Author      : SOFTING                                                    */
/*                                                                           */
/*  CHANGE_NOTES                                                             */
/*                                                                           */
/*  date      name      change                                               */
/*  -----------------------------------------------------------------------  */
/*****************************************************************************/
/*****************************************************************************/

/***************************************************************************/
/*                  PROFIboard Firmware Download                           */
/*                                                                         */
/* The firmware consists out of 3 files:                                   */
/* 1. b_boat, the loader                                                   */
/* 2. upboard, the program to write the firmware into the flash            */
/* 3. pbfw, the firmware itself                                            */
/*                                                                         */
/* Each profiboard contains a so called bootblock. If the board is         */
/* powered, it executes the code of the bootblock. This codes waits at the */
/* start of the DPRAM for a magic key word (h_ready_mask). If it finds     */
/* 0x________ it starts the firmware, if it finds 'PROG' it starts a       */
/* program found at location DPRAM startaddress + 0x100.                   */
/* Loading a new firmware is now done in the following way:                */
/* b_boat and upboard has to be written to a certain location inside the   */
/* DPRAM. Then the board is resetted and at the beginning of the DPRAM the */
/* magic code PROG has to be written.                                      */
/*                                                                         */
/* After this the firmware itself can be written record by record.         */
/* (A handshake mechanism exists to help).                                 */
/* Because different files are needed, but the user interface to the       */
/* driver should be:                                                       */
/*                cat firmwarefile > /dev/pbboard0                         */
/* all files have to concatenated into one file, which can be downloaded.  */
/* The fileformat for this file is tar.                                    */
/* The board-write function has different states:                          */
/*   0. tar mode                                                           */
/*   1. fw mode (load b_board or/and upboard)                              */
/*   2. gap mode (to read the bytes to the next block boundary)            */
/*                                                                         */
/* This module (especially PbWriteBoard) is rather sophisticated           */
/* (someone else might say complicated ;-) ). The problem is, that the     */
/* three files, belonging to the firmware (b_board.sbn, upload.sbn,        */
/* pbfw.sbn) have to be piped through the driver as one file. The three    */
/* files are kept inside a tar archive. The sequence of the files and the  */
/* filenames in the archive are important.                                 */
/* When the firmware-tar is piped into the driver, the driver is called    */
/* with a piece (usually 4096bytes) of the data. In the first step it      */
/* tries to extract the first tar header and stores the header inside a    */
/* a previous allocated memory buffer (Board->FwDwnld.TarHeader).          */
/* If the header has been read the driver switched to the mode "fw mode".  */
/* Now the Board->FwDwnld.Firmware buffer is filled until a complete       */
/* SBIN record has been stored. The function to deal with the SBIN record  */
/* is called. If the "real" SBIN record is shorter (for example a label-   */
/* record), the remaining bytes are copied to the beginning of the buffer, */
/* and the pointer to the buffer end (Board->FwDwnld.FirmwareAct) is       */
/* corrected. This way the next data will be added on the right place to   */
/* the actual firmware record.                                             */
/* If a SBIN_TYPE_EOF is read or/and the FileSize is equal to 0 the file   */
/* is completed. Because tar-headers are located only on block boundaries  */
/* (512 Bytes) the functions eats the bytes until the next boundary (GAP). */
/* Because different actions have to be done on the different files, the   */
/* action is specified by the "function" argument.                         */
/*                                                                         */
/* The following buffer exist:                                             */
/* - UserBuffer (handed over by the application to the kernel; usually     */
/*               4096 bytes)                                               */
/* - TarHeader  (dynamically inside the driver allocated buffer for        */
/*               storing the information of the actual loaded (to load)    */
/*               file; exist for each Board)                               */
/* - Firmware   (dynamically inside the driver allocated buffer for        */
/*               storing the actual complete SBIN-record; exist for each   */
/*               Board)                                                    */
/*                                                                         */
/* Closing the board device switches to state 0 (init).                    */
/*                                                                         */
/* The archive can be generated by the following command:                  */
/*                                                                         */
/* tar cvf fw.5.21.tar b_board.sbn upboard.sbn pbfw.sbn                    */
/*                                                                         */
/* XXX If the archive containing the firmware is OK, the module will work, */
/* but I never checked the other case :-(                                  */
/***************************************************************************/

#ifdef USER
#include <stdio.h>
#include <unistd.h>
#include <malloc.h>
#include <string.h>
#else
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/version.h>
#include <linux/sched.h>
#include "kernel_version.h"
#include <linux/delay.h>


#if (VERSION_CODE(2,2,0) <= LINUX_VERSION_CODE)
  #include <linux/poll.h>
#endif

#if ((VERSION_CODE(2,0,0) <= LINUX_VERSION_CODE) && \
     (LINUX_VERSION_CODE < VERSION_CODE(2,3,0)) )
#include <linux/malloc.h>
#elif (VERSION_CODE(2,4,0) <= LINUX_VERSION_CODE)
#include <linux/slab.h>
#endif

#endif
#include <asm/types.h>
#include <asm/io.h>  // for read_io_byte
#include "pb_debug.h"
#include "basetypes.h"
#include "cntrl.h"
#include "fw.h"
#include "tar.h"
#include "cmi.h"
#include "xx_if.h"


// The GAP_MODE is for the gaps between a file and the next block
// boundary, because tar headers always start at block boundaries.
#define PB_DRIVER_TAR_MODE  0
#define PB_DRIVER_FW_MODE   1
#define PB_DRIVER_GAP_MODE  2

#define PB_TAR_BLOCKSIZE 512

#define B_BOARD_SBN "b_board.sbn"
#define UPBOARD_SBN "upboard.sbn"
#define PBFW_SBN    "pbfw.sbn"
#define PBFWDPS_SBN "pbfwdps.sbn"

#ifdef USER
#include <stdio.h>
#define KERN_ERR
#define KERN_DEBUG
#define kmalloc( a, b ) malloc( a )
#define CpyFromUserToKernel memcpy
#define printk( fmt, args...) fprintf(stderr, fmt, ## args )
#define PR_DEBUG(D_FW_L1, fmt, args...) fprintf(stderr, fmt, ## args )
#else
#define CpyFromUserToKernel copy_from_user
#endif

static int WaitForReaction( struct CntrlExt *Board, int Timeout );

__u8 ReadSBinRecord( __u8 **image, __u32 *imageLen, SBinHeader **header,
  __u8 **data, __u16 *dataCheckSum)
{
  //fprintf(stderr,"ReadSBinRecord( %d )\n", *imageLen );
  if (*imageLen < sizeof(SBinHeader)) 
  { 
    // image smaller than SBIN header
    printk(KERN_ERR "error read SBin header from image (%u < %d)\n",
      *imageLen, (USIGN32)sizeof(SBinHeader));
    return FALSE;
  }

  *header = (SBinHeader *) *image;
  *image += sizeof(SBinHeader);
  *imageLen -= sizeof(SBinHeader);

  if ((*header)->len > MAX_SBIN_DATA_SIZE ) {
    printk(KERN_ERR "SBin record exceeds max size %u\n", (*header)->len);
    return FALSE;
  }

  if (*imageLen < (__u32)((*header)->len + 2)) { // data and checksum
    printk(KERN_ERR "error read SBin data from image (%u < %u))\n",
        *imageLen, (*header)->len + 2);
    return FALSE;
  }

  *data = (__u8 *) *image;
  *image += (*header)->len + 2;
  *imageLen -= (__u32)((*header)->len + 2);

  // calculate checksum
  if( dataCheckSum ) {
    __u16 headerCheckSum;
    __u8 *headerRaw;
    int i;

    // header checksum
    headerRaw = (__u8 *)(*header);
    for (i = 0, headerCheckSum = 0; i < sizeof(SBinHeader);
      headerCheckSum += headerRaw[i], i++)
      ;

    // data checksum
    for (i = 0, *dataCheckSum = 0; i < (*header)->len;
      *dataCheckSum += (*data)[i], i++)
      ;

    // check checksum
    if( (__u16)(headerCheckSum + *dataCheckSum)
      != *((__u16 *) (*data + (*header)->len)) ) {
      printk(KERN_ERR "checksum error in SBin record (%X != %X))\n",
        headerCheckSum + *dataCheckSum,
        *((__u16 *) (*data + (*header)->len)));
      return FALSE;
    }
  }
  return TRUE;
}

// There is no sscanf function in the kernel :-(

static USIGN32 AsciiToOctal( char *Octal )
{
  USIGN32 result=0;

  for( ; *Octal && *Octal==' '; Octal++ ) {
    ;
  }
  for( ; *Octal && *Octal!=' '; Octal++ ) {
    result *= 8;
    result += *Octal-'0';
  }
  return( result );
}

static void SwitchToProgMode( struct CntrlExt *Board )
{
  T_CMI_DESCRIPTOR *cmi;
  USIGN32 Delay = 1 * HZ;
  unsigned volatile char *ReceiptCode =
    (char *)(Board->DprAddr+OFFSET_RECEIPT_CODE);
  unsigned volatile char *CommandCode =
    (char *)(Board->DprAddr+OFFSET_COMMAND_CODE);

  cmi = (T_CMI_DESCRIPTOR *) Board->DprAddr;
  write_io_byte( ReceiptCode, RECEIPT_NONE );
  write_io_byte( CommandCode, COMMAND_NONE );
//   write_io_long( &cmi->h_ready_mask, 0x474f5250L ); // PROG
  cmi->h_ready_mask = 0x474f5250L;
  // wait some time, that the card can clear the flash ...
  schedule_timeout (Delay);
  if( WaitForReaction( Board, TIMEOUT_CARD_CLEAR_FLASH ) ) 
  {
    printk(KERN_ERR "profiboard: no reaction (clearing flash)\n");
  }
}

static int WaitForReaction( struct CntrlExt *Board, int Timeout )
{
  char volatile *Reaction = (char *)(Board->DprAddr + OFFSET_RECEIPT_CODE);

  StartTimeout( &Timeout, Timeout ); // currently unnecessary
  while( !SleepAWhile( &Timeout ) ) {
    if( read_io_byte( Reaction ) == RECEIPT_OK )
      return( 0 );
  }
  printk(KERN_ERR "profiboard: WaitForReaction=%d\n", read_io_byte(Reaction));
  return( 1 );
}

// Function for writing firmware records piece by piece ...

static int BurnFirmware( SBinHeader *Header, __u8 *data, __u16 CheckSum,
  struct CntrlExt *Board, void *Par1, void *Par2 )
{
  char volatile *ReceiptCode = (char *)(Board->DprAddr+OFFSET_RECEIPT_CODE);
  char volatile *CommandCode = (char *)(Board->DprAddr+OFFSET_COMMAND_CODE);
  Copymem volatile *CopyMem =
    (Copymem *) (Board->DprAddr+OFFSET_COMMAND_DATA);
  Jumpprog volatile *JumpProg =
    (Jumpprog *) (Board->DprAddr+OFFSET_COMMAND_DATA);
  ReceiptCopyMem volatile *ReceiptMem =
    (ReceiptCopyMem *) (Board->DprAddr+OFFSET_RECEIPT_DATA);

  switch( Header->type ) 
  {
  case SBIN_TYPE_DATA:
    if( Board->FwDwnld.PrevCheckSum &&
        (Board->FwDwnld.PrevCheckSum != read_io_word (&ReceiptMem->CheckSum)) ) 
    {
      printk(KERN_ERR "profiboard: Checksumfailure in SBIN record %d!\n",
         Board->FwDwnld.count );
      printk("Prev: %x - Act: %x\n", Board->FwDwnld.PrevCheckSum,
          read_io_word (&ReceiptMem->CheckSum) );
      
      //DumpMem(Board->DprAddr,0x800,"DPRAM");
    }
    //printk("%d (%x)\n", Board->FwDwnld.count, ReceiptMem->CheckSum );
    Board->FwDwnld.count++;
    Board->FwDwnld.PrevCheckSum = CheckSum;
    copy_to_io ( (char*)Board->DprAddr+OFFSET_DNL_DATA, data, Header->len );
//     write_io_long (&CopyMem->sourceAddrAbs, ADDR_DPR_FROM_BOARD_VIEW + OFFSET_DNL_DATA);
    CopyMem->sourceAddrAbs = ADDR_DPR_FROM_BOARD_VIEW + OFFSET_DNL_DATA;
//     write_io_long (&CopyMem->destinationAddrAbs, Header->addr);
    CopyMem->destinationAddrAbs = Header->addr;
    write_io_word (&CopyMem->len, Header->len);
    write_io_byte (&CopyMem->calcCS, TRUE);

    write_io_byte( ReceiptCode, RECEIPT_NONE );
    write_io_byte( CommandCode, COMMAND_COPY_MEM );
    if( WaitForReaction( Board, TIMEOUT_CARD ) ) 
    {
      printk(KERN_ERR "profiboard: no reaction from board\n");
      return( 1 );
    }
    break;
  case SBIN_TYPE_START:
    PR_DEBUG(D_FW_L1, "sbin_type_start: %x\n", Header->addr );
    if( Header->addr & 0x01L ) 
    {
      printk(KERN_ERR "profiboard: bad startaddr %x\n", Header->addr );
    } 
    else 
    {
      Board->FwDwnld.StartAddr = Header->addr;
    }
    break;
  case SBIN_TYPE_EOF:
    // Now we are ready!
    PR_DEBUG(D_FW_L1, KERN_DEBUG "Now we are ready ...\n");
    if( Board->FwDwnld.StartAddr ) 
    {
     // printk("Starting the firmware ...\n");
     //       write_io_long (&JumpProg->entryPointsAbs, Board->FwDwnld.StartAddr);
     JumpProg->entryPointsAbs = Board->FwDwnld.StartAddr;
     write_io_byte (&JumpProg->autoRestart, TRUE);
     write_io_byte( ReceiptCode, RECEIPT_NONE );
     write_io_byte( CommandCode, COMMAND_JUMP_PROG );
     if( WaitForReaction( Board, TIMEOUT_CARD ) ) 
     {
       printk(KERN_ERR "profiboard: no reaction after starting\n");
       return( 1 );
     }
    }
    break;
  default:
    break;
  }
  return( 0 );
}

// Writes b_board (boot program) and upboard (loader program)

static int WriteImageToDpr( SBinHeader *Header, __u8 *data, __u16 CheckSum,
  struct CntrlExt *Board, void *Par1, void *Par2 )
{
  PB_ADDRESS Offset = (PB_ADDRESS) Par1;
  PB_ADDRESS AddrOffset = (PB_ADDRESS) Par2;

  if( Header->type == SBIN_TYPE_DATA ) {
    PR_DEBUG(D_FW_L1,KERN_DEBUG "copy %d to %lx\n", Header->len,
        Board->DprAddr + (Header->addr-AddrOffset)+Offset );
    copy_to_io( (char *)Board->DprAddr + (Header->addr-AddrOffset)
      + Offset, data, Header->len );
    //DpramOut( Board, (Header->addr-AddrOffset)+Offset );
  }
  return( 0 );
}


static USIGN32 PbWorkOnSbinRecord( struct CntrlExt *Board,__u32 length,
  int *Type,
  int (Function)(SBinHeader*,__u8 *,__u16 cs,struct CntrlExt *,void *,void*),
  void *Par1, void *Par2 )
{
  __u8 *data, *BufferPtr=Board->FwDwnld.Firmware;
  __u16 DoChcksm;
  SBinHeader *header;
  USIGN32 Len;

  //printk(KERN_DEBUG "PbWorkOnSbinRecord( length %d )\n", length );
  if(ReadSBinRecord(&BufferPtr,&length,&header,&data,&DoChcksm)) 
  {
    Function( header, data, DoChcksm, Board, Par1, Par2 );
    switch( header->type ) 
    {
      case SBIN_TYPE_STRING:
        PR_DEBUG(D_FW_L1,"String \"%s\"\n", data);
       // printk("String \"%s\"\n", data);
        break;
      case SBIN_TYPE_DATA:
        PR_DEBUG(D_FW_L1,"Data %d\n", header->len);
       // printk("Data %d\n", header->len);
        break;
      case SBIN_TYPE_EOF:
        PR_DEBUG(D_FW_L1,"EOF %d\n", header->len);
       // printk("EOF %d\n", header->len);
        break;
      case SBIN_TYPE_START:
        PR_DEBUG(D_FW_L1,"START %x\n", header->addr);
       // printk("START %x\n", header->addr);
        break;
      default:
        PR_DEBUG(D_FW_L1,"Unknown record type %d\n", header->type );
        printk("Unknown record type %d\n", header->type );
        break;
    }
    *Type = header->type;
    // We have to copy the remaining data in the Firmware buffer
    // to the beginning of the buffer. It is the next SBIN
    // record.

    Len = (USIGN32) header->len + sizeof(SBinHeader)+2;
    Board->FwDwnld.FirmwareAct = data+length - sizeof(SBinHeader);
    copy_to_io ( Board->FwDwnld.Firmware, data+header->len+2, length );
    PR_DEBUG(D_FW_L1,"firmware now copied\n");
  } 
  else 
  {
    Len = 0;
  }

  return( Len );
}





/******************************************************/
/* PbWriteBoard is really not easy to understand.     */
/* But the comment on top of the file might help you. */
/* The function has to be reentrant! Don't use global */
/* variables!                                         */
/******************************************************/

int PbWriteBoard( struct CntrlExt *Board, char *buffer, int count )
{
  union record *Block;
  int ToCopy, Type=SBIN_TYPE_START;
  char *ptr;
  int MaxSbinRecordSize=MAX_SBIN_RECORD_SIZE;
  char volatile *ReceiptCode =(char *)(Board->DprAddr+OFFSET_RECEIPT_CODE);
  char volatile *CommandCode =(char *)(Board->DprAddr+OFFSET_COMMAND_CODE);
  __u32 volatile *BootFlag;
  char *pDigit;
  __u32 Version = 0;
  unsigned long dummy;

  BootFlag = (__u32 *)(Board->DprAddr + OFFSET_BOOT_BLOCK_FLAG);
  
  pDigit = (char*)BootFlag;
  
  Version = (pDigit[5]-'0')*100+(pDigit[6]-'0')*10+pDigit[7]-'0';
  
  
  if(Version < 250)
  {
    printk("Bootblock Version less than V250: %u\n",Version);

#ifdef DPB_PCI_SUPPORT    
      PbStartBoard( Board, 0 );
      msleep(100);
      PbStartBoard( Board, C165_RESET );
      msleep(100);
#endif // DPB_PCI_SUPPORT    
    
    
  }
  else
  {
    printk("Bootblock Version greater or equal V250: %u\n",Version);
  }
  

 // printk("PbWriteBoard( %d )\n", count );
  if( Board->FwDwnld.TarHeader == NULL ) 
  {
    Board->FwDwnld.TarHeaderAct = Board->FwDwnld.TarHeader
      = kmalloc( sizeof( union record ), GFP_KERNEL );
    Board->FwDwnld.State = PB_DRIVER_TAR_MODE;
    if( CheckForBootblock( Board ) == 1 ) 
    {
#ifdef DPB_PCI_SUPPORT
        if(Version < 250)
        {
         
          PbStartBoard( Board, C165_RESET );
          msleep(100);
        }
        
#endif
    } 
    else 
    {
      printk( KERN_ERR "profiboard: no bootblock found!\n");
      return( -1 );
    }
  }
  if( Board->FwDwnld.Firmware == NULL ) 
  {
    Board->FwDwnld.FirmwareAct = Board->FwDwnld.Firmware
      = kmalloc( MaxSbinRecordSize, GFP_KERNEL );
  }
  Block = (union record *) Board->FwDwnld.TarHeader;
  // what can we do with the current bytes?
  for( ptr=buffer; count>0 ; ) 
  {
    /*******************/
    /* Read TAR Header */
    /*******************/
    if( Board->FwDwnld.State == PB_DRIVER_TAR_MODE ) 
    {
      PR_DEBUG(D_FW_L1, "Tar-Mode\n" );
      ToCopy = Board->FwDwnld.TarHeader - Board->FwDwnld.TarHeaderAct
        + sizeof(union record);
      if( ToCopy > count )
        ToCopy = count;
      dummy = CpyFromUserToKernel( Board->FwDwnld.TarHeaderAct, ptr, ToCopy );
      count -= ToCopy;
      ptr += ToCopy;
      Board->FwDwnld.TarHeaderAct += ToCopy;
      if( (Board->FwDwnld.TarHeader+sizeof(union record))
        == Board->FwDwnld.TarHeaderAct ) 
      {
        // Tar header is complete!
        Board->FwDwnld.TarHeaderAct = Board->FwDwnld.TarHeader;
        if( *Block->header.arch_name == '\0' ) 
        {
          //printk( KERN_DEBUG "Archive end (%d)?\n", count);
          continue;
        }
        Board->FwDwnld.FileSize=AsciiToOctal((void *)
            &Block->header.size);
        Board->FwDwnld.GapSize = (PB_TAR_BLOCKSIZE
          - ( Board->FwDwnld.FileSize % PB_TAR_BLOCKSIZE ))
          % PB_TAR_BLOCKSIZE;
        printk( KERN_DEBUG "got \"%s\" length %d gap %d\n",
            Block->header.arch_name, Board->FwDwnld.FileSize,
            Board->FwDwnld.GapSize);
        Board->FwDwnld.State = PB_DRIVER_FW_MODE;
        // Because Windows often uses a capital letter as the first
        // letter in a filename, we check the string after this (+1)
        // So we are tolerant against this.
        if( strstr( Block->header.arch_name, B_BOARD_SBN+1 ) ) 
        {
          Board->FwDwnld.FilesLoaded &= PB_B_BOARD_LOADED;
	  
	  //           write_io_long( BootFlag, FLAG_NO_BOOT_BLOCK );
	  //           write_io_long( (long*) Board->DprAddr, 0 );
	  
	  *BootFlag = FLAG_NO_BOOT_BLOCK;
	  *(USIGN32*)(Board->DprAddr) = 0;
          PbInitBoard( Board, 0 );
        } 
        else if( strstr( Block->header.arch_name, UPBOARD_SBN+1 ) ) 
        {
          Board->FwDwnld.FilesLoaded &= PB_UPBOARD_LOADED;
        } 
        else if( (strstr( Block->header.arch_name, PBFW_SBN+1    )) ||
                  (strstr( Block->header.arch_name, PBFWDPS_SBN+1 )) ) 
        {
          Board->FwDwnld.PrevCheckSum = 0;
          Board->FwDwnld.count = 0;
          Board->FwDwnld.FilesLoaded &= PB_FW_LOADED;
          // Now switch to PROG Mode
          write_io_byte( ReceiptCode, RECEIPT_NONE );
          write_io_byte( CommandCode, COMMAND_NONE );
	  //           write_io_long( BootFlag, FLAG_NO_BOOT_BLOCK );
	  *BootFlag = FLAG_NO_BOOT_BLOCK;
          PbStartCntrl( Board );
          WaitForBootFlag( Board );
          //WaitForReaction( Board, TIMEOUT_CARD_CLEAR_FLASH );
          printk(KERN_DEBUG "switch to PROG mode ...\n");
          SwitchToProgMode( Board );
          printk(KERN_DEBUG "done\n");
        } 
        else 
        {
          printk(KERN_DEBUG "unknown file in archive!\n");
        }
        // Check, whether both FW-startup-files have been written
      }
    }
    /**********************/
    /* Read Firmware File */
    /**********************/
    if( Board->FwDwnld.State == PB_DRIVER_FW_MODE ) 
    {
      PR_DEBUG(D_FW_L1,"FW-Mode\n");
      // is the File smaller as the SBIN record buffer?
      MaxSbinRecordSize = (Board->FwDwnld.FileSize<MAX_SBIN_RECORD_SIZE) ?
        Board->FwDwnld.FileSize : MAX_SBIN_RECORD_SIZE;
      ToCopy = Board->FwDwnld.Firmware + MaxSbinRecordSize
        - Board->FwDwnld.FirmwareAct;
      if( ToCopy > count )
        ToCopy = count;
      dummy = CpyFromUserToKernel( Board->FwDwnld.FirmwareAct, ptr, ToCopy );
      Board->FwDwnld.FirmwareAct += ToCopy;
      ptr += ToCopy;
      count -= ToCopy;
      if( (Board->FwDwnld.Firmware+MaxSbinRecordSize)
        ==Board->FwDwnld.FirmwareAct ) 
      {
        // FW-Download of the current SBIN-RECORD completed
        Board->FwDwnld.FirmwareAct = Board->FwDwnld.Firmware;
        if( strstr( Block->header.arch_name, B_BOARD_SBN ) ) 
        {
          Board->FwDwnld.FileSize -= PbWorkOnSbinRecord( Board,
            MaxSbinRecordSize, &Type, WriteImageToDpr,
            (void *)OFFSET_BOOT_IMAGE,
            (void *)ADDR_OFFSET_BOOT_IMAGE);
        }
        if( strstr( Block->header.arch_name, UPBOARD_SBN ) ) 
        {
          Board->FwDwnld.FileSize -= PbWorkOnSbinRecord( Board,
            MaxSbinRecordSize, &Type, WriteImageToDpr,
            (void *)OFFSET_LOAD_IMAGE,
            (void *)ADDR_OFFSET_LOAD_IMAGE);
        }
        if( (strstr( Block->header.arch_name, PBFW_SBN+1    )) ||
            (strstr( Block->header.arch_name, PBFWDPS_SBN+1 )) ) 
        {
          Board->FwDwnld.FileSize -= PbWorkOnSbinRecord( Board,
            MaxSbinRecordSize, &Type, BurnFirmware,
            (void *)NULL, (void *)NULL );
        }
        //fprintf(stderr,"FileSize: %ld\n", Board->FwDwnld.FileSize );
        if( Board->FwDwnld.FileSize==0 || Type==SBIN_TYPE_EOF ) 
        {
          // Download of File completed
          Board->FwDwnld.State = PB_DRIVER_GAP_MODE;
          if( (strstr( Block->header.arch_name, PBFW_SBN+1    )) ||
              (strstr( Block->header.arch_name, PBFWDPS_SBN+1 )) ) 
          {
            char LocBuf[256];
            unsigned short BufSize=sizeof(LocBuf);

            if ( PbStartBoard( Board, C165_RESET ) == 0 ) 
            {
              if (( cmi_get_data(Board,ID_FW_VERS_IMAGE,0,&BufSize,LocBuf,
                      PB_KERNEL_SPACE ) == E_OK))
              {
                if(Board->FWVersion)
                  kfree(Board->FWVersion);
                Board->FWVersion = kmalloc( strlen(LocBuf), GFP_ATOMIC);
                if( Board->FWVersion ) 
                {
                  memcpy( Board->FWVersion, LocBuf, strlen(LocBuf) );
                } 
                else 
                {
                  if(Board->FWVersion)
                      Board->FWVersion[0]=0; 
                }
              }
            }
          }
        }
      }
    }
    /*****************************/
    /* Read until block boundary */
    /*****************************/
    if( Board->FwDwnld.State == PB_DRIVER_GAP_MODE ) 
    {
      ToCopy = Board->FwDwnld.GapSize;
      if( ToCopy > count ) 
      {
        ToCopy = count;
      }
      //fprintf(stderr,"gaping (%d) %d\n",ToCopy,Board->FwDwnld.GapSize);
      // nothing to copy here, just ignore the bytes
      ptr += ToCopy;
      count -= ToCopy;
      if( (Board->FwDwnld.GapSize -= ToCopy) == 0 ) 
      {
        // gap completed, read next file from archive
        Board->FwDwnld.State = PB_DRIVER_TAR_MODE;
      }
    }
  }
  // let's wait for the next bytes ...
  return( 1 );
}

# if 0 //#if USER
int main( int argc, char **argv )
{
  __u8 buffer[MAX_SBNRECORD_SIZE];
  int count;
  struct CntrlExt Board;

  memset( &Board, 0, sizeof(Board) );
  printf("MAX_SBIN_RECORD_SIZE=%d\n", MAX_SBIN_RECORD_SIZE );
  while( (count=read( STDIN_FILENO, buffer, sizeof(buffer))) ) 
  {
    //fprintf(stderr,"read %d bytes from stdin\n", count);
    PbWriteBoard( &Board, buffer, count );
  }
  return( 0 );
}
#endif
