/*****************************************************************************/
/*                                                                           */
/*                              SOFTING AG                                   */
/*                        Richard-Reitzner-Allee 6                           */
/*                              D-85540 Haar                                 */
/*                      Phone: (++49)-(0)89-45656-0                          */
/*                      Fax:   (++49)-(0)89-45656-399                        */
/*                                                                           */
/*                  Copyright (C) SOFTING AG 1995-2007                       */
/*                            All Rights Reserved                            */
/*                                                                           */
/*****************************************************************************/
/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*  Filename    : DOWNLOAD.C                                                 */
/*  Author      : SOFTING-AG                                                 */
/*                                                                           */
/*  Description : Some functions for firmmmware download                     */
/*                                                                           */
/*  Note:         The bootloader of the PB-PROx-PCI interface contains the   */
/*                functions to flash the on-chip flash memory                */
/*                                                                           */
/*  CHANGE_NOTES                                                             */
/*                                                                           */
/*  date      name      change                                               */
/*  -----------------------------------------------------------------------  */
/*****************************************************************************/

#include "basetypes.h"

#include <linux/module.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,33)
#include <linux/autoconf.h>
#endif
#include <asm/io.h>
#include <asm/segment.h>

#include "kernel_version.h"

#include <linux/poll.h>


#include <linux/slab.h>
#include <linux/proc_fs.h>


#include "pb_debug.h"
#include "cntrl.h"
#include "pb_hw_pci.h"

#define FW_BUFFER_SIZE 500000





typedef struct
{
  u32  fwBufPtr;
  BOOL isUsed;
  u8   buffer[FW_BUFFER_SIZE];
} FW_BUFFER;

extern void ReleaseSpinlock(struct CntrlExt *Board, unsigned long Flags, unsigned long line);
extern void GetSpinlock(struct CntrlExt *Board, unsigned long *Flags, unsigned long line);
extern int PbInitBoard( struct CntrlExt *Board, char Reset );
extern void DumpMem(void *pMem, unsigned long size, char *pText);



static FW_BUFFER load_buffer;

const unsigned short Crc16Table [] =
{
  0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
  0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
  0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
  0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
  0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
  0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
  0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
  0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
  0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
  0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
  0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
  0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
  0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
  0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
  0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
  0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
  0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
  0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
  0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
  0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
  0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
  0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
  0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
  0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
  0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
  0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
  0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
  0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
  0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
  0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
  0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
  0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0
};

/*---------------------------------------------------------------------------*/


typedef unsigned short   T_BOOTIF_SIGNAL;

/*--- CONTROLLER -> HOST signals -----------------------------------------------*/
#define BOOTIF_SIGNAL_ACKNOWLEDGE                       ((T_BOOTIF_SIGNAL) 0x0000)

/*--- HOST -> CONTROLLER signals -----------------------------------------------*/

#define BOOTIF_SIGNAL_SERVICE_GET_DEVICE_INFO           ((T_BOOTIF_SIGNAL) 0x0001)
#define BOOTIF_SIGNAL_SERVICE_PROGRAM_FLASH             ((T_BOOTIF_SIGNAL) 0x0004)
#define BOOTIF_SIGNAL_SERVICE_CALCULATE_FLASH_CHECKSUM  ((T_BOOTIF_SIGNAL) 0x0009)

#define BOOTIF_SIGNAL_RESET_FIRMWARE                    ((T_BOOTIF_SIGNAL) 0x4000)


/*---------------------------------------------------------------------------*/

#define BOOTBLOCK_VERSION_SIZE                72
#define PROFIBUS_FIRMWARE_VERSION_SIZE        128

#pragma pack(push,2)

typedef struct _T_BOOT_IF_DESCR   /* 232 bytes */             /* must be located at offset 0x00 */
{
  volatile u32      HostBootSelectMask;             /* set to START_BOOTBLOCK_FIRMWARE to use bottloader functionality */

  unsigned char               Reserved1 [22];                 /* !!! DO NOT USE THIS AREA !!! */

  u32               LiveSign;                       /* Counter to indicate taht the controller is still alive */

  volatile T_BOOTIF_SIGNAL    SignalFromHost;                 /* set by the host and reset by the controller */
  volatile T_BOOTIF_SIGNAL    SignalFromController;           /* set by the controller and reset by the host */

  unsigned short              IfId;                           /* current interface id/version      -> set by the controller */
  unsigned short              BootIfId;                       /* current T_BOOT_IF_DESCR id/version-> set by the controller */
  unsigned short              ServiceIfId;                    /* current T_BOOT_IF_DESCR id/version-> set by the controller */

  unsigned char               Reserved2 [64];

  u32               SerialNumber;                   /* SerialDeviceNumber -> set by the controller */
  unsigned short              DeviceId;                       /* PB-PRO device id -> set by the controller */
  unsigned short              HwRevision;                     /* set by the controller */

  volatile u32      BootLoaderMask;                 /* set by the controller Note: must be located at offset 0x70 */
  u32               BootLoaderVersion;              /* set by the controller Note: must be located at offset 0x74 */

  char                        BootBlockVersionString [BOOTBLOCK_VERSION_SIZE];  /* set by the controller */
  char                        FirmwareVersionString [PROFIBUS_FIRMWARE_VERSION_SIZE]; /* set by the controller Note: must be located at offset 0xC0 */

} T_BOOT_IF_DESCR;


/*---------------------------------------------------------------------------*/

typedef struct _T_BOOT_DEVICE_INFO
{
  u32               SerialNumber;
  unsigned short              DeviceId;
  unsigned short              HwRevision;

} T_BOOT_DEVICE_INFO;


/*---------------------------------------------------------------------------*/


#define PB_PRO_FLASH_BASE_ADDRESS 0xC08000   /* PB-PRO flash base address to store located firmware */

#define MAX_BINARY_LENGTH         2048u
#define TYPE_PROGRAM_DATA         0
#define TYPE_PROGRAM_END          1

typedef struct _T_PROGRAM_FLASH
{
  unsigned short              Type;              /* TYPE_PROGRAM_DATA, TYPE_PROGRAM_END */

  u32               Offset;            /* flash address to store data block */
  u32               Length;            /* size of data block */

  unsigned char               BinaryData [MAX_BINARY_LENGTH];

} T_PROGRAM_FLASH;

/*---------------------------------------------------------------------------*/

typedef struct _T_FLASH_CHECKSUM
{
  unsigned short    BootblockFlashChecksum;
  unsigned short    ProtocolFwFlashChecksum;

} T_FLASH_CHECKSUM;


typedef struct _T_SERVICE_REQ_IF   /* 2060 bytes */
{
  signed short                ReturnCode;

  union
  {
    T_BOOT_DEVICE_INFO        DeviceInfo;
    T_PROGRAM_FLASH           ProgramFlash;
    T_FLASH_CHECKSUM          FlashChecksum;
  }                           UseAs;
} T_SERVICE_REQ_IF;



typedef struct _T_BOOT_HOST_SHARED_MEMORY
{
  T_BOOT_IF_DESCR             BootIfDescr;
  T_SERVICE_REQ_IF            ServiceReq;

} T_BOOT_HOST_SHARED_MEMORY;


#pragma pack(pop)

#define ENABLE_IRQ_CHANNEL_0      0x00000001
#define ENABLE_IRQ_CHANNEL_1      0x00000008
#define DISABLE_IRQ_CHANNEL_0     (~ENABLE_IRQ_CHANNEL_0)
#define DISABLE_IRQ_CHANNEL_1     (~ENABLE_IRQ_CHANNEL_1)

#define RUN_CHANNEL_0             0x00000020
#define RUN_CHANNEL_1             0x00000800
#define RESET_CHANNEL_0           (~RUN_CHANNEL_0)
#define RESET_CHANNEL_1           (~RUN_CHANNEL_1)

#define SET_IRQ_CHANNEL_0         0x00000004
#define SET_IRQ_CHANNEL_1         0x00000100
#define CLEAR_IRQ_CHANNEL_0       (~SET_IRQ_CHANNEL_0)
#define CLEAR_IRQ_CHANNEL_1       (~SET_IRQ_CHANNEL_1)

/*--- pci board identifiers --------------------------------------------------*/
#define PLX_VENDOR_ID         0x10B5
#define PLX_9050_DEVICE_ID    0x9050
#define PLX_9030_DEVICE_ID    0x9030
#define PLX_86E1_DEVICE_ID    0x86E1
#define SOFTING_VENDOR_ID     0x14A0
#define SOFTING_DEVICE_ID     0x0028
#define SOFTING_PRO_DEVICE    0x002A



#define BOOTIF_OK                                   0
#define BOOTIF_ERROR_INVALID_SERIAL_NUMBER         -1
#define BOOTIF_ERROR_EEPROM_WRITE_FAILED           -2
#define BOOTIF_ERROR_EEPROM_READ_FAILED            -3
#define BOOTIF_ERROR_INVALID_LENGTH                -4
#define BOOTIF_ERROR_INVALID_OFFSET                -5
#define BOOTIF_ERROR_FLASH_ERASE_ERROR             -6
#define BOOTIF_ERROR_FLASH_WRITE_ERROR             -7
#define BOOTIF_ERROR_FLASH_READ_ERROR              -8
#define BOOTIF_ERROR_CHECK_SUM_ERROR               -9
#define BOOTIF_ERROR_INVALID_PROGRAM_TYPE          -10
#define BOOTIF_ERROR_INVALID_INTERRUPT_NUMBER      -11
#define BOOTIF_ERROR_SERVICE_NOT_SUPPORTED         -12


#define TYP_BOOTIF_SIGNAL_ACK_TIME        1                // 1ms
#define MAX_BOOTIF_SIGNAL_ACK_TIME        1000             // 1sec

#define TYP_START_FIRMWARE_ACK_TIME       1                // 1ms
#define MAX_START_FIRMWARE_ACK_TIME       2000             // 2sec


/*--- boot loader mask ------------------------------------------------------*/
#define BOOT_BLOCK_EXIST             0x544F4F42uL          /* "BOOT" Bootblock-Exist-Flag */
#define START_BOOTBLOCK_FIRMWARE     0x544F4F42uL          /* "BOOT" command to start Bootblock firmware */

#define _ASSERT(Condition)  do { assert (Condition); } while (0)

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/


/*********************************************************************************/
/*                                                                               */
/* Functions to download firmware for PBpro-Pci and PBpro-PC104+                 */
/*                                                                               */
/*********************************************************************************/

static BOOL pb_pro_start_device(struct CntrlExt     *pCtrlExt)
  /*-------------------------------------------------------------------------*/
{
 volatile T_BOOT_HOST_SHARED_MEMORY*   pBootHostSharedMemory;
  //u32 ResetCommand;
  u32 PollingCounter = 0, MaxTimeout = 100;
//   unsigned long flags;
  
  if(pCtrlExt->DeviceType == SOFTING_PB_PCIE_DEVICE)
  {
    disable_pcie_irq(pCtrlExt);
  }


  /*--- get host shared memory base address ---------------------------------*/
  pBootHostSharedMemory = (T_BOOT_HOST_SHARED_MEMORY*) pCtrlExt->DprAddr;

  /*--- reset and start PB-PRO ----------------------------------------------*/
  pBootHostSharedMemory->BootIfDescr.BootLoaderMask = 0uL;

  PbInitBoard(pCtrlExt,0); //halt
  PbInitBoard(pCtrlExt,1); // run again
  
  if(pCtrlExt->DeviceType == SOFTING_PB_PCIE_DEVICE)
  {
    disable_pcie_irq(pCtrlExt);
  }
  
  do
  {

    PollingCounter++;

    if (PollingCounter > MaxTimeout)
    {
      
      PbInitBoard(pCtrlExt,0); //halt
      printk("pb_pro_start_device: no response from controller!\n");
      return (FALSE);
    }
    udelay(10000);

  } while (pBootHostSharedMemory->BootIfDescr.BootLoaderMask != BOOT_BLOCK_EXIST);


  /*--- sart bootblock firmware -------------------------------------------------*/
  pBootHostSharedMemory->BootIfDescr.HostBootSelectMask = START_BOOTBLOCK_FIRMWARE;

  return (TRUE);
}


/*===========================================================================*/


static BOOL pb_pro_restart_device
  (
     struct CntrlExt   *pCtrlExt
  )
  /*-------------------------------------------------------------------------*/
{
  return(pb_pro_start_device(pCtrlExt));
}


/*===========================================================================*/


static BOOL pb_pro_send_signal_to_controller
  (
      struct CntrlExt   *pCtrlExt,
      T_BOOTIF_SIGNAL  BootIfSignal
  )
  /*-------------------------------------------------------------------------*/
{
  volatile T_BOOT_HOST_SHARED_MEMORY*   pBootHostSharedMemory;
  volatile u32                PollingCounter = 0;
 
  pBootHostSharedMemory = (T_BOOT_HOST_SHARED_MEMORY*) pCtrlExt->DprAddr;

  /*--- send signal to controller ---------------------------------------*/
 // _ASSERT (pBootHostSharedMemory->BootIfDescr.SignalFromHost == BOOTIF_SIGNAL_ACKNOWLEDGE);
  pBootHostSharedMemory->BootIfDescr.SignalFromHost = BootIfSignal;

  /* --- wait for Signal Acknowledge -------------------------------------*/
  do
  {
    mdelay(10);
    PollingCounter++;

    if (PollingCounter > 100)
    {
      printk("pb_pro_send_signal_to_controller: controller does not respond\n");
       return (FALSE);
    }

  } while (pBootHostSharedMemory->BootIfDescr.SignalFromHost != BOOTIF_SIGNAL_ACKNOWLEDGE );

  return (TRUE);
}

/*===========================================================================*/


static u16 pb_pro_generate_crc16
  (
     u8        *pBuffer,
     u32         Size
  )
{
  u32  i;
  u16 Crc16 = 0;
//  u16 Value;

  for (i=0; i<Size; i++)
  {
    Crc16 = Crc16Table[((Crc16 >> 8) ^ pBuffer[i]) & 0xFF] ^ ((Crc16 << 8) & 0xFFFF);
  }

  return (Crc16);
}


/*===========================================================================*/

static BOOL pb_pro_download_firmware
  (
     struct CntrlExt *pCtrlExt,
     u8 *pBuffer,
     u32 size /* size of the firmware */
  )
  /*-------------------------------------------------------------------------*/
{
  
  T_BOOT_HOST_SHARED_MEMORY* pBootHostSharedMemory;
  u16                     CheckSum;
  long                     FwDataLength;
  u32                      FlashAddress;
  u32                      RecordBufferOffset;
  u32                      BinaryLength;
  u8                      *pRecordBuffer = load_buffer.buffer;

   

  //--- read and download data records until EOF (End Of File) detected ------------------
  FwDataLength       = load_buffer.fwBufPtr;
  FlashAddress       = PB_PRO_FLASH_BASE_ADDRESS;  // PB-PRO flash base address to store located firmware 
                                                   // do not change it 
  CheckSum           = pb_pro_generate_crc16(pRecordBuffer,0x38000 ); //FileInformation.EndOfFile.LowPart
  RecordBufferOffset = 0;
  BinaryLength       = 0;

  pBootHostSharedMemory = (T_BOOT_HOST_SHARED_MEMORY *) pCtrlExt->DprAddr;
  
  

 // _ASSERT(pBootHostSharedMemory);
  do
  {
    // max chunk size == MAX_BINARY_LENGTH --------------------------------//
    if (FwDataLength >= MAX_BINARY_LENGTH) 
    {
      BinaryLength = MAX_BINARY_LENGTH;
    }
    else                     
    {              
      BinaryLength = FwDataLength;
    }

    //--- send service request to controller ------------------------------
    pBootHostSharedMemory->ServiceReq.UseAs.ProgramFlash.Type   = TYPE_PROGRAM_DATA;
    pBootHostSharedMemory->ServiceReq.UseAs.ProgramFlash.Length = BinaryLength;
    pBootHostSharedMemory->ServiceReq.UseAs.ProgramFlash.Offset = FlashAddress;

    printk("\n Offset =%x Length =%x", FlashAddress,BinaryLength);

    memcpy(pBootHostSharedMemory->ServiceReq.UseAs.ProgramFlash.BinaryData, &pRecordBuffer[RecordBufferOffset], BinaryLength);

    //--- send signal to controller ---------------------------------------
    //_ASSERT (pBootHostSharedMemory->BootIfDescr.SignalFromHost == BOOTIF_SIGNAL_ACKNOWLEDGE);
    pBootHostSharedMemory->BootIfDescr.SignalFromHost = BOOTIF_SIGNAL_SERVICE_PROGRAM_FLASH;

    if (!pb_pro_send_signal_to_controller (pCtrlExt, BOOTIF_SIGNAL_SERVICE_PROGRAM_FLASH))
    {
      printk("\nPROFIbrd/pb_pro_download_firmware() - leave - Status=ERR_NO_CONTROLLER_SIGNAL_ACK\n");
      //ExFreePool(pRecordBuffer);
      // ZwClose(hFile);
      return (FALSE);
    }

    //--- check service return code --------------------------------------
    switch (pBootHostSharedMemory->ServiceReq.ReturnCode)
    {
      case BOOTIF_ERROR_INVALID_LENGTH:
      {
        printk("\nPROFIbrd/pb_pro_download_firmware() - leave - Status=ERR_INVALID_LENGTH\n");
       // ExFreePool(pRecordBuffer);
        // ZwClose(hFile);
        return (FALSE);
      }
      case BOOTIF_ERROR_INVALID_OFFSET:
      {
        printk("\nPROFIbrd/pb_pro_download_firmware() - leave - Status=ERR_INVALID_OFFSET\n");
       // ExFreePool(pRecordBuffer);
        // ZwClose(hFile);
        return (FALSE);
      }
      case BOOTIF_ERROR_FLASH_ERASE_ERROR:
      {
        printk("\nPROFIbrd/pb_pro_download_firmware() - leave - Status=ERR_FLASH_ERASE_ERROR\n");
      //  ExFreePool(pRecordBuffer);
        // ZwClose(hFile);
        return (FALSE);
      }
      case BOOTIF_ERROR_FLASH_WRITE_ERROR:
      {
        printk("\nPROFIbrd/pb_pro_download_firmware() - leave - Status=ERR_FLASH_WRITE_ERROR\n");
       // ExFreePool(pRecordBuffer);
        // ZwClose(hFile);
        return (FALSE);
      }
      case BOOTIF_ERROR_FLASH_READ_ERROR:
      {
        printk("\nPROFIbrd/pb_pro_download_firmware() - leave - Status=ERR_FLASH_READ_ERROR\n");
      //  ExFreePool(pRecordBuffer);
        // ZwClose(hFile);
        return (FALSE);
      }
      case BOOTIF_ERROR_CHECK_SUM_ERROR:
      {
        printk("\nPROFIbrd/pb_pro_download_firmware() - leave - Status=ERR_CHECK_SUM_ERROR\n");
       // ExFreePool(pRecordBuffer);
        // ZwClose(hFile);
        return (FALSE);
      }
      case BOOTIF_ERROR_INVALID_PROGRAM_TYPE:
      {
        printk("\nPROFIbrd/pb_pro_download_firmware() - leave - Status=ERR_INVALID_PROGRAM_TYPE\n");
      //  ExFreePool(pRecordBuffer);
        // ZwClose(hFile);
        return (FALSE);
      }
    }

    FwDataLength       -= BinaryLength;
    FlashAddress       += BinaryLength;                           // next flash base address 
    RecordBufferOffset += BinaryLength;

  } while (FwDataLength > 0);

  printk("\n");

  // ZwClose(hFile);
 // ExFreePool(pRecordBuffer);

  //--- finish flash procedure ------------------------------
  pBootHostSharedMemory->ServiceReq.UseAs.ProgramFlash.Type = TYPE_PROGRAM_END;

  if (!pb_pro_send_signal_to_controller (pCtrlExt, BOOTIF_SIGNAL_SERVICE_PROGRAM_FLASH))
  {
    printk("\nPROFIbrd/pb_pro_download_firmware() - leave - Status=ERR_NO_CONTROLLER_SIGNAL_ACK.\n");
    return (FALSE);
  }

  if (pBootHostSharedMemory->ServiceReq.ReturnCode != BOOTIF_OK)
  {
    printk("\nPROFIbrd/pb_pro_download_firmware() - leave - Status=0x%x\n",pBootHostSharedMemory->ServiceReq.ReturnCode);
    return (FALSE);
  }

  // build checksum 
  pBootHostSharedMemory->ServiceReq.UseAs.FlashChecksum.BootblockFlashChecksum = 0;
  pBootHostSharedMemory->ServiceReq.UseAs.FlashChecksum.ProtocolFwFlashChecksum = CheckSum;

  if (!pb_pro_send_signal_to_controller (pCtrlExt, BOOTIF_SIGNAL_SERVICE_CALCULATE_FLASH_CHECKSUM))
  {
    printk("\nPROFIbrd/pb_pro_download_firmware() - leave - Status=ERR_NO_CONTROLLER_SIGNAL_ACK!\n");
    return (FALSE);
  }

  printk("\n PROFIbrd/pb_pro_download_firmware() - BootblockCheckSum = 0x%x ProtocolFwFlashChecksum = 0x%x 0x%x\n",
             pBootHostSharedMemory->ServiceReq.UseAs.FlashChecksum.BootblockFlashChecksum,
             pBootHostSharedMemory->ServiceReq.UseAs.FlashChecksum.ProtocolFwFlashChecksum,CheckSum
          );

  if (pBootHostSharedMemory->ServiceReq.ReturnCode != BOOTIF_OK)
  {
    printk("\nPROFIbrd/pb_pro_download_firmware() - leave - Status=0x%x\n",pBootHostSharedMemory->ServiceReq.ReturnCode);
    return (FALSE);
  }

  return (TRUE);
}



/*********************************************************************************/
/*                                                                               */
/* Main-Function to download firmware                                            */
/*                                                                               */
/*********************************************************************************/


//-----------------------------------------------------------------------------
// doFirmwareDownload
//
//      STATUS_SUCCESS
//
u32 doFirmwareDownload(struct CntrlExt  *pCtrlExt)
{
  
  pb_pro_start_device(pCtrlExt);
  pb_pro_download_firmware(pCtrlExt, load_buffer.buffer, load_buffer.fwBufPtr);
  pb_pro_restart_device(pCtrlExt);
  
  
 // DumpMem(load_buffer.buffer, load_buffer.fwBufPtr, "FIRMWARE BUFFER");
  printk("doFirmwareDownload done!\n");
  load_buffer.fwBufPtr = 0;
  load_buffer.isUsed = FALSE;

  return 0;
} // doFirmwareDownload

BOOL isBufferFree(void)
{
  return !load_buffer.isUsed;
}

BOOL copyToFwBuffer(u8 *pBuffer, long size)
{
  u32 i,j;
  long fwSize = size;
  
  if(!load_buffer.isUsed)
  {
    memset(load_buffer.buffer, 0, FW_BUFFER_SIZE);
  }
  
  load_buffer.isUsed = TRUE;
  
  // DumpMem(&pBuffer[0], 0x10, "Firmware Start");
  

  
  for(i=load_buffer.fwBufPtr, j=0; ((i<FW_BUFFER_SIZE) && (fwSize > 0)); i++, j++, fwSize--)
  {
    load_buffer.buffer[i] = pBuffer[j];
  }
  
#if 0
  printk("i=%u, size=%u\n",i,size);
  
  DumpMem(&pBuffer[j-0x10], 0x10, "Firmware End");
  
  if(load_buffer.fwBufPtr > 0x10)
  {
    DumpMem(&load_buffer.buffer[load_buffer.fwBufPtr-0x10], 0x20, "Firmware Buffer Start");
  }
  DumpMem(&load_buffer.buffer[i-0x10], 0x20, "Firmware Buffer End");
#endif //0
  
  load_buffer.fwBufPtr = i;
  
  return (i<FW_BUFFER_SIZE);
}




#define PCI_MANUFACTURER_ID       "SOFTING GmbH"
#define PCI_PROFIBUS_DEVICE_ID    "PROFIBUS PCI Interface"

#define PCI_CONFIG_EEPROM_SIZE    (PCI_COMMON_HDR_LENGTH + 192)    // size of configuration EEPROM

#define PCI_VENDOR_ID             0x10b5  // von PCI Controller PLX 9050
#define PCI_SYSTEM_9050_ID        0x9050  // von PCI Controller PLX 9050
#define PCI_SYSTEM_9030_ID        0x9030  // von PCI Controller PLX 9050
#define PCI_SUBVENDOR_ID          0x14A0  // Softing Vendor ID from PCI SIG

#define PRODUCT_PROFIPCI          0x0028  // PROFIboard-PCI
#define PRODUCT_PROFIPRO_PCI      0x002A  // PBpro-PCI  / PBpro-PCIe
#define PRODUCT_PROFIPRO_PC104    0x0023  // PBpro-PC104+

#define PCI_DEVICE_CHANNEL_0      0x00    // channel 0
#define PCI_DEVICE_CHANNEL_1      0x01    // channel 1

#define PCI_CFG_REGISTERS         0x00    // offset of CONFIG RAM address in common config area
#define PCI_DPRAM_CHANNEL_0       0x02    // offset of DPRAM address of channel 0 in common config area
#define PCI_DPRAM_CHANNEL_1       0x03    // offset of DPRAM address of channel 1 in common config area

#define SELECT_BUS_DEVICE         0x00
#define SELECT_BUS_DEVICE_CHANNEL 0x01

#define PCI_PORT_INTCSR           0x004CL // offset Port Interrupt Control/Status Register
#define PCI_PORT_CNTRL            0x0050L // offset Port User I/O, PCI Target Response, EEPROM,


