/*****************************************************************************/
/*                                                                           */
/*                              SOFTING AG                                   */
/*                        Richard-Reitzner-Allee 6                           */
/*                              D-85540 Haar                                 */
/*                      Phone: (++49)-(0)89-45656-0                          */
/*                      Fax:   (++49)-(0)89-45656-399                        */
/*                                                                           */
/*                  Copyright (C) SOFTING AG 1995-2007                       */
/*                                                                           */
/* See file CREDITS for list of people who contributed to this               */
/* project.                                                                  */
/*                                                                           */
/* This program is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU General Public License as            */
/* published by the Free Software Foundation; either version 2 of            */
/* the License, or (at your option) any later version.                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston,                    */
/* MA 02111-1307, USA.                                                       */
/*                                                                           */
/* email: support.automation@softing.com                                     */
/* Softing AG, Richard-Reitzner-Allee 6, 85540 Haar, Germany                 */
/*                                                                           */
/*****************************************************************************/
/*****************************************************************************/
/*                                                                           */
/*  Author      : SOFTING                                                    */
/*                                                                           */
/*  CHANGE_NOTES                                                             */
/*                                                                           */
/*  date      name      change                                               */
/*  -----------------------------------------------------------------------  */
/*****************************************************************************/
/*****************************************************************************/

/***************************************************************************/
/*                                                                         */
/*          LINUX PROFIBUS DEVICE DRIVER FOR SOFTING PROFIBOARD            */
/*                                                                         */
/* The data structure CtrlExt has entries for 4 possible boards. For each  */
/* board the hw-resources are stored. In addition this data structure      */
/* contains sub structures for (direct) pointers to the DP-Ram memory      */
/* of the different services (for example start address of DP input        */
/* and output areas).                                                      */
/*                                                                         */
/* For DP two bottom half interrupt service threads exist. Only one        */
/* of them is used, they are switched - if necessary - in the ISR.         */
/*                                                                         */
/* The device driver uses _NOT_ a dynamic major number. A fixed value has  */
/* been choosen (122). It has been written, in mind, that several boards   */
/* have to be supported (later) and that it can be ported to other         */
/* hardware platforms.                                                     */
/*                                                                         */
/* Data device: The driver offers two access modes to the data device:     */
/*              - the normal mode and the                                  */
/*              - optimized mode.                                          */
/*              In the NORMAL mode a read call to the data device returns  */
/*              immideatly with the requested data.                        */
/*              In the OPTIMIZED mode a read call to the data device       */
/*              will only return immideatly, if the data image had been    */
/*              changed since the last successful read call. In case the   */
/*              device had been opened in blocking mode, the calling       */
/*              thread/process will be blocked, until real new data is     */
/*              available, in case of non-blocking mode the function       */
/*              returns immideatly but without any data (count=0).         */
/*              The OPTIMIZED mode can only be enabled globally. This      */
/*              means, that if the data device is opened several times     */
/*              and via one file descriptor the device is set to the       */
/*              OPTIMIZED mode, this mode is used for every fd!            */
/*                                                                         */
/* init_module:                                                            */
/*   - searching of HW (board)                                             */
/*   - allocation of HW-resources                                          */
/*   - initializing of CtrlExt                                             */
/*                                                                         */
/*                                                                         */
/*                                                                         */
/* ----                                                                    */
/* A good resource for information about writing Linux-Device-Drivers is:  */
/* Alessandro Rubini: LINUX Device Drivers, O'Reilly & Associates, INC,    */
/*                    February 1998                                        */
/***************************************************************************/

#ifdef PB_ISA_SUPPORT

#if defined(PB_ISA_SUPPORT) && defined (PB_PCI_SUPPORT)
#error this driver does not support mixing of isa and pci boards
#endif


#include <linux/module.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,33)
#include <linux/autoconf.h>
#endif
#include <asm/io.h>
#include <asm/segment.h>

#include "basetypes.h"
#include "kernel_version.h"

#include <linux/poll.h>


#include <linux/slab.h>
#include <linux/proc_fs.h>

#include "pb_debug.h"
#include "cntrl.h"
#include "cmi.h"
#include "xx_if.h"
#include "xx_err.h"
#include "pb_xxx.h"
#include "fw.h"
#include "pb_ioctl.h"




#define PB_IO_START 0x200
#define PB_IO_END   0x400
#define PB_IO_RANGE 0x04


static char PbIrqList[] = { 5, 10, 11, 12, 15, 0 };
static unsigned int       major_number;

extern struct file_operations PrFileOps;



void PbDataISTFirst( TASKPARATYPE Args );
void PbDataIST( TASKPARATYPE Args );
void PbSlaveIST( TASKPARATYPE Args );

void await_irq_reset( struct CntrlExt *CtrlExt);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
IRQRETTYPE profibus_irq( int irq, void *dev_id, struct pt_regs *regs );
#else
irqreturn_t profibus_irq( int irq, void *dev_id );
#endif
    
extern int cmi_init( struct CntrlExt *CtrlExt );


extern char *Version;
extern char *Copyright;
extern char *Warranty;
extern struct semaphore PbGlobalSema;

#ifdef MODULE
// The names of the following data structure are exact the name,
// which have to be specified in the "commandline" (when loading the module).
int irq[] = { 0,0,0,0 };
int io[] = { 0,0,0,0 };
int mem[] = { 0,0,0,0 };
int dbg[] = { 0,0,0,0 };
#endif



static int PbCheckCard( int IOPort )
{
  // magic identification for PROFIboard 0x96 0x26 0x6b
  // magic identification for PROFI104  0x2c 0x12 0x41
  // magic identification for PROFI104s 0x2c 0x12 0x91
  
  {
    if (inb_p(IOPort)  ==0x96 &&
        inb_p(IOPort+1)==0x26 &&
        inb_p(IOPort+2)==0x6b) 
    {
      pr_info("PROFIboard found at 0x%x!\n", IOPort);
      return( 0 );
    } 
    else if (inb_p(IOPort)  ==0x2c &&
             inb_p(IOPort+1)==0x12 &&
             inb_p(IOPort+2)==0x41) 
               
    {
      pr_info("PROFI104 found at 0x%x!\n", IOPort);
      return( 0 );
    } 
    else if (inb_p(IOPort)  ==0x2c &&
             inb_p(IOPort+1)==0x12 &&
             inb_p(IOPort+2)==0x91) 
    {
      pr_info("PROFI104s found at 0x%x!\n", IOPort);
      return( 0 );
    }
  } /* check_region */

  return( -1 );
}


void PbStartCntrl( struct CntrlExt *Board )
{
  Board->BoardCtrl |= C165_RESET;
  outb_p( Board->BoardCtrl, Board->port+BOARD_CTRL );
}


int PbInitBoard( struct CntrlExt *Board, char Reset )
{
  
 // printk("PbInitBoard\n");
  PR_DEBUG(D_ALL_L1, "profiboard: starting board (0x%x - %d - 0x%lx)\n",
           Board->port, Board->interrupt, Board->RealDprAddr );

  if (Board->port)
  {
    outb_p( 0, Board->port+PAGE_SEL );
    outb_p( 0, Board->port+BOARD_CTRL );

    // XXX - Autsch, busy loop!
    udelay( 35 ); // IRQ has to be activated for at least 25 usec

    outb_p( (u8)(Board->RealDprAddr>>16), Board->port+BASE_ADD );

    outb_p( PAGE_0, Board->port+PAGE_SEL );
    Board->BoardCtrl = 0;
    Board->BoardCtrl |= SIZE_16_64;
    Board->BoardCtrl |= EN_DPRAM;
    Board->BoardCtrl &= ~(BASE_A14 + BASE_A15);
    Board->BoardCtrl |= ((u8)(Board->RealDprAddr>>11)&(BASE_A14+BASE_A15));
    Board->BoardCtrl |= MEMCS16_LA_SA;
    Board->BoardCtrl |= Reset; // C165_RESET or 0

    outb_p( Board->BoardCtrl, Board->port+BOARD_CTRL );
  }

 // printk("PbInitBoard done!\n");
  
  return( 0 );
}

/***********************/
/* Resource Management */
/***********************/

static int PbGetNextFreeIRQ( int Irq, struct CntrlExt *Board )
{
  int i; 
  size_t count;
  T_PROFI_SERVICE_DESCR sdb;

  if( Irq==0 ) {
    for( i=0, Irq=PbIrqList[0]; Irq; Irq=PbIrqList[++i] ) 
    {
      PR_DEBUG(D_ALL_L1,"checking irq %d\n", Irq );
      if(request_irq(Irq,profibus_irq,IRQF_DISABLED,"Pb104",(void*)Board)==0) 
      {
        // now test the interrupt
        Board->NrIrqs = 0;
        Board->interrupt = Irq;
        printk("checking irq %u\n",Irq);
        if( PbStartBoard( Board, C165_RESET ) >= 0 ) 
        {
          if( Board->TypeOfFW == PB_MASTER_FW ) 
          {
            sdb.layer = FMB;
            sdb.service = FMB_EXIT;
          } else {
            sdb.layer = DPS;
            sdb.service = DPS_EXIT_SLAVE;
          }
          sdb.comm_ref = 0;
          sdb.primitive = REQ;
          sdb.invoke_id = 0;
          count=sizeof(T_PROFI_SERVICE_DESCR);
          cmi_write(Board,&sdb, &count, PB_KERNEL_SPACE);
        }

        if( Board->NrIrqs == 0 ) 
        {
          printk(KERN_INFO "profiboard: no irq reported by probe\n");
          free_irq( Irq, Board );
          Irq = 0;
          continue;
        }
        //free_irq( Irq, Board );
        //request_irq(Irq,profibus_irq,0,PB_NAME,(void*)Board);
        break;
      }
    }
  } 
  else if( request_irq( Irq, profibus_irq, IRQF_DISABLED, "Pb104", Board )!=0 ) 
  {
    printk("request_irq %u,.. failed\n",Irq);
    Irq = 0;
  }
  PR_DEBUG(D_ALL_L1,"PbGetNextFreeIrq: %d\n", Irq );
  return( Irq );
}

/* The function for probing for an ISA memory area has been written by */
/* Allessandro Rubini, Linux Device Drivers, page 32-33.               */

static PB_ADDRESS PbProbeMem( PB_ADDRESS add )
{
  volatile unsigned char *ptr;
  unsigned char oldval, newval;
  USIGN32 i;

  ptr = (unsigned char*) add;

  //printk("probemem %lx\n", add);
  /* Expansion rom (executed at boot time by the bios) has a signature */
  /* of 0x55, 0xaa, and the third byte tells the size of such rom      */
  if ((read_io_byte (ptr) == 0x55) && (read_io_byte (ptr+1) == 0xaa)) {
    PR_DEBUG(D_ALL_L1,"%lx: Expansion ROM\n", (long)ptr );
    return (0L);
  }

  oldval = read_io_byte (ptr);
  write_io_byte (ptr, oldval^0xff);
  newval = read_io_byte (ptr);
  write_io_byte (ptr, oldval);

  /* FIXME--user getmem_fromio or such */
  if ((oldval ^ newval) == 0xff) 
  {
    PR_DEBUG(D_ALL_L1,"%lx: RAM\n", (long) ptr);
    return (0L);
  }
  if ((oldval ^ newval) != 0) 
  {
    PR_DEBUG(D_ALL_L1,"%lx: empty\n", (long) ptr);
    return (add);
  }

  /* If the tests above failed, we still don't know if it is ROM or empty. */
  /* Since empty memory can appear as 0x00, 0xff, or the low address byte, */
  /* we must probe multiple bytes: if at least one of them is different    */
  /* from these three values, then this is rom (though not boot rom).      */
  for (i = 0; i < 5; i++) 
  {
    ptr += 57; /* a random value */
    if ((read_io_byte (ptr)                       ) &&
         (read_io_byte (ptr) != 0xff               ) &&
         (read_io_byte (ptr) != ((long) ptr & 0xff)) )
      break;
  }
  if (i == 5) 
  {
    PR_DEBUG(D_ALL_L1,"%lx: DPRAM found\n", (long) ptr);
    return (add);
  } 
  else 
  {
    PR_DEBUG(D_ALL_L1,"%lx: ROM\n", (long) ptr);
    return (0L);
  }
}

static int PbGetNextFreeMem( int Io, int Irq, int Mem )
{
  static int i=0;
  static PB_ADDRESS UnusedMem[] = {
    0xc0000, 0xc4000, 0xc8000, 0xcc000,
    0xd0000, 0xd4000, 0xd8000, 0xdc000,
    0xb0000, 0xb4000, 0xb8000, 0xbc000,
    0xe0000, 0xe4000, 0xe8000, 0xec000,
    0xf0000, 0xf4000, 0xf8000, 0xfc000, 0 };
    
    PB_ADDRESS Memory = Mem;
    
    if (Mem)
    {
      Memory = (PB_ADDRESS) io_remap (Mem, 0x4000);
      
      if (check_mem_region (Memory, 0x4000))
	return (0);
    }
    
    if( Mem && PbProbeMem( Memory ) ) 
    {
      PR_DEBUG(D_ALL_L1,"PbGetNextFreeMem: 0x%x\n", Mem );
      io_unmap ((void*) Memory);
      return( Mem );
    } 
    else 
    {
      PB_ADDRESS j=0L;
      while( !j && UnusedMem[i] ) 
      {
	// Check, whether the start of the page is free
	Memory = (PB_ADDRESS) io_remap (UnusedMem[i], 0x4000);
	for( ; !PbProbeMem(Memory); i++ ) 
	{
	  io_unmap ((void*) Memory);
	  Memory = (PB_ADDRESS) io_remap (UnusedMem[i], 0x4000);
	  ;
	}
	// check, whether the page is completly free
	for( j=UnusedMem[i]+0x4000; j<(UnusedMem[i]+0x4000L); j+=0x4000 ) 
	{
	  if( PbProbeMem( Memory ) == 0 ) 
	  { 
	    // no, page is not free
	    j = 0;
	    i++; // check next page
	    io_unmap ((void*) Memory);
	    break;
	  }
	}
      }
      //PR_DEBUG("PbGetNextFreeMem: 0x%x\n", UnusedMem[i] );
      PR_DEBUG(D_ALL_L1,"PbGetNextFreeMem: 0x%lx\n", UnusedMem[i] );
      io_unmap ((void*) Memory);
      return( UnusedMem[i++] );
    }
}

void irq_to_cntrl( struct CntrlExt *CtrlExt, u8 IrqValue)
{
  char volatile *h_to_c_irq_val;


  await_irq_reset( CtrlExt );
  h_to_c_irq_val = (char *)CtrlExt->DprAddr + H_TO_C_IRQ_VAL_OFFSET;
    
  if( IrqValue == 0 )
    printk(KERN_ERR "illegal Interrupt: irq_to_cntrl( %p, %d)\n",
           CtrlExt, IrqValue );

  write_io_byte( h_to_c_irq_val, IrqValue); // set IRQ identifier

  outb_p( 0, CtrlExt->port + CP_IRQ );

  PR_DEBUG(D_OTHER_L3,"profiboard(irq_to_cntrl): %d h2c: 0x%X\n", IrqValue,
           read_io_byte( h_to_c_irq_val ) );
  return;
}

int PbStartBoard( struct CntrlExt *Board, char Reset )
{
  char version_string [VERSION_STRING_LENGTH];

  //  printk("PbStartBoard\n");
  if ( (Board->port == 0) && (Board->ConfigBase == 0) ) 
  {
    return( -1 );
  }

  PbInitBoard( Board, Reset );
  if( cmi_init( Board ) < 0 ) 
  {
    printk("profiboard: CONTROLLER NOT RESPONDING\n");
    return( -1 );
  }
  Board->Service.CMIClosed = 0;
  // check board type by reading the version string at offset 0x60
  // The board needs a few micro-seconds to write the magic strings into the
  // DPRAM.
  Board->TypeOfFW = PB_MASTER_FW;
  copy_from_io (version_string,
                (char*) (Board->DprAddr + PB_VERSION_OFFSET),
                sizeof (version_string));

  if (strstr (version_string, "PROFI")) 
  {
    PR_DEBUG(D_ALL_L1, "now checking type ...");
    if (strstr (version_string, "DPS")) 
    {
      PR_DEBUG(D_ALL_L1, "DP-Slave software\n");
      Board->TypeOfFW = PB_DP_SLAVE_FW;
    } 
    else if (strstr (version_string, "FMS")) 
    {
      PR_DEBUG(D_ALL_L1, "Master-Software\n");
      Board->TypeOfFW = PB_MASTER_FW;
    } 
    else 
    {
      printk("Firmware type (master/slave) not identified!\n");
    }
  }

//  printk("PbStartBoard done!\n");
  
  return( 0 );
}


void unloadProfiIsaDriver(void)
{
  int i;

  for( i=0; i<PB_MAX_NR_BOARDS; i++ ) 
  {
    tasklet_kill(&CtrlExt [i].PbSlaveTasklet);
    tasklet_kill(&CtrlExt [i].PbFirstTasklet);
    tasklet_kill(&CtrlExt [i].PbDataTasklet);
    
    if( io[i] ) 
    {
      outb_p( 0, CtrlExt[i].port+BOARD_CTRL );
      release_region( io[i], PB_IO_RANGE );
      if(CtrlExt[i].interrupt)
      {
        free_irq( CtrlExt[i].interrupt, &CtrlExt[i] );
        CtrlExt[i].interrupt = 0;
      }
      if(CtrlExt[i].DprAddr)
      {
        io_unmap ((void*) CtrlExt[i].DprAddr);
        CtrlExt[i].DprAddr = 0;
      }
      CtrlExt[i].DprAddr = 0;
      CtrlExt[i].RealDprAddr = 0;
      
      if( CtrlExt[i].FWVersion ) 
      {
        kfree( CtrlExt[i].FWVersion );
        CtrlExt[i].FWVersion = NULL;
      }
    }
  }


  return;

  
}


int initProfiIsaDriver(void)
{
  struct CntrlExt *Board;
  int i, IOCheck = PB_IO_START;

  char LocBuf[256];
  unsigned short BufSize=sizeof(LocBuf);
  int  Length;
  
  
  
  printk( Version );
  printk( Copyright );
  printk( Warranty );
  
  printk("\n **** initProfiIsaDriver: ****\n");
  
  // initialize structures (esp. WaitQueue)
  memset( CtrlExt, 0, sizeof(CtrlExt) );
  for( i=0; i< PB_MAX_NR_BOARDS; i++ ) 
  {
    init_waitqueue_head (&CtrlExt [i].Service.QRead);
    init_waitqueue_head (&CtrlExt [i].Service.QWrite);
    init_waitqueue_head (&CtrlExt [i].Data.QNewInputData);
    init_waitqueue_head (&CtrlExt [i].SlaveIn.QInputAtController);
    init_my_wait(&CtrlExt [i].Service.wait_for_read, "wait_for_read");
    init_my_wait(&CtrlExt [i].Service.wait_for_write, "wait_for_write");
    init_my_wait(&CtrlExt [i].Data.wait_for_new_input_data, "wait_for_new_input_data");
    
    
    spin_lock_init(&CtrlExt [i].Spinlock);
    tasklet_init(&CtrlExt [i].PbSlaveTasklet, PbSlaveIST, (TASKPARATYPE) &CtrlExt [i]);
    tasklet_init(&CtrlExt [i].PbFirstTasklet, PbDataISTFirst, (TASKPARATYPE)&CtrlExt [i]);
    tasklet_init(&CtrlExt [i].PbDataTasklet, PbDataIST, (TASKPARATYPE)&CtrlExt [i]);
    CtrlExt [i].DeviceType = SOFTING_PB_ISA_DEVICE;
    CtrlExt [i].asic = 0; // can not be detected ==> unknown == 0

  }

  for( i=0; i< PB_MAX_NR_BOARDS; i++ ) 
  {
    // find the card in the IO space
    printk("io[%d]=0x%x irq[%d]=%d mem[%d]=0x%x dbg[%d]=0x%x\n",
           i, io[i], i, irq[i], i, mem[i], i, dbg[i] );

    if( io[i] == 0 ) 
    {
      for( ; IOCheck<PB_IO_END; IOCheck+=PB_IO_RANGE ) 
      {
        if( PbCheckCard(IOCheck) == 0 ) 
        {
          io[i] = IOCheck;
          IOCheck += PB_IO_RANGE;
          break;
        }
      }
    } 
    else 
    {
      if( PbCheckCard(IOCheck) != 0 ) 
      {
        printk("profiboard: No board found at io=0x%x\n", io[i] );
        io[i] = 0;
        break;
      } 
      else 
      {
        IOCheck = io[i];
      }
    }
    if( io[i] ) 
    {
      // trying to find a valid memspace
      // but, first we reserve the IO space
      request_region( io[i], PB_IO_RANGE, PB_NAME );
      mem[i] = PbGetNextFreeMem( io[i], irq[i], mem[i] );
      if( mem[i]==0 ) 
      {
        printk("profiboard: Can't find DPR-Address.\n");
        release_region( io[i], PB_IO_RANGE );
        io[i]=0;
        break;
      }
      else
      {
        printk("profiboard: dpram 0x%x.\n", mem[i]);
      }
      Board = &CtrlExt[i];
      Board->port = io[i];
      Board->RealDprAddr = mem[i];
  
      Board->DprAddr = (PB_ADDRESS) io_remap (Board->RealDprAddr, 0x4000);
      Board->c_to_h_irq_val =  (unsigned short *)(Board->DprAddr + C_TO_H_IRQ_VAL_OFFSET);
      Board->h_to_c_irq_val =  (unsigned short *)(Board->DprAddr + H_TO_C_IRQ_VAL_OFFSET);
     
      // printk("profiboard: 0x%lx - 0x%lx.\n", Board->RealDprAddr, Board->DprAddr);
      if( Board->DprAddr==0  ) 
      {
        printk("profiboard: Can't remap DPR-Address.\n");
        release_region( io[i], PB_IO_RANGE );
        io[i]=0;
        break;
      }
      PbInitBoard( Board, 0 );
      irq[i] = Board->interrupt = PbGetNextFreeIRQ( irq[i], Board );
      if( irq[i] == 0 ) {
        printk("profiboard: Can't get free IRQ - giving up.\n");
        release_region( io[i], PB_IO_RANGE );
        io[i]=0;
        break;
      }
      PbInitBoard( Board, 0 );
      if( PbStartBoard( Board, C165_RESET ) >= 0 ) 
      {
        cmi_get_data(Board,ID_FW_VERS_IMAGE,0,&BufSize,LocBuf,
                     PB_KERNEL_SPACE );
        if(Board->FWVersion)
          kfree(Board->FWVersion);
        Board->FWVersion = kmalloc( strlen(LocBuf)+1, GFP_ATOMIC );
        if( Board->FWVersion ) 
        {
          memcpy( Board->FWVersion, LocBuf, strlen(LocBuf) );
        } 
        else 
        {
          Board->FWVersion=NULL;
        }
        Length = sizeof(Board->SerialNumber);
        cmi_get_data(Board,ID_SERIAL_DEVICE_NUMBER,0,(void *)&Length,
                     &Board->SerialNumber, PB_KERNEL_SPACE );
      }
      PbInitBoard( Board, 0 );
    }
  }
  if( io[0] ) 
  {
    if ((major_number = register_chrdev (PB_MAJOR, PB_NAME, &PrFileOps)) < 0) 
    {
      printk("profiboard: unable to register device\n" );
      return(major_number);
    }

    return( 0 );
  } 
  else 
  {
    printk(KERN_ERR "No resources available, driver not loaded!\n");
    return( -1 );
  } 
  return 0;
}

int init_module()
{
  
  int ret=0;
  
  printk( Version );
  printk( Copyright );
  printk( Warranty );
  
  // initialize structures (esp. WaitQueue)
  
  sema_init(&PbGlobalSema, 1);        // only 1 task may hold the semaphore
  
  
  memset( CtrlExt, 0, sizeof(CtrlExt) );



  ret = initProfiIsaDriver();


  
  return ret; 
}

void cleanup_module()
{

  unloadProfiIsaDriver();
 
  unregister_chrdev(major_number, PB_NAME);
  
}

#endif // PB_ISA_SUPPORT


