/*****************************************************************************/
/*                                                                           */
/*                              SOFTING AG                                   */
/*                        Richard-Reitzner-Allee 6                           */
/*                              D-85540 Haar                                 */
/*                      Phone: (++49)-(0)89-45656-0                          */
/*                      Fax:   (++49)-(0)89-45656-399                        */
/*                                                                           */
/*                  Copyright (C) SOFTING AG 1995-2004                       */
/*                                                                           */
/* See file CREDITS for list of people who contributed to this               */
/* project.                                                                  */
/*                                                                           */
/* This program is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU General Public License as            */
/* published by the Free Software Foundation; either version 2 of            */
/* the License, or (at your option) any later version.                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston,                    */
/* MA 02111-1307, USA.                                                       */
/*                                                                           */
/* email: support.automation@softing.com                                     */
/* Softing AG, Richard-Reitzner-Allee 6, 85540 Haar, Germany                 */
/*                                                                           */
/*****************************************************************************/
/*****************************************************************************/
/*                                                                           */
/*  Author      : SOFTING                                                    */
/*                                                                           */
/*  CHANGE_NOTES                                                             */
/*                                                                           */
/*  date      name      change                                               */
/*  -----------------------------------------------------------------------  */
/*****************************************************************************/
/*****************************************************************************/

#ifndef _PB_XXX_H_
#define _PB_XXX_H_

/* ------------------------------------------------------------------------- */
/* --- DEFINITIONS FOR FMB                                                   */
/* ------------------------------------------------------------------------- */
#define FMB_RESET                 20
#define FMB_EXIT                  21
#define FMB_EXCEPTION             28


/* ------------------------------------------------------------------------- */
/* --- DEFINITIONS FOR FMS                                                   */
/* ------------------------------------------------------------------------- */
#define FMS_ABORT                 38
#define FDLIF_SAP_DEACTIVATE      10

/* ------------------------------------------------------------------------- */
/* --- DEFINITIONS FOR DP-Master                                             */
/* ------------------------------------------------------------------------- */
#define DP_INIT_MASTER                  0xC1
#define DP_ACT_PARAM_LOC                0xC2
#define DP_EXIT_MASTER                  0xC4


#define DP_AAM_IO_BLOCKS                0x03
#define DP_MAX_NUMBER_STATIONS          127
#define DP_MAX_OUTPUT_DATA_LEN          244
#define DP_MAX_INPUT_DATA_LEN           DP_MAX_OUTPUT_DATA_LEN

/*--- DP STATION STATUS ----------------------------------------------------*/
#define DP_STATE_STATION_TYPE           0x80         /* station status bits */

#define DP_STATE_MASTER_STATION         0x80
#define DP_STATE_SLAVE_STATION          0x00

#define DP_STATE_SLAVE_DIAG_DATA        0x02
#define DP_STATE_MASTER_DIAG_DATA       0x02

#define DP_STATE_SLAVE_ERROR            0x01
#define DP_STATE_MASTER_UP_DOWN_LOAD    0x01




/*--- DP AREA / ACTIVATION CODES -------------------------------------------*/

/*--- DP AREA CODES --------------------------------------------------------*/

#define DP_AREA_SLAVE_PARAM         0x00
#define DP_AREA_BUS_PARAM           0x7F
#define DP_AREA_SET_MODE            0x80
#define DP_AREA_STAT_COUNT          0x81
#define DP_AREA_NO_PROTECTION       0xFF


#define DP_OP_MODE_OFFLINE          0x00
#define DP_OP_MODE_STOP             0x40
#define DP_OP_MODE_CLEAR            0x80
#define DP_OP_MODE_RED_CLEAR        0x81            /* new redundancy state */
#define DP_OP_MODE_OPERATE          0xC0
#define DP_OP_MODE_RED_OPERATE      0xC1            /* new redundancy state */


typedef struct _T_DP_INIT_MASTER_REQ
{
  USIGN8    master_default_address;       /* station address for default BP */
  PB_BOOL   master_class2;              /* use M2 requester functionality ? */
  STRINGV   master_class2_name [32];                   /* vendor name of M2 */

  USIGN8    lowest_slave_address;      /* station address of first DP Slave */

  USIGN8    slave_io_address_mode;  /* Address Assignment Mode of slave I/O */

  PB_BOOL   clear_outputs;      /* clear output areas in DP_OP_MODE_CLEAR ? */
  USIGN8    auto_remote_services;                            /* DP_AUTO_xxx */
  PB_BOOL   cyclic_data_transfer;    /* no DATA_TRANSFER service requests ? */

  USIGN8    dummy;                                        /* alignment byte */

} T_DP_INIT_MASTER_REQ;


typedef struct _T_DP_ACT_PARAM_REQ          /* MM: function_num --> rem_add */
{
  USIGN8    rem_add;                 /* station address DP Master (class 1) */
  USIGN8    area_code;                                       /* DP_AREA_xxx */
  USIGN8    activate;  /* DP_SLAVE_xxx, DP_BUS_PAR_ACTIVATE, DP_OP_MODE_xxx */
  USIGN8    dummy;                                    /* SDU alignment byte */

} T_DP_ACT_PARAM_REQ;

/*--------------------------------------------------------------------------*/

typedef struct _T_DP_ACT_PARAM_IND          /* MM: req_add --> function_num */
{
  USIGN8    req_add;                 /* station address DP Master (class 2) */
  USIGN8    area_code;                                       /* DP_AREA_xxx */
  USIGN8    activate;  /* DP_SLAVE_xxx, DP_BUS_PAR_ACTIVATE, DP_OP_MODE_xxx */
  USIGN8    dummy;                                    /* SDU alignment byte */

} T_DP_ACT_PARAM_IND;

/*--------------------------------------------------------------------------*/

typedef struct _T_DP_ACT_PARAM_RES_CON
{
  USIGN16   status;                  /* OK, DS, NA, RS, RR, UE, TO, FE, RE, */
                                          /* NE, AD, IP, SC, NI, DI, EA, LE */

} T_DP_ACT_PARAM_RES_CON;




/* ------------------------------------------------------------------------- */
/* --- DEFINITIONS FOR DP-Slave -------------------------------------------- */
/* ------------------------------------------------------------------------- */
#define DPS_EXIT_SLAVE   0x02
#define DPS_EXCEPTION    0x07


/*--- DPS INPUT / OUTPUT STATUS --------------------------------------------*/

#define DPS_INPUT_STATE_FREEZE_ENABLED    0x01
#define DPS_INPUT_STATE_FREEZE_COMMAND    0x02

#define DPS_OUTPUT_STATE_SYNC_ENABLED     0x01
#define DPS_OUTPUT_STATE_SYNC_COMMAND     0x02
#define DPS_OUTPUT_STATE_CLEAR_DATA       0x04
#define DPS_OUTPUT_STATE_VALID_DATA       0x08
#define DPS_OUTPUT_STATE_NEW_DATA         0x10
#define DPS_OUTPUT_STATE_GLOBAL_CONTROL   0x20


/*--- access to T_DPS_INPUT_DATA_HEADER --------------------------------------*/

#define DPS_INPUT_DATA_LEN_OFFSET     0x00
#define DPS_INPUT_STATE_OFFSET        0x01
#define DPS_INPUT_DATA_OFFSET         0x02


/*--- access to T_DPS_OUTPUT_DATA_HEADER -------------------------------------*/

#define DPS_OUTPUT_DATA_LEN_OFFSET    0x00
#define DPS_OUTPUT_STATE_OFFSET       0x01
#define DPS_OUTPUT_DATA_OFFSET        0x02

typedef struct _T_DPS_OUTPUT_DATA_HEADER
{
  USIGN8    output_data_len;     /* length of DP master output data, 0..244 */
  USIGN8    output_state;   /* status of output image: DPS_OUTPUT_STATE_xxx */

  USIGN8    output_data [DP_MAX_OUTPUT_DATA_LEN];
} T_DPS_OUTPUT_DATA_HEADER;

typedef struct _T_DPS_INPUT_DATA_HEADER
{
  USIGN8    input_data_len;     /* length of DP master output data, 0..244 */
  USIGN8    input_state;   /* status of output image: DPS_OUTPUT_STATE_xxx */

  USIGN8    input_data[DP_MAX_INPUT_DATA_LEN];
} T_DPS_INPUT_DATA_HEADER;

#endif
