#! /bin/sh
#

### BEGIN INIT INFO
# Provides:          Profibus interface
# Required-Start:    $syslog
# Should-Start: $time 
# Required-Stop:     $syslog
# Should-Stop: $time 
# Default-Start:     3 5
# Default-Stop:      0 1 2 6
# Short-Description: Softing AG Profibus driver
# Description:       Start profiboard.sh to provide access to the Softing Profibus hardware
#	
### END INIT INFO

#MODULE_PATH=/lib/modules/`uname -r`/kernel/drivers/softing
MODULE_PATH=`/sbin/modinfo -F filename profiboard`
module=profiboard
board_device=pbboard
service_device=pbservice
data_device=pbdata
slave_in_device=pbslin
slave_out_device=pbslout
mode=666

test -x /sbin/insmod || exit 0
test -x /sbin/rmmod  || exit 0
test -x /bin/egrep   || exit 0
test -f $MODULE_PATH || echo "module profiboard not found"

#test -f $MODULE_PATH/profiboard.ko || exit 0

case "$1" in
  start)
      echo "starting profibus driver"
      /sbin/modprobe profiboard

      echo "removing old profibus device files"

      rm -rf /dev/${board_device}*
      rm -rf /dev/${service_device}*
      rm -rf /dev/${data_device}*
      rm -rf /dev/${slave_in_device}*
      rm -rf /dev/${slave_out_device}*

      echo "creating new profibus device files"

      major=`cat /proc/devices | awk '$2~/profiboard/ {print $1}'`;
      for i in 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15; do
        /bin/mknod /dev/${board_device}$i   c ${major} $(($i+0));
        /bin/mknod /dev/${service_device}$i c ${major} $(($i+16));
        /bin/mknod /dev/${data_device}$i    c ${major} $(($i+32));
        /bin/mknod /dev/${slave_in_device}$i  c ${major} $(($i+48));
        /bin/mknod /dev/${slave_out_device}$i c ${major} $(($i+64));
      done

      echo "setting access rights for new profibus device files"

      /bin/chmod ${mode} /dev/${board_device}*
      /bin/chmod ${mode} /dev/${service_device}*
      /bin/chmod ${mode} /dev/${data_device}*
      /bin/chmod ${mode} /dev/${slave_in_device}*
      /bin/chmod ${mode} /dev/${slave_out_device}*

    ;;
  stop)

      echo "shutting down profibus driver"
      /sbin/rmmod profiboard

      echo "removing profibus device files"
      rm -rf /dev/${board_device}*
      rm -rf /dev/${service_device}*
      rm -rf /dev/${data_device}*
      rm -rf /dev/${slave_in_device}*
      rm -rf /dev/${slave_out_device}*
    ;;
  *)
    echo "Usage: profiboard.sh {start|stop}"
    exit 1
esac

echo "leaving profibus.sh"

exit 0
