/*****************************************************************************/
/*                                                                           */
/*                              SOFTING AG                                   */
/*                        Richard-Reitzner-Allee 6                           */
/*                              D-85540 Haar                                 */
/*                      Phone: (++49)-(0)89-45656-0                          */
/*                      Fax:   (++49)-(0)89-45656-399                        */
/*                                                                           */
/*                  Copyright (C) SOFTING AG 1995-2004                       */
/*                                                                           */
/* See file CREDITS for list of people who contributed to this               */
/* project.                                                                  */
/*                                                                           */
/* This program is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU General Public License as            */
/* published by the Free Software Foundation; either version 2 of            */
/* the License, or (at your option) any later version.                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston,                    */
/* MA 02111-1307, USA.                                                       */
/*                                                                           */
/* email: support.automation@softing.com                                     */
/* Softing AG, Richard-Reitzner-Allee 6, 85540 Haar, Germany                 */
/*                                                                           */
/*****************************************************************************/
/*****************************************************************************/
/*                                                                           */
/*  Author      : SOFTING                                                    */
/*                                                                           */
/*  CHANGE_NOTES                                                             */
/*                                                                           */
/*  date      name      change                                               */
/*  -----------------------------------------------------------------------  */
/*****************************************************************************/
/*****************************************************************************/

#ifndef _XX_ERR_H_
#define _XX_ERR_H_

/*****************************************************************************/
/*************            INTERFACE ERRORS               *********************/
/*****************************************************************************/

#define   E_IF_BOOT_DIAG_ERROR             5 /* HW error recognized by loader   */
#define   E_IF_STARTUP_DIAG_ERROR          6 /* HW error recognized by startup  */
#define   E_IF_FATAL_ERROR                 7 /* unrecoverable error on board    */
#define   E_IF_INIT_INVALID_PARAMETER      8 /* invalid initialize parameter    */
#define   E_IF_LOADER_ERROR                9 /* download firmware error         */
#define   E_IF_NO_CNTRL_RES               10 /* controller does not respond     */
#define   E_IF_INVALID_CNTRL_TYPE_VERSION 11 /* inv. controller type or SW vers.*/
#define   E_IF_INVALID_LAYER              12 /* invalid layer                   */
#define   E_IF_INVALID_SERVICE            13 /* invalid service identifier      */
#define   E_IF_INVALID_PRIMITIVE          14 /* invalid service primitive       */
#define   E_IF_INVALID_DATA_SIZE          15 /* not enough cmi data block memory*/
#define   E_IF_INVALID_COMM_REF           16 /* invalid communication reference */
#define   E_IF_INVALID_CMI_CALL           19 /* invalid CMI call                */
#define   E_IF_CMI_ERROR                  20 /* error occured in CMI            */
#define   E_IF_RESOURCE_UNAVAILABLE       21 /* no resource available           */
#define   E_IF_NO_PARALLEL_SERVICES       22 /* no parallel services allowed    */
#define   E_IF_SERVICE_CONSTR_CONFLICT    23 /* serv. tempor. not executable    */
#define   E_IF_SERVICE_NOT_SUPPORTED      24 /* service not supported           */
#define   E_IF_SERVICE_NOT_EXECUTABLE     25 /* service not executable          */
#define   E_IF_INVALID_VERSION            26 /* invalid version                 */
#define   E_IF_NO_CNTRL_PRESENT       28
#define   E_IF_INVALID_PARAMETER          30   /* wrong parameter in REQ or RES */
#define   E_IF_INIT_FAILED                31  /* init. API or Controller failed */
#define   E_IF_EXIT_FAILED                32   /* exit API or Controller failed */

/* additional error codes available only in Win NT */
#define E_IF_SLAVE_DIAG_DATA   0xf0     /* no data available new diagnostics data available */
#define E_IF_SLAVE_ERROR       0xf1     /* no data exchange */
#define E_IF_INVALID_DP_STATE  0xf2     /* DP is not in state operate */
#define E_IF_NT_ERROR          0xff     /* NT system error, get error detail with GetLastError */
/*****************************************************************************/

/*****************************************************************************/
/*************      EXCEPTION STRUCTURES         *****************************/
/*****************************************************************************/

#ifdef FF
typedef struct _T_EXCEPTION
{
  USIGN8  task_id;            /* task identifier in which execption occurs  */
  USIGN8  modul_id;           /* modul identifier in which execption occurs */
  USIGN16 line;               /* line in which execption occurs             */
  USIGN16 class_code;         /* exception class and code                   */
  USIGN16 add_detail;         /* addition detail                            */
} T_EXCEPTION;
#endif

#ifdef PB
typedef struct _T_EXCEPTION
{
  USIGN8  task_id;              /* task identifier in which execption occurs */
  USIGN8  par1;                                               /* parameter 1 */
  USIGN16 par2;                                               /* parameter 2 */
  USIGN16 par3;                                               /* parameter 3 */
} T_EXCEPTION;
#endif

#endif

