/******************************************************************************
*                                                                             *
*                                 SOFTING AG                                  *
*                          Richard-Reitzner-Allee 6                           *
*                                D-85540 Haar                                 *
*                        Phone: (++49)-(0)89-45656-0                          *
*                        Fax:   (++49)-(0)89-45656-399                        *
*                                                                             *
*                      Copyright (C) SOFTING AG 1995-2003                     *
*                              All Rights Reserved                            *
*                                                                             *
*******************************************************************************

FILE_NAME               FMSDEMO.C

PROJECT_NAME            PROFIBUS

MODULE                  FMSDEMO

COMPONENT_LIBRARY       --------

AUTHOR                  SOFTING AG

VERSION                 5.24

DATE                    31.07.2001

STATUS                  finished

FUNCTIONAL_MODULE_DESCRIPTION

FMS Demo Application
--------------------

RELATED_DOCUMENTS

============================================================================*/
#include <keywords.h>

INCLUDES

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined (WIN16) || defined (WIN32)
#include <windows.h>
#include <conio.h>
#include <dos.h>
#else
#ifndef _LINUX
#include <graph.h>
#endif
#endif

#ifdef _LINUX
#include <unistd.h>
#include <ctype.h>
#endif

#include <errno.h>

#include "pb_type.h"
#include "pb_conf.h"
#include "pb_if.h"
#include "pb_fms.h"
#include "pb_fm7.h"
#include "pb_err.h"
#include "pb_fmb.h"
#define __PB_FM7__
#include "pb_tools.h"


GLOBAL_DEFINES

LOCAL_DEFINES


/*----------- constants for program configuration ---------------*/

#if defined (WIN32) || defined (_LINUX)
#define BOARD_NUMBER                    0
#else
#define H_DPR_BASE_ADDRESS              0xD0000000
#define IO_ADDRESS                      0x0240
#endif
#define TS5_TXT                         "station 5  (client)"
#define TS5                             5
#define TS10_TXT                        "station 10 (server)"
#define TS10                            10

#define WAIT                            0xFF
#define DONT_WAIT                       0x00

/* --- defines for OD object descriptions ----------------------------------*/
#define LEN_ST_OD             19
#define LEN_S_OD              10
#define LEN_DV_OD              0
#define LEN_DP_OD              0
#define NO_OF_OD_OBJECT_DESCR  (LEN_ST_OD + LEN_S_OD + LEN_DV_OD + LEN_DP_OD + 1)

#ifdef _LINUX
#define stricmp(str1, str2)         strcasecmp(str1, str2)
#endif


EXPORT_TYPEDEFS

LOCAL_TYPEDEFS

FUNCTION_DECLARATIONS


static VOID  wait_for_cnf_ind(USIGN8);
static VOID  print_version(VOID);
static VOID  print_papi_firmware_version(VOID);

EXPORT_DATA

IMPORT_DATA

LOCAL_DATA

#if defined (WIN16) || defined (WIN32)
static char                    exe_fname [_MAX_FNAME];  /* name of .exe file */
#endif

#if defined (WIN32) || defined (_LINUX)
static short unsigned int      board_number;
#else
static long  unsigned int      dpr_base_address;
static short unsigned int      io_address;
#endif

static T_PROFI_SERVICE_DESCR   sdb;
static USIGN8                  cnf_ind_buffer[512];
static USIGN8                  req_rsp_buffer[512];
static INT16                   ret_val;
static USIGN32                 default_variable_object_20_data;
static USIGN32                 default_variable_object_21_data;
static INT8                    invoke_id;
static USIGN8                  ts;

static PB_BOOL                 multi_read_variable=PB_FALSE;
static PB_BOOL                 multi_write_variable=PB_FALSE;
static USIGN16                 wait_for_exit;
static USIGN8                  info_rpt_priority;
static USIGN16                 act_cr;
static USIGN32                 irq_rcv_ctr;

static T_LOAD_CRL_REQ          crl[10];
static T_LOAD_OD_REQ           od[NO_OF_OD_OBJECT_DESCR+1];


/* ------------------------------------------------------------------------- */
/* local aux. functions:                                                     */
/* - to compute an invoke id                                                 */
/* - to print interface errors                                               */
/* - build CRL                                                               */
/* - build OD                                                                */
/* ------------------------------------------------------------------------- */


FUNCTION static INT8 
    compute_invoke_id(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to compute the invoke_id

possible return values:
- invoke_id

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

invoke_id++;
if (invoke_id == 127) invoke_id = 0;

return(invoke_id);
}


FUNCTION static VOID 
    print_if_error(
           IN USIGN8           primitive,
           IN INT16            error
         )

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

This function is used to print interface errors

possible return values:
- NONE

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_EXCEPTION exception;
USIGN16     size = sizeof(T_EXCEPTION);

FUNCTION_BODY

if (primitive == REQ || primitive == RES)
{
    printf("\n profi_snd_req_res failed\n");
}
else if (primitive == IND || primitive == CON)
{
    printf("\n profi_rcv_con_ind returns with error\n");
}
else
{
}

switch (error)
{
  case E_IF_FATAL_ERROR:
       printf(" unrecoverable error on board     \n");
       while (profi_get_data(ID_EXCEPTION_IMAGE,
                             0,
                             (USIGN16 FAR*) &size,
                             (VOID FAR*) &exception
                            ) == E_IF_SERVICE_CONSTR_CONFLICT);

       printf(" task_id : %d (dez) %X (hex)\n",exception.task_id,exception.task_id);
       printf(" par1    : %d (dez) %X (hex)\n",exception.par1,exception.par1);
       printf(" par2    : %d (dez) %X (hex)\n",exception.par2,exception.par2);
       printf(" par3    : %d (dez) %X (hex)\n",exception.par3,exception.par3);
       break;

  case E_IF_INIT_INVALID_PARAMETER:    printf(" invalid initialize parameter     \n");
                                       break;
  case E_IF_LOADER_ERROR:              printf(" download firmware error          \n");
                                       break;
  case E_IF_NO_CNTRL_RES:              printf(" controller does not respond      \n");
                                       break;
  case E_IF_INVALID_CNTRL_TYPE_VERSION:printf(" inv. controller type or SW vers. \n");
                                       break;
  case E_IF_INVALID_LAYER:             printf(" invalid layer                    \n");
                                       break;
  case E_IF_INVALID_SERVICE:           printf(" invalid service identifier       \n");
                                       break;
  case E_IF_INVALID_PRIMITIVE:         printf(" invalid service primitive        \n");
                                       break;
  case E_IF_INVALID_DATA_SIZE:         printf(" not enough cmi data block memory \n");
                                       break;
  case E_IF_INVALID_COMM_REF:          printf(" invalid communication reference  \n");
                                       break;
  case E_IF_INVALID_CMI_CALL:          printf(" invalid CMI call                 \n");
                                       break;
  case E_IF_CMI_ERROR:                 printf(" error occured in CMI             \n");
                                       break;
  case E_IF_RESOURCE_UNAVAILABLE:      printf(" no resource available            \n");
                                       break;
  case E_IF_NO_PARALLEL_SERVICES:      printf(" no parallel services allowed     \n");
                                       break;
  case E_IF_SERVICE_CONSTR_CONFLICT:   printf(" serv. tempor. not executable     \n");
                                       break;
  case E_IF_SERVICE_NOT_SUPPORTED:     printf(" service not supported            \n");
                                       break;
  case E_IF_SERVICE_NOT_EXECUTABLE:    printf(" service not executable           \n");
                                       break;
  case E_IF_INVALID_VERSION:           printf(" no valid access to controller    \n");
                                       break;
  case E_IF_NO_CNTRL_PRESENT:          printf(" controller not available         \n");
                                       printf(" additonal error: 0x%2x \n",profi_get_last_error());
                                       break;
  case E_IF_INVALID_PARAMETER:         printf(" wrong parameter in REQ or RES    \n"); break;
  case E_IF_INIT_FAILED:               printf(" initialize of PAPI failed\n");
                                       printf(" additonal error: %x \n",profi_get_last_error());
                                       break;
  case E_IF_EXIT_FAILED:               printf(" exit of PAPI failed     \n");
                                       printf(" additonal error: %x \n",profi_get_last_error());
                                       break;
  case E_IF_OS_ERROR:                  printf(" Operating system error: (0x%02x)",profi_get_last_error());
                                       break;
  default:                             printf(" unknown IF error   %d            \n",error); break;
}
printf("\n -> press return \n"); getch();
fflush(stdout);
return;
}


FUNCTION static VOID 
    build_crl(
           IN USIGN8 ts
         )

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to build the crl

possible return values:

- == E_OK    -> no error
- != E_OK    -> error

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY


/* --- build CRL header ---------------------------------------------------- */
crl[0].desired_cr                          = 0;
crl[0].id.crl_hdr.nr_of_entries            = 8;
crl[0].id.crl_hdr.poll_sap                 = 0;
crl[0].id.crl_hdr.ass_abt_ci               = 1000;
crl[0].id.crl_hdr.symbol_length            = 20;
crl[0].id.crl_hdr.vfd_pointer_supported    = PB_FALSE;

crl[1].desired_cr                           = (USIGN16) 1;
crl[1].id.crl_static.loc_lsap               = (USIGN8)  1;
crl[1].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
crl[1].id.crl_static.rem_lsap               = (USIGN8)  255;
crl[1].id.crl_static.lli_sap                = (USIGN8)  1;
crl[1].id.crl_static.conn_type              = (USIGN8)  MSAC;
crl[1].id.crl_static.conn_attr              = (USIGN8)  O_CONN;
crl[1].id.crl_static.ci                     = (USIGN32) 1000;
crl[1].id.crl_static.multiplier             = (USIGN8)  0;
crl[1].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
crl[1].id.crl_static.max_pdu_snd_low        = (USIGN8)  0;
crl[1].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
crl[1].id.crl_static.max_pdu_rcv_low        = (USIGN8)  0;
crl[1].id.crl_static.rem_add                = (USIGN8)  255;
crl[1].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;   /* as client */
crl[1].id.crl_static.feature_supp[1]        = (USIGN8)  0x7E;   /* as server */
crl[1].id.crl_static.feature_supp[2]        = (USIGN8)  0x00;
crl[1].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
crl[1].id.crl_static.feature_supp[4]        = (USIGN8)  0x00;
crl[1].id.crl_static.feature_supp[5]        = (USIGN8)  0x00;
crl[1].id.crl_static.max_scc                = (USIGN8)  1;
crl[1].id.crl_static.max_rcc                = (USIGN8)  1;
crl[1].id.crl_static.max_sac                = (USIGN8)  0;
crl[1].id.crl_static.max_rac                = (USIGN8)  0;
crl[1].id.crl_static.vfd_pointer            = 1;
crl[1].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
crl[1].id.crl_static.symbol[0]              = 20;
strcpy(&crl[1].id.crl_static.symbol[1],"MSAC (Slave) (FM7)");


/* --- build CRL entry ----------------------------------------------------- */
if (ts == TS5)
{
    crl[2].desired_cr                           = (USIGN16) 2;
    crl[2].id.crl_static.loc_lsap               = (USIGN8)  2;
    crl[2].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[2].id.crl_static.rem_lsap               = (USIGN8)  2;
    crl[2].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[2].id.crl_static.conn_type              = (USIGN8)  MMAC;
    crl[2].id.crl_static.conn_attr              = (USIGN8)  D_CONN;
    crl[2].id.crl_static.ci                     = (USIGN32) 1000;
    crl[2].id.crl_static.multiplier             = (USIGN8)  0;
    crl[2].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
    crl[2].id.crl_static.max_pdu_snd_low        = (USIGN8)  241;
    crl[2].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
    crl[2].id.crl_static.max_pdu_rcv_low        = (USIGN8)  241;
    crl[2].id.crl_static.rem_add                = (USIGN8)  10;
    crl[2].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[2].id.crl_static.feature_supp[1]        = (USIGN8)  0x30;
    crl[2].id.crl_static.feature_supp[2]        = (USIGN8)  0x80;
    crl[2].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[2].id.crl_static.feature_supp[4]        = (USIGN8)  0x30;
    crl[2].id.crl_static.feature_supp[5]        = (USIGN8)  0x80;
    crl[2].id.crl_static.max_scc                = (USIGN8)  3;
    crl[2].id.crl_static.max_rcc                = (USIGN8)  3;
    crl[2].id.crl_static.max_sac                = (USIGN8)  3;
    crl[2].id.crl_static.max_rac                = (USIGN8)  3;
    crl[2].id.crl_static.vfd_pointer            = 1;
    crl[2].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[2].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[2].id.crl_static.symbol[1],"MMAC               ");
}
else
{
    crl[2].desired_cr                           = (USIGN16) 2;
    crl[2].id.crl_static.loc_lsap               = (USIGN8)  2;
    crl[2].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[2].id.crl_static.rem_lsap               = (USIGN8)  2;
    crl[2].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[2].id.crl_static.conn_type              = (USIGN8)  MMAC;
    crl[2].id.crl_static.conn_attr              = (USIGN8)  D_CONN;
    crl[2].id.crl_static.ci                     = (USIGN32) 1000;
    crl[2].id.crl_static.multiplier             = (USIGN8)  0;
    crl[2].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
    crl[2].id.crl_static.max_pdu_snd_low        = (USIGN8)  241;
    crl[2].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
    crl[2].id.crl_static.max_pdu_rcv_low        = (USIGN8)  241;
    crl[2].id.crl_static.rem_add                = (USIGN8)  5;
    crl[2].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[2].id.crl_static.feature_supp[1]        = (USIGN8)  0x30;
    crl[2].id.crl_static.feature_supp[2]        = (USIGN8)  0x80;
    crl[2].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[2].id.crl_static.feature_supp[4]        = (USIGN8)  0x30;
    crl[2].id.crl_static.feature_supp[5]        = (USIGN8)  0x80;
    crl[2].id.crl_static.max_scc                = (USIGN8)  3;
    crl[2].id.crl_static.max_rcc                = (USIGN8)  3;
    crl[2].id.crl_static.max_sac                = (USIGN8)  3;
    crl[2].id.crl_static.max_rac                = (USIGN8)  3;
    crl[2].id.crl_static.vfd_pointer            = 1;
    crl[2].id.crl_static.extension[0]           = 0;      /* NO EXTENSION */
    crl[2].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[2].id.crl_static.symbol[1],"MMAC               ");
}

/* --- build CRL entry ----------------------------------------------------- */
if (ts == TS5)
{
    crl[3].desired_cr                           = (USIGN16) 3;
    crl[3].id.crl_static.loc_lsap               = (USIGN8)  0;
    crl[3].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[3].id.crl_static.rem_lsap               = (USIGN8)  3;
    crl[3].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[3].id.crl_static.conn_type              = (USIGN8)  MSAC;
    crl[3].id.crl_static.conn_attr              = (USIGN8)  D_CONN;
    crl[3].id.crl_static.ci                     = (USIGN32) 1000;
    crl[3].id.crl_static.multiplier             = (USIGN8)  1;
    crl[3].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
    crl[3].id.crl_static.max_pdu_snd_low        = (USIGN8)  241;
    crl[3].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
    crl[3].id.crl_static.max_pdu_rcv_low        = (USIGN8)  241;
    crl[3].id.crl_static.rem_add                = (USIGN8)  10;
    crl[3].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[3].id.crl_static.feature_supp[1]        = (USIGN8)  0x30;
    crl[3].id.crl_static.feature_supp[2]        = (USIGN8)  0x80;
    crl[3].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[3].id.crl_static.feature_supp[4]        = (USIGN8)  0x30;
    crl[3].id.crl_static.feature_supp[5]        = (USIGN8)  0x80;
    crl[3].id.crl_static.max_scc                = (USIGN8)  3;
    crl[3].id.crl_static.max_rcc                = (USIGN8)  0;
    crl[3].id.crl_static.max_sac                = (USIGN8)  3;
    crl[3].id.crl_static.max_rac                = (USIGN8)  0;
    crl[3].id.crl_static.vfd_pointer            = 1;
    crl[3].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[3].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[3].id.crl_static.symbol[1],"MSAC (Master)       ");
}
else
{
    crl[3].desired_cr                           = (USIGN16) 3;
    crl[3].id.crl_static.loc_lsap               = (USIGN8)  3;
    crl[3].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[3].id.crl_static.rem_lsap               = (USIGN8)  0;
    crl[3].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[3].id.crl_static.conn_type              = (USIGN8)  MSAC;
    crl[3].id.crl_static.conn_attr              = (USIGN8)  D_CONN;
    crl[3].id.crl_static.ci                     = (USIGN32) 1000;
    crl[3].id.crl_static.multiplier             = (USIGN8)  1;
    crl[3].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
    crl[3].id.crl_static.max_pdu_snd_low        = (USIGN8)  241;
    crl[3].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
    crl[3].id.crl_static.max_pdu_rcv_low        = (USIGN8)  241;
    crl[3].id.crl_static.rem_add                = (USIGN8)  5;
    crl[3].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[3].id.crl_static.feature_supp[1]        = (USIGN8)  0x30;
    crl[3].id.crl_static.feature_supp[2]        = (USIGN8)  0x80;
    crl[3].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[3].id.crl_static.feature_supp[4]        = (USIGN8)  0x30;
    crl[3].id.crl_static.feature_supp[5]        = (USIGN8)  0x80;
    crl[3].id.crl_static.max_scc                = (USIGN8) 0;
    crl[3].id.crl_static.max_rcc                = (USIGN8) 3;
    crl[3].id.crl_static.max_sac                = (USIGN8) 0;
    crl[3].id.crl_static.max_rac                = (USIGN8) 3;
    crl[3].id.crl_static.vfd_pointer            = 1;
    crl[3].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[3].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[3].id.crl_static.symbol[1],"MSAC (Slave)        ");
}

/* --- build CRL entry ----------------------------------------------------- */
if (ts == TS5)
{
    crl[4].desired_cr                           = (USIGN16) 4;
    crl[4].id.crl_static.loc_lsap               = (USIGN8)  0;
    crl[4].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[4].id.crl_static.rem_lsap               = (USIGN8)  4;
    crl[4].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[4].id.crl_static.conn_type              = (USIGN8)  MSAC_SI;
    crl[4].id.crl_static.conn_attr              = (USIGN8)  D_CONN;
    crl[4].id.crl_static.ci                     = (USIGN32) 1000;
    crl[4].id.crl_static.multiplier             = (USIGN8)  1;
    crl[4].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
    crl[4].id.crl_static.max_pdu_snd_low        = (USIGN8)  241;
    crl[4].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
    crl[4].id.crl_static.max_pdu_rcv_low        = (USIGN8)  241;
    crl[4].id.crl_static.rem_add                = (USIGN8)  10;
    crl[4].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[4].id.crl_static.feature_supp[1]        = (USIGN8)  0x30;
    crl[4].id.crl_static.feature_supp[2]        = (USIGN8)  0x80;
    crl[4].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[4].id.crl_static.feature_supp[4]        = (USIGN8)  0x30;
    crl[4].id.crl_static.feature_supp[5]        = (USIGN8)  0x80;
    crl[4].id.crl_static.max_scc                = (USIGN8)  3;
    crl[4].id.crl_static.max_rcc                = (USIGN8)  0;
    crl[4].id.crl_static.max_sac                = (USIGN8)  3;
    crl[4].id.crl_static.max_rac                = (USIGN8)  3;
    crl[4].id.crl_static.vfd_pointer            = 1;
    crl[4].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[4].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[4].id.crl_static.symbol[1],"MSAC-SI (Master)    ");
}
else
{
    crl[4].desired_cr                           = (USIGN16) 4;
    crl[4].id.crl_static.loc_lsap               = (USIGN8)  4;
    crl[4].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[4].id.crl_static.rem_lsap               = (USIGN8)  0;
    crl[4].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[4].id.crl_static.conn_type              = (USIGN8)  MSAC_SI;
    crl[4].id.crl_static.conn_attr              = (USIGN8)  D_CONN;
    crl[4].id.crl_static.ci                     = (USIGN32) 1000;
    crl[4].id.crl_static.multiplier             = (USIGN8)  1;
    crl[4].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
    crl[4].id.crl_static.max_pdu_snd_low        = (USIGN8)  241;
    crl[4].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
    crl[4].id.crl_static.max_pdu_rcv_low        = (USIGN8)  241;
    crl[4].id.crl_static.rem_add                = (USIGN8)  5;
    crl[4].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[4].id.crl_static.feature_supp[1]        = (USIGN8)  0x30;
    crl[4].id.crl_static.feature_supp[2]        = (USIGN8)  0x80;
    crl[4].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[4].id.crl_static.feature_supp[4]        = (USIGN8)  0x30;
    crl[4].id.crl_static.feature_supp[5]        = (USIGN8)  0x80;
    crl[4].id.crl_static.max_scc                = (USIGN8) 0;
    crl[4].id.crl_static.max_rcc                = (USIGN8) 3;
    crl[4].id.crl_static.max_sac                = (USIGN8) 3;
    crl[4].id.crl_static.max_rac                = (USIGN8) 3;
    crl[4].id.crl_static.vfd_pointer            = 1;
    crl[4].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[4].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[4].id.crl_static.symbol[1],"MSAC-SI (Slave)     ");
}

if (ts == TS5)
{
    crl[5].desired_cr                           = (USIGN16) 5;
    crl[5].id.crl_static.loc_lsap               = (USIGN8)  0;
    crl[5].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[5].id.crl_static.rem_lsap               = (USIGN8)  5;
    crl[5].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[5].id.crl_static.conn_type              = (USIGN8)  MSCY;
    crl[5].id.crl_static.conn_attr              = (USIGN8)  D_CONN;
    crl[5].id.crl_static.ci                     = (USIGN32) 1000;
    crl[5].id.crl_static.multiplier             = (USIGN8)  1;
    crl[5].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
    crl[5].id.crl_static.max_pdu_snd_low        = (USIGN8)  241;
    crl[5].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
    crl[5].id.crl_static.max_pdu_rcv_low        = (USIGN8)  241;
    crl[5].id.crl_static.rem_add                = (USIGN8)  10;
    crl[5].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[5].id.crl_static.feature_supp[1]        = (USIGN8)  0x30;
    crl[5].id.crl_static.feature_supp[2]        = (USIGN8)  0x80;
    crl[5].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[5].id.crl_static.feature_supp[4]        = (USIGN8)  0x30;
    crl[5].id.crl_static.feature_supp[5]        = (USIGN8)  0x80;
    crl[5].id.crl_static.max_scc                = (USIGN8)  1;
    crl[5].id.crl_static.max_rcc                = (USIGN8)  0;
    crl[5].id.crl_static.max_sac                = (USIGN8)  0;
    crl[5].id.crl_static.max_rac                = (USIGN8)  0;
    crl[5].id.crl_static.vfd_pointer            = 1;
    crl[5].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[5].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[5].id.crl_static.symbol[1],"MSCY (Master)       ");
}
else
{
    crl[5].desired_cr                           = (USIGN16) 5;
    crl[5].id.crl_static.loc_lsap               = (USIGN8)  5;
    crl[5].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[5].id.crl_static.rem_lsap               = (USIGN8)  0;
    crl[5].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[5].id.crl_static.conn_type              = (USIGN8)  MSCY;
    crl[5].id.crl_static.conn_attr              = (USIGN8)  D_CONN;
    crl[5].id.crl_static.ci                     = (USIGN32) 1000;
    crl[5].id.crl_static.multiplier             = (USIGN8)  1;
    crl[5].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
    crl[5].id.crl_static.max_pdu_snd_low        = (USIGN8)  241;
    crl[5].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
    crl[5].id.crl_static.max_pdu_rcv_low        = (USIGN8)  241;
    crl[5].id.crl_static.rem_add                = (USIGN8)  5;
    crl[5].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[5].id.crl_static.feature_supp[1]        = (USIGN8)  0x30;
    crl[5].id.crl_static.feature_supp[2]        = (USIGN8)  0x80;
    crl[5].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[5].id.crl_static.feature_supp[4]        = (USIGN8)  0x30;
    crl[5].id.crl_static.feature_supp[5]        = (USIGN8)  0x80;
    crl[5].id.crl_static.max_scc                = (USIGN8) 0;
    crl[5].id.crl_static.max_rcc                = (USIGN8) 1;
    crl[5].id.crl_static.max_sac                = (USIGN8) 0;
    crl[5].id.crl_static.max_rac                = (USIGN8) 0;
    crl[5].id.crl_static.vfd_pointer            = 1;
    crl[5].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[5].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[5].id.crl_static.symbol[1],"MSCY (Slave)        ");
}

if (ts == TS5)
{
    crl[6].desired_cr                           = (USIGN16) 6;
    crl[6].id.crl_static.loc_lsap               = (USIGN8)  0;
    crl[6].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[6].id.crl_static.rem_lsap               = (USIGN8)  6;
    crl[6].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[6].id.crl_static.conn_type              = (USIGN8)  MSCY_SI;
    crl[6].id.crl_static.conn_attr              = (USIGN8)  D_CONN;
    crl[6].id.crl_static.ci                     = (USIGN32) 1000;
    crl[6].id.crl_static.multiplier             = (USIGN8)  1;
    crl[6].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
    crl[6].id.crl_static.max_pdu_snd_low        = (USIGN8)  241;
    crl[6].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
    crl[6].id.crl_static.max_pdu_rcv_low        = (USIGN8)  241;
    crl[6].id.crl_static.rem_add                = (USIGN8)  10;
    crl[6].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[6].id.crl_static.feature_supp[1]        = (USIGN8)  0x30;
    crl[6].id.crl_static.feature_supp[2]        = (USIGN8)  0x80;
    crl[6].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[6].id.crl_static.feature_supp[4]        = (USIGN8)  0x30;
    crl[6].id.crl_static.feature_supp[5]        = (USIGN8)  0x80;
    crl[6].id.crl_static.max_scc                = (USIGN8)  1;
    crl[6].id.crl_static.max_rcc                = (USIGN8)  0;
    crl[6].id.crl_static.max_sac                = (USIGN8)  3;
    crl[6].id.crl_static.max_rac                = (USIGN8)  3;
    crl[6].id.crl_static.vfd_pointer            = 1;
    crl[6].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[6].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[6].id.crl_static.symbol[1],"MSCY-SI (Master)    ");
}
else
{
    crl[6].desired_cr                           = (USIGN16) 6;
    crl[6].id.crl_static.loc_lsap               = (USIGN8)  6;
    crl[6].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[6].id.crl_static.rem_lsap               = (USIGN8)  0;
    crl[6].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[6].id.crl_static.conn_type              = (USIGN8)  MSCY_SI;
    crl[6].id.crl_static.conn_attr              = (USIGN8)  D_CONN;
    crl[6].id.crl_static.ci                     = (USIGN32) 1000;
    crl[6].id.crl_static.multiplier             = (USIGN8)  1;
    crl[6].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
    crl[6].id.crl_static.max_pdu_snd_low        = (USIGN8)  241;
    crl[6].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
    crl[6].id.crl_static.max_pdu_rcv_low        = (USIGN8)  241;
    crl[6].id.crl_static.rem_add                = (USIGN8)  5;
    crl[6].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[6].id.crl_static.feature_supp[1]        = (USIGN8)  0x30;
    crl[6].id.crl_static.feature_supp[2]        = (USIGN8)  0x80;
    crl[6].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[6].id.crl_static.feature_supp[4]        = (USIGN8)  0x30;
    crl[6].id.crl_static.feature_supp[5]        = (USIGN8)  0x80;
    crl[6].id.crl_static.max_scc                = (USIGN8) 0;
    crl[6].id.crl_static.max_rcc                = (USIGN8) 1;
    crl[6].id.crl_static.max_sac                = (USIGN8) 3;
    crl[6].id.crl_static.max_rac                = (USIGN8) 3;
    crl[6].id.crl_static.vfd_pointer            = 1;
    crl[6].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[6].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[6].id.crl_static.symbol[1],"MSCY-SI (Slave)     ");
}

if (ts == TS5)
{
    crl[7].desired_cr                           = (USIGN16) 7;
    crl[7].id.crl_static.loc_lsap               = (USIGN8)  7;
    crl[7].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[7].id.crl_static.rem_lsap               = (USIGN8)  63;
    crl[7].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[7].id.crl_static.conn_type              = (USIGN8)  BRCT;
    crl[7].id.crl_static.conn_attr              = (USIGN8)  I_CONN;
    crl[7].id.crl_static.ci                     = (USIGN32) 0;
    crl[7].id.crl_static.multiplier             = (USIGN8)  0;
    crl[7].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
    crl[7].id.crl_static.max_pdu_snd_low        = (USIGN8)  241;
    crl[7].id.crl_static.max_pdu_rcv_high       = (USIGN8)  0;
    crl[7].id.crl_static.max_pdu_rcv_low        = (USIGN8)  0;
    crl[7].id.crl_static.rem_add                = (USIGN8)  127;
    crl[7].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[7].id.crl_static.feature_supp[1]        = (USIGN8)  0x00;
    crl[7].id.crl_static.feature_supp[2]        = (USIGN8)  0x80;
    crl[7].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[7].id.crl_static.feature_supp[4]        = (USIGN8)  0x00;
    crl[7].id.crl_static.feature_supp[5]        = (USIGN8)  0x00;
    crl[7].id.crl_static.max_scc                = (USIGN8)  0;
    crl[7].id.crl_static.max_rcc                = (USIGN8)  0;
    crl[7].id.crl_static.max_sac                = (USIGN8)  1;
    crl[7].id.crl_static.max_rac                = (USIGN8)  0;
    crl[7].id.crl_static.vfd_pointer            = 1;
    crl[7].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[7].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[7].id.crl_static.symbol[1],"BRCT (Requester)    ");
}
else
{
    crl[7].desired_cr                           = (USIGN16) 7;
    crl[7].id.crl_static.loc_lsap               = (USIGN8)  63;
    crl[7].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[7].id.crl_static.rem_lsap               = (USIGN8)  255;
    crl[7].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[7].id.crl_static.conn_type              = (USIGN8)  BRCT;
    crl[7].id.crl_static.conn_attr              = (USIGN8)  O_CONN;
    crl[7].id.crl_static.ci                     = (USIGN32) 0;
    crl[7].id.crl_static.multiplier             = (USIGN8)  0;
    crl[7].id.crl_static.max_pdu_snd_high       = (USIGN8)  0;
    crl[7].id.crl_static.max_pdu_snd_low        = (USIGN8)  0;
    crl[7].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
    crl[7].id.crl_static.max_pdu_rcv_low        = (USIGN8)  241;
    crl[7].id.crl_static.rem_add                = (USIGN8)  255;
    crl[7].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[7].id.crl_static.feature_supp[1]        = (USIGN8)  0x00;
    crl[7].id.crl_static.feature_supp[2]        = (USIGN8)  0x00;
    crl[7].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[7].id.crl_static.feature_supp[4]        = (USIGN8)  0x00;
    crl[7].id.crl_static.feature_supp[5]        = (USIGN8)  0x80;
    crl[7].id.crl_static.max_scc                = (USIGN8)  0;
    crl[7].id.crl_static.max_rcc                = (USIGN8)  0;
    crl[7].id.crl_static.max_sac                = (USIGN8)  0;
    crl[7].id.crl_static.max_rac                = (USIGN8)  1;
    crl[7].id.crl_static.vfd_pointer            = 1;
    crl[7].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[7].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[7].id.crl_static.symbol[1],"BRCT (Receiver)     ");
}

if (ts == TS5)
{
    crl[8].desired_cr                           = (USIGN16) 8;
    crl[8].id.crl_static.loc_lsap               = (USIGN8)  8;
    crl[8].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[8].id.crl_static.rem_lsap               = (USIGN8)  8;
    crl[8].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[8].id.crl_static.conn_type              = (USIGN8)  MULT;
    crl[8].id.crl_static.conn_attr              = (USIGN8)  I_CONN;
    crl[8].id.crl_static.ci                     = (USIGN32) 0;
    crl[8].id.crl_static.multiplier             = (USIGN8)  0;
    crl[8].id.crl_static.max_pdu_snd_high       = (USIGN8)  241;
    crl[8].id.crl_static.max_pdu_snd_low        = (USIGN8)  241;
    crl[8].id.crl_static.max_pdu_rcv_high       = (USIGN8)  0;
    crl[8].id.crl_static.max_pdu_rcv_low        = (USIGN8)  0;
    crl[8].id.crl_static.rem_add                = (USIGN8)  127;
    crl[8].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[8].id.crl_static.feature_supp[1]        = (USIGN8)  0x00;
    crl[8].id.crl_static.feature_supp[2]        = (USIGN8)  0x80;
    crl[8].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[8].id.crl_static.feature_supp[4]        = (USIGN8)  0x00;
    crl[8].id.crl_static.feature_supp[5]        = (USIGN8)  0x00;
    crl[8].id.crl_static.max_scc                = (USIGN8)  0;
    crl[8].id.crl_static.max_rcc                = (USIGN8)  0;
    crl[8].id.crl_static.max_sac                = (USIGN8)  1;
    crl[8].id.crl_static.max_rac                = (USIGN8)  0;
    crl[8].id.crl_static.vfd_pointer            = 1;
    crl[8].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[8].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[8].id.crl_static.symbol[1],"MULT (Requester)    ");
}
else
{
    crl[8].desired_cr                           = (USIGN16) 8;
    crl[8].id.crl_static.loc_lsap               = (USIGN8)  8;
    crl[8].id.crl_static.rem_segm               = (USIGN8)  NO_SEGMENT;
    crl[8].id.crl_static.rem_lsap               = (USIGN8)  255;
    crl[8].id.crl_static.lli_sap                = (USIGN8)  0;
    crl[8].id.crl_static.conn_type              = (USIGN8)  MULT;
    crl[8].id.crl_static.conn_attr              = (USIGN8)  O_CONN;
    crl[8].id.crl_static.ci                     = (USIGN32) 0;
    crl[8].id.crl_static.multiplier             = (USIGN8)  0;
    crl[8].id.crl_static.max_pdu_snd_high       = (USIGN8)  0;
    crl[8].id.crl_static.max_pdu_snd_low        = (USIGN8)  0;
    crl[8].id.crl_static.max_pdu_rcv_high       = (USIGN8)  241;
    crl[8].id.crl_static.max_pdu_rcv_low        = (USIGN8)  241;
    crl[8].id.crl_static.rem_add                = (USIGN8)  255;
    crl[8].id.crl_static.feature_supp[0]        = (USIGN8)  0x00;
    crl[8].id.crl_static.feature_supp[1]        = (USIGN8)  0x00;
    crl[8].id.crl_static.feature_supp[2]        = (USIGN8)  0x00;
    crl[8].id.crl_static.feature_supp[3]        = (USIGN8)  0x00;
    crl[8].id.crl_static.feature_supp[4]        = (USIGN8)  0x00;
    crl[8].id.crl_static.feature_supp[5]        = (USIGN8)  0x80;
    crl[8].id.crl_static.max_scc                = (USIGN8)  0;
    crl[8].id.crl_static.max_rcc                = (USIGN8)  0;
    crl[8].id.crl_static.max_sac                = (USIGN8)  0;
    crl[8].id.crl_static.max_rac                = (USIGN8)  1;
    crl[8].id.crl_static.vfd_pointer            = 1;
    crl[8].id.crl_static.extension[0]           = 0; /* NO EXTENSION */
    crl[8].id.crl_static.symbol[0]              = 20;
    strcpy(&crl[8].id.crl_static.symbol[1],"MULT (Receiver)     ");
}


return;
}


FUNCTION static VOID 
    build_od(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to build the OD

possible return values:
- NONE

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES
USIGN16 i;

FUNCTION_BODY

/* --- initilize OD with NULL objects -------------------------------------- */
for (i=0; i<NO_OF_OD_OBJECT_DESCR; i++)
{
  od[i].vfd_number                           = 1;
  od[i].obj_descr.id.null_obj_descr.index    = i;
  od[i].obj_descr.id.null_obj_descr.obj_code = NULL_OBJECT;
}

/* --- build OD header ----------------------------------------------------- */
od[0].vfd_number                                 = 1;
od[0].obj_descr.id.od_obj_descr.index            = 0;
od[0].obj_descr.id.od_obj_descr.obj_code         = OD_OBJECT;
od[0].obj_descr.id.od_obj_descr.flag             = PB_TRUE;
od[0].obj_descr.id.od_obj_descr.length           = 9;
od[0].obj_descr.id.od_obj_descr.protection       = 0;
od[0].obj_descr.id.od_obj_descr.version          = 1;
od[0].obj_descr.id.od_obj_descr.len_st_od        = LEN_ST_OD;
od[0].obj_descr.id.od_obj_descr.first_index_s_od = LEN_ST_OD + 1;
od[0].obj_descr.id.od_obj_descr.len_s_od         = LEN_S_OD;
od[0].obj_descr.id.od_obj_descr.first_index_dv_od= 0;
od[0].obj_descr.id.od_obj_descr.len_dv_od        = LEN_DV_OD;
od[0].obj_descr.id.od_obj_descr.first_index_dp_od= 0;
od[0].obj_descr.id.od_obj_descr.len_dp_od        = LEN_DP_OD;
od[0].obj_descr.id.od_obj_descr.int_addr         = 0xFFFFFFFF;
od[0].obj_descr.id.od_obj_descr.int_addr_st_od   = 0xFFFFFFFF;
od[0].obj_descr.id.od_obj_descr.int_addr_s_od    = 0xFFFFFFFF;
od[0].obj_descr.id.od_obj_descr.int_addr_dv_od   = 0xFFFFFFFF;
od[0].obj_descr.id.od_obj_descr.int_addr_dp_od   = 0xFFFFFFFF;


/* --- build static type descriptions -------------------------------------- */
od[7].vfd_number                           = 1;
od[7].obj_descr.id.dt_obj_descr.index      = 7;
od[7].obj_descr.id.dt_obj_descr.obj_code   = TYPE_OBJECT;
od[7].obj_descr.id.dt_obj_descr.meaning[0] = 4;
strcpy (&od[7].obj_descr.id.dt_obj_descr.meaning[1], "U-32");

/* --- build variable object descriptions ---------------------------------- */
od[20].vfd_number                                    = 1;
od[20].obj_descr.id.s_var_obj_descr.index            = 20;
od[20].obj_descr.id.s_var_obj_descr.obj_code         = SIMPLE_VAR_OBJECT;
od[20].obj_descr.id.s_var_obj_descr.length           = 4;
od[20].obj_descr.id.s_var_obj_descr.index_of_type    = OD_USIGN32;
od[20].obj_descr.id.s_var_obj_descr.access.pass_word = 0;
od[20].obj_descr.id.s_var_obj_descr.access.acc_groups= 0;
od[20].obj_descr.id.s_var_obj_descr.access.acc_right = 255;
od[20].obj_descr.id.s_var_obj_descr.local_address    = 0xFFFFFFFF;
od[20].obj_descr.id.s_var_obj_descr.extension[0]     = 0;
od[20].obj_descr.id.s_var_obj_descr.name[0]          = 8;
strcpy(&od[20].obj_descr.id.s_var_obj_descr.name[1],"S_VAR_20");

od[21].vfd_number = 1;
od[21].obj_descr.id.s_var_obj_descr.index            = 21;
od[21].obj_descr.id.s_var_obj_descr.obj_code         = SIMPLE_VAR_OBJECT;
od[21].obj_descr.id.s_var_obj_descr.length           = 4;
od[21].obj_descr.id.s_var_obj_descr.index_of_type    = OD_USIGN32;
od[21].obj_descr.id.s_var_obj_descr.access.pass_word = 0;
od[21].obj_descr.id.s_var_obj_descr.access.acc_groups= 0;
od[21].obj_descr.id.s_var_obj_descr.access.acc_right = 255;
od[21].obj_descr.id.s_var_obj_descr.local_address    = 0xFFFFFFFF;
od[21].obj_descr.id.s_var_obj_descr.extension[0]     = 0;
od[21].obj_descr.id.s_var_obj_descr.name[0]          = 8;
strcpy(&od[21].obj_descr.id.s_var_obj_descr.name[1],"S_VAR_21");

return;

}

/* ------------------------------------------------------------------------- */
/* FMS functions:                                                            */
/* - to establish a FMS connection                                           */
/* - to deestablish  a FMS connection                                        */
/* - to read a simple variable object                                        */
/* - to write a simple variable object                                       */
/* - for information report                                                  */
/* - to indicate an abort                                                    */
/* - to indicate a reject                                                    */
/* ------------------------------------------------------------------------- */

FUNCTION static INT16 
    fms_initiate_req(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to establish the default connection

possible return values:

== E_OK    -> no error
!= E_OK    -> interface errors

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_CTXT_INIT_REQ    init_req;

FUNCTION_BODY

printf("\nFMS_INITIATE.req:\n");
fflush(stdout);

/* --- set data block parameters ------------------------------------------- */
init_req.profile_number[0] = 0;
init_req.profile_number[1] = 0;
init_req.password          = 0;
init_req.access_groups     = 0xFF;

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = crl[act_cr].desired_cr;
sdb.layer      = FMS;
sdb.service    = FMS_INITIATE;
sdb.primitive  = REQ;
sdb.invoke_id  = invoke_id=0;

return(profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,(VOID FAR*) &init_req,PB_TRUE));
}



FUNCTION extern INT16 
    fms_initiate_ind(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to receive the initiate indication to establish
the default connection

possible return values:

== E_OK    -> no error
!= E_OK    -> interface errors

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_CTXT_INIT_CNF FAR*  init_rsp = (T_CTXT_INIT_CNF FAR*) cnf_ind_buffer;

FUNCTION_BODY

info_rpt_priority = LOW;

default_variable_object_20_data = 0;       /* reset default variable objects */
default_variable_object_21_data = 0;

act_cr = sdb.comm_ref;                      /* set act cr for server device */

printf("\nFMS_INITIATE.ind:\n");
fflush(stdout);

/* --- NOTE: the server user has to analyse the indication data structure
             to check profile, password and access groups ------------------ */

/* --- send initiate response ---------------------------------------------- */

/* --- set data block parameters ------------------------------------------- */
init_rsp->profile_number[0] = 0;
init_rsp->profile_number[1] = 0;
init_rsp->password          = 0;
init_rsp->access_groups     = 0xFF;

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = crl[act_cr].desired_cr;
sdb.layer      = FMS;
sdb.service    = FMS_INITIATE;
sdb.primitive  = RES;
sdb.invoke_id  = invoke_id=0;

return(profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,(VOID FAR*) init_rsp,PB_TRUE));
}



FUNCTION static VOID 
    fms_initiate_cnf(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to receive the confirmation of the FMS-Initiate service

possible return values:

- NONE

-----------------------------------------------------------------------------*/
{

  T_CTXT_INIT_ERR_CNF FAR* init_err_cnf = (T_CTXT_INIT_ERR_CNF FAR*) cnf_ind_buffer;

     
  printf("\nfms_initiate_cnf:\n");

  if (sdb.result == POS)
  {
    default_variable_object_20_data = 0;
    default_variable_object_21_data = 0;
    info_rpt_priority               = LOW;

    /* --- positive confirmation -------------------------------------------- */
    printf("\nFMS_INITIATE.con(+):\n");
  }
  else
  {
    /* --- negative confirmation -------------------------------------------- */
    printf("\nFMS_INITIATE.con(-): %x\n",init_err_cnf->class_code);
   }
  fflush(stdout);
  return;
}



FUNCTION static INT16 
    fms_read_req(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to read the data of the default variable object

possible return values:

- == E_OK    -> no error
- != E_OK    -> interface errors

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_VAR_READ_REQ     read_req;
USIGN8             c = '\0';

FUNCTION_BODY

if (!multi_read_variable)
{
  printf("\n                'S' : SINGLE READ VARIABLE");
  printf("\n                'M' : MULTI  READ VARIABLE");
  printf("\n");
  fflush(stdout);

  c = (USIGN8) getch();
  if (toupper(c) == 'M') multi_read_variable = PB_TRUE;
  else                   multi_read_variable = PB_FALSE;
}

/* --- set data block parameters ------------------------------------------- */
read_req.acc_spec.tag      = ACCESS_INDEX;

read_req.acc_spec.id.index = 20;
read_req.subindex          = 0;

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = crl[act_cr].desired_cr;
sdb.layer      = FMS;
sdb.service    = FMS_READ;
sdb.primitive  = REQ;
sdb.invoke_id  = compute_invoke_id();

return(profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,(VOID FAR*) &read_req,PB_TRUE));
}



FUNCTION extern INT16 
    fms_read_ind(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to receive the read indication to read data of the
default variable object

possible return values:

== E_OK    -> no error
!= E_OK    -> interface errors

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_VAR_READ_CNF FAR *read_rsp  = (T_VAR_READ_CNF FAR*) cnf_ind_buffer;

USIGN8        FAR  *read_data = (USIGN8         FAR*) (read_rsp + 1);

FUNCTION_BODY

printf("\nFMS_READ.ind:\n");
fflush(stdout);

/* --- NOTE: the server user has to analyse the indication data structure
             has to read the object description from the OD if he does not
             know anything about the variable object --------------------- */

/* --- set data block parameters ------------------------------------------- */
/* --- read and auto increment the value of the default variable object -- */
read_rsp->length = 4;


/* --- swap high/low byte (MOTORALA to INTEL) of the default variable object */
read_data[0] = ((USIGN8 FAR*) &default_variable_object_20_data) [3];
read_data[1] = ((USIGN8 FAR*) &default_variable_object_20_data) [2];
read_data[2] = ((USIGN8 FAR*) &default_variable_object_20_data) [1];
read_data[3] = ((USIGN8 FAR*) &default_variable_object_20_data) [0];

default_variable_object_20_data++;

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = crl[act_cr].desired_cr;
sdb.layer      = FMS;
sdb.service    = FMS_READ;
sdb.primitive  = RES;

/* --- send read response ------------------------------------------------- */
return(profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,(VOID FAR*) read_rsp,PB_TRUE));
}



FUNCTION static VOID fms_read_cnf(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to receive the confirmation of the FMS-Read service

possible return values:

- NONE

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_VAR_READ_CNF  FAR* read_cnf     = (T_VAR_READ_CNF FAR*) cnf_ind_buffer;
T_ERROR         FAR* read_err_cnf = (T_ERROR        FAR*) cnf_ind_buffer;

USIGN8          FAR* read_data    = (USIGN8         FAR*) (read_cnf + 1);
USIGN32              data32;

FUNCTION_BODY

if (sdb.result == POS)
{
   /* --- positive confirmation -------------------------------------------- */

   /* --- swap high/low byte (MOTORALA to INTEL) --------------------------- */
   ((USIGN8 FAR*) &data32) [0]  = read_data[3];
   ((USIGN8 FAR*) &data32) [1]  = read_data[2];
   ((USIGN8 FAR*) &data32) [2]  = read_data[1];
   ((USIGN8 FAR*) &data32) [3]  = read_data[0];

   printf("\nFMS_READ.con(+): index: 20  value: %lu\n",data32);

   if (multi_read_variable) fms_read_req();
}
else
{
   /* --- negative confirmation -------------------------------------------- */
   printf("\nFMS_READ.con(-): %X\n",read_err_cnf->class_code);

   multi_read_variable = PB_FALSE;
}
fflush(stdout);
return;
}


FUNCTION static INT16 fms_write_req(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to write the data of the default variable object

possible return values:

- == E_OK    -> no error
- != E_OK    -> interface errors

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_VAR_WRITE_REQ FAR* write_req  = (T_VAR_WRITE_REQ FAR*) req_rsp_buffer;
USIGN8          FAR* write_data = (USIGN8 FAR*) (write_req + 1);
USIGN8               c = '\0';

FUNCTION_BODY

if (!multi_write_variable)
{
  printf("\n                'S' : SINGLE WRITE VARIABLE");
  printf("\n                'M' : MULTI  WRITE VARIABLE");
  printf("\n");

  c = (USIGN8) getch();
  if (toupper(c) == 'M') multi_write_variable = PB_TRUE;
  else                   multi_write_variable = PB_FALSE;
}

/* --- set data block parameters ------------------------------------------- */
write_req->acc_spec.tag      = ACCESS_INDEX;

write_req->acc_spec.id.index = 21;
write_req->subindex          = 0;
write_req->length            = 4;

/* --- swap high/low byte (MOTORALA to INTEL) --------------------------- */
write_data[0] = ((USIGN8 FAR*) &default_variable_object_21_data) [3];
write_data[1] = ((USIGN8 FAR*) &default_variable_object_21_data) [2];
write_data[2] = ((USIGN8 FAR*) &default_variable_object_21_data) [1];
write_data[3] = ((USIGN8 FAR*) &default_variable_object_21_data) [0];

default_variable_object_21_data++;

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = crl[act_cr].desired_cr;
sdb.layer      = FMS;
sdb.service    = FMS_WRITE;
sdb.primitive  = REQ;
sdb.invoke_id  = compute_invoke_id();

return(profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,(VOID FAR*) write_req,PB_TRUE));
}



FUNCTION extern INT16 fms_write_ind(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to receive the write indication to write data of the
default variable object

possible return values:

== E_OK    -> no error
!= E_OK    -> interface errors

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_VAR_WRITE_REQ FAR *write_ind  = (T_VAR_WRITE_REQ FAR*) cnf_ind_buffer;
USIGN8          FAR *write_data    = (USIGN8  FAR*) (write_ind + 1);
USIGN32             data32;
FUNCTION_BODY


/* --- NOTE: the server user has to analyse the indication data structure
             has to read the object description from the OD if he does not
             know anything about the variable object ----------------------- */

/* --- swap high/low byte (MOTORALA to INTEL) ------------------------------ */
((USIGN8 FAR*) &data32) [0]  = write_data[3];
((USIGN8 FAR*) &data32) [1]  = write_data[2];
((USIGN8 FAR*) &data32) [2]  = write_data[1];
((USIGN8 FAR*) &data32) [3]  = write_data[0];

printf("\nFMS_WRITE.ind: index: 21 value: %lu\n", data32);
fflush(stdout);

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = crl[act_cr].desired_cr;
sdb.layer      = FMS;
sdb.service    = FMS_WRITE;
sdb.primitive  = RES;

/* --- send write response ------------------------------------------------- */
return(profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,NULL,PB_TRUE));
}



FUNCTION static VOID fms_write_cnf(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to receive the confirmation of the FMS-write service

possible return values:

- NONE

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_ERROR FAR* write_err_cnf = (T_ERROR FAR*) cnf_ind_buffer;

FUNCTION_BODY

if (sdb.result == POS)
{
   /* --- positive confirmation -------------------------------------------- */
   printf("\nFMS_WRITE.con(+):\n");

   if (multi_write_variable) fms_write_req();
}
else
{
   /* --- negative confirmation -------------------------------------------- */
   printf("\nFMS_WRITE.con(-): %X\n",write_err_cnf->class_code);

   multi_write_variable = PB_FALSE;
}
fflush(stdout);
return;
}


FUNCTION static INT16 fms_info_rpt_req(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to report the data of the default variable object

possible return values:

- == E_OK    -> no error
- != E_OK    -> interface errors

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_VAR_INFO_RPT_REQ FAR* info_rpt_req  = (T_VAR_INFO_RPT_REQ FAR*) req_rsp_buffer;
USIGN8             FAR* info_rpt_data = (USIGN8 FAR*) (info_rpt_req + 1);

FUNCTION_BODY

printf("\nFMS_INFO_RPT.req:\n");
fflush(stdout);

/* --- set data block parameters ------------------------------------------- */
info_rpt_req->priority          = info_rpt_priority;
info_rpt_req->acc_spec.tag      = ACCESS_INDEX;
info_rpt_req->acc_spec.id.index = 20;
info_rpt_req->subindex          = 0;
info_rpt_req->length            = 4;

/* --- swap high/low byte (MOTORALA to INTEL) --------------------------- */
info_rpt_data[0] = 0;
info_rpt_data[1] = 0;
info_rpt_data[2] = 0;
info_rpt_data[3] = 0;

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = crl[act_cr].desired_cr;
sdb.layer      = FMS;
sdb.service    = FMS_INFO_RPT;
sdb.primitive  = REQ;
sdb.invoke_id  = 0;

if (info_rpt_priority == LOW) info_rpt_priority = HIGH;
else                          info_rpt_priority = LOW;

return(profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,(VOID FAR*) info_rpt_req,PB_TRUE));
}



FUNCTION extern VOID fms_info_rpt_ind(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to receive the information report indication
to write data of the default variable object

possible return values:

== E_OK    -> no error
!= E_OK    -> interface errors

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_VAR_INFO_RPT_REQ FAR* info_rpt_ind  = (T_VAR_INFO_RPT_REQ FAR*) cnf_ind_buffer;
USIGN8             FAR* info_rpt_data = (USIGN8 FAR*) (info_rpt_ind + 1);

FUNCTION_BODY


/* --- NOTE: the server user has to analyse the indication data structure
             has to read the object description from the OD if he does not
             know anything about the variable object ----------------------- */

/* --- swap high/low byte (MOTORALA to INTEL) ------------------------------ */
((USIGN8 FAR*) &default_variable_object_20_data) [0]  = info_rpt_data[3];
((USIGN8 FAR*) &default_variable_object_20_data) [1]  = info_rpt_data[2];
((USIGN8 FAR*) &default_variable_object_20_data) [2]  = info_rpt_data[1];
((USIGN8 FAR*) &default_variable_object_20_data) [3]  = info_rpt_data[0];

if (info_rpt_ind->priority == LOW) printf("\nFMS_INFO_RPT.ind: LOW  index: 20 value: %lu\n", default_variable_object_20_data);
else                               printf("\nFMS_INFO_RPT.ind: HIGH index: 20 value: %lu\n", default_variable_object_20_data);

fflush(stdout);
return;
}


FUNCTION static INT16 fms_abort_req(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to deestablish the open default connection

possible return values:

- == E_OK    -> no error
- != E_OK    -> interface errors

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_CTXT_ABORT_REQ abort_req;

FUNCTION_BODY

printf("\nFMS_ABORT.req:\n");
fflush(stdout);

/* --- set data block parameters ------------------------------------------- */
abort_req.abort_id      = USR;
abort_req.reason        = USR_ABT_RC1;          /* ---- disconect ----- */
abort_req.detail_length = 10;
strcpy((VOID*)abort_req.detail,"USER_ABORT");

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = crl[act_cr].desired_cr;
sdb.layer      = FMS;
sdb.service    = FMS_ABORT;
sdb.primitive  = REQ;
sdb.invoke_id  = 0;

return(profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,(VOID FAR*) &abort_req,PB_TRUE));
}



FUNCTION static VOID fms_abort_ind(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to receive an FMS abort indication

possible return values:

- NONE

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES


T_CTXT_ABORT_REQ FAR *abort = (T_CTXT_ABORT_REQ FAR*) cnf_ind_buffer;
USIGN8               i;

FUNCTION_BODY

printf("\nFMS_ABORT.ind: local: %3u id: %u rc: %2u ad:\n",
       abort->local,
       abort->abort_id,
       abort->reason
      );

for (i=0; i<abort->detail_length; i++) printf("%3x",abort->detail[i]);


fflush(stdout);
return;
}


FUNCTION static VOID fms_reject_ind(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to receive an FMS reject indication

possible return values:

- NONE

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES


T_CTXT_REJECT_IND FAR *reject = (T_CTXT_REJECT_IND FAR*) cnf_ind_buffer;

FUNCTION_BODY

printf("\nFMS_REJECT.ind: local: %3u iv-id: %3u pdu-type: %u rc: %u\n",
        reject->detected_here,
        reject->orig_invoke_id,
        reject->pdu_type,
        reject->reject_code
       );

fflush(stdout);
return;
}


/* ------------------------------------------------------------------------- */
/* FM7 functions:                                                            */
/* - to indicate FM7 events (LLI-FAULT-INDICATIONS and FM2-EVENTS)           */
/* - to reset and leave the Demo application                                 */
/* ------------------------------------------------------------------------- */

FUNCTION static VOID fm7_event_ind(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to receive a fm7 event indication
- fm2 event
- lli fault indication
- lli event

possible return values:

- NONE

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_FM7_EVENT_IND FAR* event = (T_FM7_EVENT_IND FAR*) cnf_ind_buffer;

FUNCTION_BODY

printf("\nFM7_EVENT.ind: cr: %3d id: %1d rc: %2d\n",
        event->comm_ref,
        event->instance_id,
        event->reason
       );

fflush(stdout);
return;
}

/* ------------------------------------------------------------------------- */
/* FMB functions:                                                            */
/* - to configurate system memory                                            */
/* - to load busparameters                                                   */
/* - to shutdown controller                                                  */
/* ------------------------------------------------------------------------- */

FUNCTION static INT16 fmb_set_configuration(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to set the system configuration

possible return values:

- == E_OK    -> no error
- != E_OK    -> error

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_FMB_SET_CONFIGURATION_REQ fmb_config_req;
T_FMB_CONFIG_CRL            crl_config_req;
INT16                       i;

FUNCTION_BODY

printf("\nFMB_SET_CONFIGURATION.req:\n");
fflush(stdout);

/* --- calculate CRL resources --------------------------------------------- */
ccrl_resrces_init((T_CRL_HDR FAR*) &crl[0].id.crl_hdr);

for (i=1;i<crl[0].id.crl_hdr.nr_of_entries+1;i++)
{
   ccrl_resrces_add_entry(crl[i].desired_cr,
                          (T_CRL_STATIC     FAR*) &(crl[i].id.crl_static),
                          (T_FMB_CONFIG_CRL FAR*) &crl_config_req
                         );
}

fmb_config_req.fms_active                = PB_TRUE;
fmb_config_req.dp_active                 = PB_FALSE;
fmb_config_req.fdlif_active              = PB_FALSE;
fmb_config_req.sm7_active                = PB_FALSE;
fmb_config_req.fdl_evt_receiver          = FM7_USR;
fmb_config_req.data_buffer_length        = 255;

fmb_config_req.vfd.max_no_vfds           = MAX_VFD;
fmb_config_req.vfd.max_no_obj_descr      = 1500;
fmb_config_req.vfd.max_obj_name_length   = OBJECT_NAME_LENGTH;
fmb_config_req.vfd.max_obj_ext_length    = EXTENSION_LENGTH;

fmb_config_req.crl.max_no_fal_sdbs       = crl_config_req.max_no_fal_sdbs;
fmb_config_req.crl.max_no_fdl_sdbs       = crl_config_req.max_no_fdl_sdbs;
fmb_config_req.crl.max_no_data_buffer    = crl_config_req.max_no_data_buffer;
fmb_config_req.crl.max_no_api_buffer     = crl_config_req.max_no_api_buffer;
fmb_config_req.crl.max_no_poll_entries   = crl_config_req.max_no_poll_entries;
fmb_config_req.crl.max_no_subscr_entries = 0;

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = 0;
sdb.layer      = FMB;
sdb.service    = FMB_SET_CONFIGURATION;
sdb.primitive  = REQ;

/* --- send request ------------------------------------------------------- */
if ((ret_val = profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                                (VOID FAR*) &fmb_config_req,
                                PB_TRUE
                               )) != E_OK) return(ret_val);

wait_for_cnf_ind(WAIT);
return(E_OK);
}



FUNCTION static INT16 fmb_load_bus_param(
           IN USIGN8 ts
         )

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to load the bus parameter

possible return values:

- == E_OK    -> no error
- != E_OK    -> error

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_FMB_SET_BUSPARAMETER_REQ bus_par_req;

FUNCTION_BODY

printf("\nFMB_SET_BUSPARAMTER.req:\n");
fflush(stdout);

/* --- set data block parameters ------------------------------------------ */
/* set default busparameter for baudrate 500KBit/s ------------------------ */
bus_par_req.loc_add         = ts;

bus_par_req.in_ring_desired = 255;
bus_par_req.baud_rate       = 4;
bus_par_req.tset            = 1;
bus_par_req.loc_segm        = 255;
bus_par_req.medium_red      = 0;
bus_par_req.tsl             = 3500;
bus_par_req.min_tsdr        = 500;
bus_par_req.max_tsdr        = 1000;
bus_par_req.tqui            = 0;
bus_par_req.ttr             = 100000;
bus_par_req.g               = 1;
bus_par_req.hsa             = 126;
bus_par_req.max_retry_limit = 1;

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = 0;
sdb.layer      = FMB;
sdb.service    = FMB_SET_BUSPARAMETER;
sdb.primitive  = REQ;

if ((ret_val = profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                               (VOID FAR *) &bus_par_req,
                               (USIGN8) PB_TRUE )) != E_OK )
{
  return(ret_val);
}

/* --- wait for confirmation ----------------------------------------------- */
wait_for_cnf_ind(WAIT);

return(E_OK);
}



FUNCTION static INT16 fmb_exit(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

This function is used to shutdown the controler and leave the communication
software

Note: For a new start up of the communication software, initialize
      the communication interface and communication software
      using by the init_profibus() function


possible return values:

- == E_OK    -> no error
- != E_OK    -> error


-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES


FUNCTION_BODY

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = 0;
sdb.layer      = FMB;
sdb.service    = FMB_EXIT;
sdb.primitive  = REQ;

if ((ret_val = profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                               NULL,
                               (USIGN8) PB_TRUE )) != E_OK )
{
  return(ret_val);
}


/* --- wait for confirmation and leave the program ------------------------- */
wait_for_cnf_ind(WAIT);

return(E_OK);
}





/* ------------------------------------------------------------------------- */
/* scheduling functions:                                                     */
/* - to enter TS5 funtionality                                               */
/* - to enter TS10 funtionality                                              */
/* - to schedule the demo application and communication interface            */
/* ------------------------------------------------------------------------- */

FUNCTION static VOID wait_for_cnf_ind(
           IN USIGN8 wait
         )

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

This function is used to receive an indication or confirmation from
the communication interface


-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_ERROR        FAR  *error          = (T_ERROR        FAR*) cnf_ind_buffer;

USIGN16 i;
USIGN16 buffer_len;
USIGN8  c = '\0';

FUNCTION_BODY


for(;;)
{
   if (wait == WAIT)
   {
    if (kbhit())
    {
      c = (USIGN8) getch();
      if (toupper(c) == 'Z')
      {
        profi_end();
        exit(1);
       }
    }
   }


   buffer_len = 512;
   ret_val = E_OK;

   ret_val = profi_rcv_con_ind((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                               (VOID FAR *) cnf_ind_buffer,
                               &buffer_len
                              );
#ifdef WIN32
   Sleep(1);
#endif

#ifdef _LINUX
   //sleep (1);
#endif

   if (ret_val == CON_IND_RECEIVED)
   {
     switch(sdb.layer)
     {
        case FMS_USR:
             switch(sdb.service)
             {
               case FMS_INITIATE:
                    if (sdb.primitive == IND) fms_initiate_ind();
                    else                      fms_initiate_cnf();
                    break;

               case FMS_INFO_RPT:
                    if (sdb.primitive == IND) fms_info_rpt_ind();
                    break;

               case FMS_READ:
                    if (sdb.primitive == IND) fms_read_ind();
                    else                      fms_read_cnf();
                    break;

               case FMS_WRITE:
                    if (sdb.primitive == IND) fms_write_ind();
                    else                      fms_write_cnf();
                    break;

               case FMS_ABORT:
                    fms_abort_ind();
                    break;

               case FMS_REJECT:
                    fms_reject_ind();
                    break;

               case FMS_INIT_LOAD_OD_LOC:
                    if (sdb.result == POS) printf("\nFMS_INIT_LOAD_OD_LOC.con(+):\n");
                    else                   printf("\nFMS_INIT_LOAD_OD_LOC.con(-): 0x%x\n",error->class_code);
                    if (sdb.result != POS)
                    {
                      profi_end();
                      exit(0);
                    };
                    fflush(stdout);
                    if (wait == WAIT) return;
                    break;

               case FMS_LOAD_OD_LOC:
                    if (sdb.result == POS) printf("\nFMS_LOAD_OD_LOC.con(+):\n");
                    else                   printf("\nFMS_LOAD_OD_LOC.con(-): 0x%x\n",error->class_code);
                    if (sdb.result != POS)
                    {
                      profi_end();
                      exit(0);
                    };
                    fflush(stdout);
                    if (wait == WAIT) return;
                    break;

               case FMS_TERM_LOAD_OD_LOC:
                    if (sdb.result == POS) printf("\nFMS_TERM_LOAD_OD_LOC.con(+):\n");
                    else                   printf("\nFMS_TERM_LOAD_OD_LOC.con(-): 0x%x\n",error->class_code);
                    if (sdb.result != POS)
                    {
                      profi_end();
                      exit(0);
                    };
                    fflush(stdout);
                    if (wait == WAIT) return;
                    break;

               case FMS_CREATE_VFD_LOC:
                    if (sdb.result == POS) printf("\nFMS_CREATE_VFD_LOC.con(+):\n");
                    else                   printf("\nFMS_CREATE_VFD_LOC.con(-): 0x%x\n",error->class_code);
                    if (sdb.result != POS)
                    {
                      profi_end();
                      exit(0);
                    };
                    fflush(stdout);
                    if (wait == WAIT) return;
                    break;

               case FMS_VFD_SET_PHYS_STATUS_LOC:
                    if (sdb.result == POS) printf("\nFMS_VFD_SET_PHYS_STATUS_LOC.con(+):\n");
                    else                   printf("\nFMS_VFD_SET_PHYS_STATUS_LOC.con(-): 0x%x\n",error->class_code);
                    if (sdb.result != POS)
                    {
                      profi_end();
                      exit(0);
                    };
                    fflush(stdout);
                    if (wait == WAIT) return;
                    break;

               default:
                    printf("\n FMS service not supported\n");
                    printf("service: %d primitive: %d\n",sdb.service,sdb.primitive);
                    break;
             }

             fflush(stdout);
             if (wait == DONT_WAIT) return;
             break;

        case FM7_USR:
             switch(sdb.service)
             {
               case FM7_INIT_LOAD_CRL_LOC:
                    if (sdb.result == POS) printf("\nFM7_INIT_LOAD_CRL_LOC.con(+):\n");
                    else                   printf("\nFM7_INIT_LOAD_CRL_LOC.con(-): 0x%x\n",error->class_code);
                    if (sdb.result != POS)
                    {
                      profi_end();
                      exit(0);
                    };
                    fflush(stdout);
                    if (wait == WAIT) return;
                    break;

               case FM7_LOAD_CRL_LOC:
                    if (sdb.result == POS) printf("\nFM7_LOAD_CRL_LOC.con(+):\n");
                    else                   printf("\nFM7_LOAD_CRL_LOC.con(-): 0x%x\n",error->class_code);
                    if (sdb.result != POS)
                    {
                      profi_end();
                      exit(0);
                    };
                    fflush(stdout);
                    if (wait == WAIT) return;
                    break;

               case FM7_TERM_LOAD_CRL_LOC:
                    if (sdb.result == POS) printf("\nFM7_TERM_LOAD_CRL_LOC.con(+):\n");
                    else                   printf("\nFM7_TERM_LOAD_CRL_LOC.con(-):\n");
                    if (sdb.result != POS)
                    {
                      profi_end();
                    };
                    fflush(stdout);
                    if (wait == WAIT) return;
                    break;

               case FM7_EVENT:
                    fm7_event_ind();
                    break;

             default:
                    printf("\nFM7 service not supported\n");
                    printf("service: %d primitive: %d\n",sdb.service,sdb.primitive);
                    break;
             }
             fflush(stdout);
             if (wait == DONT_WAIT) return;
             break;

        case FMB_USR:
             switch(sdb.service)
             {
               case FMB_SET_BUSPARAMETER:
                    if (sdb.result == POS) printf("\nFMB_SET_BUSPARAMETER.con(+):\n");
                    else                   printf("\nFMB_SET_BUSPARAMETER.con(-): 0x%x\n",error->class_code);
                    if (sdb.result != POS)
                    {
                      profi_end();
                      exit(0);
                    };
                    fflush(stdout);
                    if (wait == WAIT) return;
                    break;

               case FMB_SET_CONFIGURATION:
                    if (sdb.result == POS) printf("\nFMB_SET_CONFIGURATION.con(+):\n");
                    else                   printf("\nFMB_SET_CONFIGURATION.con(-): 0x%x\n",error->class_code);
                    if (sdb.result != POS)
                    {
                      profi_end();
                      exit(0);
                    };
                    fflush(stdout);
                    if (wait == WAIT) return;
                    break;

               case FMB_EXIT:
                    if (sdb.result == POS) printf("\nFMB_EXIT.con(+):\n");
                    else                   printf("\nFMB_EXIT.con(-): 0x%x\n",error->class_code);
                    for (i=0;i<0xFFFF;i++) wait_for_exit++;
                    profi_end();
                    exit(0);

             default:
                    printf("\nFMB service not supported\n");
                    printf("service: %d primitive: %d\n",sdb.service,sdb.primitive);
                    break;
             }
             fflush(stdout);
             if (wait == DONT_WAIT) return;
             break;
     }
   }
   else
   {
     if (ret_val == NO_CON_IND_RECEIVED)
     {
       fflush(stdout);
        if (wait == DONT_WAIT) return;
     }
     else
     {
       print_if_error(IND,ret_val);
       profi_end();
       exit(ret_val);
     }
   }
}
fflush(stdout);
return;
}



FUNCTION static VOID station5(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

TS5 waits for keyboard inputs of the user

- with the demo application the user can do the following FMS services

  1.) INITIATE:         establish the default connection
  2.) READ VARIABLE:    read the data of the default variable 20 object
  3.) WRITE VARIABLE:   write the data of the default variable 21 object
  4.) INFO REPORT:      information report on default variable 20 object
  5.) ABORT:            deestablish the open default connection
  6.) EXIT:             leave the demo application


-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

USIGN8 c = '\0';

FUNCTION_BODY

irq_rcv_ctr = 0;

act_cr = 2;

#ifdef DOS16
_clearscreen ( _GCLEARSCREEN );
#endif

while (PB_TRUE)
{
#ifdef DOS16
   _clearscreen ( _GCLEARSCREEN );
#endif
   ret_val = E_OK;

   printf("\n                             PROFIBUS FMS STATION 5");
   printf("\n                             ----------------------\n");
   printf("\n");
   printf("      '2..8' : GET CR                 CR: %d  CONN-TYPE: %s\n",act_cr,&crl[act_cr].id.crl_static.symbol[1]);
   printf("\n");
   printf("      'I'    : INITIATE\n");
   printf("      'R'    : READ VARIABLE\n");
   printf("      'W'    : WRITE VARIABLE\n");
   printf("      'F'    : INFORMATION REPORT\n");
   printf("      'A'    : ABORT\n");
   printf("\n");
   printf("      'E'    : EXIT\n\n");

   while (!kbhit()) wait_for_cnf_ind(DONT_WAIT);

   fflush(stdout);
   c = (USIGN8) getch();

   switch (toupper(c))
   {
     case '2': act_cr = 2; break;
     case '3': act_cr = 3; break;
     case '4': act_cr = 4; break;
     case '5': act_cr = 5; break;
     case '6': act_cr = 6; break;
     case '7': act_cr = 7; break;
     case '8': act_cr = 8; break;

     case 'I':
          ret_val = fms_initiate_req();
          break;

     case 'R':
          multi_read_variable = PB_FALSE;
          ret_val = fms_read_req();
          break;

     case 'W':
          multi_write_variable = PB_FALSE;
          ret_val = fms_write_req();
          break;

     case 'F':
          ret_val = fms_info_rpt_req();
          break;

     case 'A':
          ret_val = fms_abort_req();
          break;

     case 'E':
          ret_val = fmb_exit();
          break;

     default:
          ret_val = E_OK;
          break;

   }
   if (ret_val != E_OK) print_if_error(REQ,ret_val);
}
return;
}


FUNCTION static VOID station10(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

TS10 waits for keyboard inputs of the user

- with the demo application the user can do the following FMS services

  1.) INITIATE:         establish the default connection
  2.) READ VARIABLE:    read the data of the default variable 20 object
  3.) WRITE VARIABLE:   write the data of the default variable 21 object
  4.) INFO REPORT:      information report on default variable 20 object
  5.) ABORT:            deestablish the open default connection
  6.) EXIT:             leave the demo application


-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

USIGN8 c = '\0';

FUNCTION_BODY

irq_rcv_ctr = 0;

act_cr = 2;

#ifdef DOS16
 _clearscreen ( _GCLEARSCREEN );
#endif
 while (PB_TRUE)
 {
   ret_val = E_OK;

#ifdef DOS16
   _clearscreen ( _GCLEARSCREEN );
#endif

   printf("\n                            PROFIBUS FMS STATION 10");
   printf("\n                            -----------------------\n");
   printf("\n");
   printf("      '2..8' : GET CR                 CR: %d  CONN-TYPE: %s\n",act_cr,&crl[act_cr].id.crl_static.symbol[1]);
   printf("\n");
   printf("      'I'    : INITIATE\n");
   printf("      'R'    : READ VARIABLE\n");
   printf("      'W'    : WRITE VARIABLE\n");
   printf("      'F'    : INFORMATION REPORT\n");
   printf("      'A'    : ABORT\n");
   printf("\n");
   printf("      'E'    : EXIT\n\n");

   while (!kbhit()) wait_for_cnf_ind(DONT_WAIT);

   fflush(stdout);
   c = (USIGN8) getch();

   switch (toupper(c))
   {
     case '2': act_cr = 2; break;
     case '3': act_cr = 3; break;
     case '4': act_cr = 4; break;
     case '5': act_cr = 5; break;
     case '6': act_cr = 6; break;
     case '7': act_cr = 7; break;
     case '8': act_cr = 8; break;


     case 'I':
          ret_val = fms_initiate_req();
          break;

     case 'R':
          multi_read_variable = PB_FALSE;
          ret_val = fms_read_req();
          break;

     case 'W':
          multi_write_variable = PB_FALSE;
          ret_val = fms_write_req();
          break;

     case 'F':
          ret_val = fms_info_rpt_req();
          break;

     case 'A':
          ret_val = fms_abort_req();
          break;

     case 'E':
          ret_val = fmb_exit();
          break;

     default:
          ret_val = E_OK;
          break;

   }
   if (ret_val != E_OK) print_if_error(REQ,ret_val);
}
return;
}



/* ------------------------------------------------------------------------- */
/* intialization                                                             */
/* - initialize the communication interface                                  */
/* - download firmware                                                       */
/* - build CRL                                                               */
/* - set configuration                                                       */
/* - create a VFD                                                            */
/* - load the OD                                                             */
/* - load the bus parameter                                                  */
/* - load the crl                                                            */
/* - set the physical device status in the VFD                               */
/* ------------------------------------------------------------------------- */


FUNCTION static INT16 fms_create_vfd(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to create a VFD

possible return values:

- == E_OK    -> no error
- != E_OK    -> error

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_VFD_CREATE_REQ   create_vfd;

FUNCTION_BODY

printf("\nFMS_CREATE_VFD_LOC.req:\n");

/* --- set data block parameters ------------------------------------------ */
create_vfd.vfd_number = (USIGN32) 1;

/* vendor name (byte 0 contains the length of the vendor name) ------------- */
create_vfd.vendor_name[0] = 7;
strcpy(&create_vfd.vendor_name[1],"SOFTING");

/* model name (byte 0 contains the length of the model name) --------------- */
create_vfd.model_name[0]  = 16;
strcpy(&create_vfd.model_name[1],"DEMO APPLICATION");

/* revision (byte 0 contains the length of the revision) ------------------- */
create_vfd.revision[0] = 4;
strcpy(&create_vfd.revision[1],"5.24");

/* -- profile number (no profile is specified) ----------------------------- */
create_vfd.profile_number[0] = 0;
create_vfd.profile_number[1] = 0;


/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = 0;
sdb.layer      = FMS;
sdb.service    = FMS_CREATE_VFD_LOC;
sdb.primitive  = REQ;

/* --- send request ------------------------------------------------------- */
if ((ret_val = profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                                (VOID FAR*) &create_vfd,
                                PB_TRUE
                               )) != E_OK) return(ret_val);

wait_for_cnf_ind(WAIT);
return(E_OK);
}



FUNCTION static INT16 fms_load_od(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to load the OD

possible return values:

- == E_OK    -> no error
- != E_OK    -> error

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_INIT_LOAD_OD_REQ     init_load_od_req;
T_LOAD_OD_REQ          load_od_req;
T_TERM_LOAD_OD_REQ     term_load_od_req;
USIGN16                i,j;

FUNCTION_BODY

/* ------------------------------------------------------------------------ */
/* ----------------- initiate load od ------------------------------------- */
/* ------------------------------------------------------------------------ */

printf("\nFMS_INIT_LOAD_OD_LOC.req:\n");

/* --- set data block parameters ------------------------------------------ */
init_load_od_req.vfd_number   = 1;
init_load_od_req.consequence  = CONSEQUENCE_2;

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = 0;
sdb.layer      = FMS;
sdb.service    = FMS_INIT_LOAD_OD_LOC;
sdb.primitive  = REQ;

if ((ret_val = profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                               (VOID FAR *) &init_load_od_req,
                               (USIGN8) PB_TRUE)) != E_OK)
{
  return(ret_val);
}

/* --- wait for confirmation ----------------------------------------------- */
wait_for_cnf_ind(WAIT);

/* ------------------------------------------------------------------------- */
/* ------------------load od descriptions ---------------------------------- */
/* ------------------------------------------------------------------------- */

/* --- load OD object descriptions ----------------------------------------- */
for (j=0; j< NO_OF_OD_OBJECT_DESCR;j++)
{
   printf("\nFMS_LOAD_OD_LOC.req:  index: %d\n",j);
   for (i=0;i< 0x7FFF;i++);      /* wait to display INFO */

   load_od_req = od[j];

   /* --- set parameter block parameters ----------------------------------- */
   sdb.comm_ref   = 0;
   sdb.layer      = FMS;
   sdb.service    = FMS_LOAD_OD_LOC;
   sdb.primitive  = REQ;

   if ((ret_val = profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                                    (VOID FAR*) &load_od_req,
                                    (USIGN8) PB_TRUE )) != E_OK )
   {
     return(ret_val);
   }

   /* --- wait for confirmation -------------------------------------------- */
   wait_for_cnf_ind(WAIT);
}

/* ------------------------------------------------------------------------- */
/* ------------------terminate load od  ------------------------------------ */
/* ------------------------------------------------------------------------- */


printf("\nFMS_TERM_LOAD_OD_LOC.req:\n");
for (i=0;i< 0x7FFF;i++);      /* wait to display INFO */

/* -- set data nblock parameters ------------------------------------------- */
term_load_od_req.vfd_number   = 1;

/* -- set parameter block parameter -------------------------------------- */
sdb.comm_ref   = 0;
sdb.layer      = FMS;
sdb.service    = FMS_TERM_LOAD_OD_LOC;
sdb.primitive  = REQ;

if ((ret_val = profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                               (VOID FAR *) &term_load_od_req,
                               (USIGN8) PB_TRUE )) != E_OK )
{
  return(ret_val);
}

/* --- wait for confirmation ----------------------------------------------- */
wait_for_cnf_ind(WAIT);

return(E_OK);
}




FUNCTION static INT16 fm7_load_crl(
           VOID
         )

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to load the crl

possible return values:

- == E_OK    -> no error
- != E_OK    -> error

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_LOAD_CRL_REQ   crl_req;
INT16            i,j;

FUNCTION_BODY

/* ------------------------------------------------------------------------ */
/* ----------------- initiate load crl ------------------------------------ */
/* ------------------------------------------------------------------------ */
printf("\nFM7_INIT_LOAD_CRL_LOC.req:\n");
for (i=0;i< 0x7FFF;i++);      /* wait to display INFO */

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = 0;
sdb.layer      = FM7;
sdb.service    = FM7_INIT_LOAD_CRL_LOC;
sdb.primitive  = REQ;

if ((ret_val = profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                               NULL,
                               (USIGN8) PB_TRUE )) != E_OK )
{
  return(ret_val);
}

/* --- wait for confirmation ----------------------------------------------- */
wait_for_cnf_ind(WAIT);

/* ------------------------------------------------------------------------ */
/* ----------------- load crl entries  ------------------------------------ */
/* ------------------------------------------------------------------------ */

/* --- crl header --------------------------------------------------------- */
printf("\nFM7_LOAD_CRL_LOC.req:  CRL-HEADER\n");
for (i=0;i< 0x7FFF;i++);      /* wait to display INFO */

/* --- set data block parameters ------------------------------------------ */
crl_req = crl[0];

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = 0;
sdb.layer      = FM7;
sdb.service    = FM7_LOAD_CRL_LOC;
sdb.primitive  = REQ;

if ((ret_val = profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                               (VOID FAR*) &crl_req,
                               (USIGN8) PB_TRUE )) != E_OK )
{
  return(ret_val);
}

/* --- wait for confirmation ----------------------------------------------- */
wait_for_cnf_ind(WAIT);


/* --- crl entries --------------------------------------------------------- */
/* --- set data block parameters ------------------------------------------- */
for (j=1; j<crl[0].id.crl_hdr.nr_of_entries+1;j++)
{
  printf("\nFM7_LOAD_CRL_LOC.req:  %d. CRL-ENTRY\n",j);
  for (i=0;i< 0x7FFF;i++);      /* wait to display INFO */

  crl_req = crl[j];

  /* --- set parameter block parameters ------------------------------------ */
  sdb.comm_ref   = 0;
  sdb.layer      = FM7;
  sdb.service    = FM7_LOAD_CRL_LOC;
  sdb.primitive  = REQ;

  if ((ret_val = profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                                   (VOID FAR*) &crl_req,
                                   (USIGN8) PB_TRUE )) != E_OK )
  {
    return(ret_val);
  }

  /* --- wait for confirmation --------------------------------------------- */
  wait_for_cnf_ind(WAIT);
}


/* ------------------------------------------------------------------------ */
/* ----------------- terminate load crl ----------------------------------- */
/* ------------------------------------------------------------------------ */

printf("\nFM7_TERM_LOAD_CRL_LOC.req\n");
for (i=0;i< 0x7FFF;i++);      /* wait to display INFO */
/* --- set parameter block parameters ------------------------------------- */
sdb.comm_ref   = 0;
sdb.layer      = FM7;
sdb.service    = FM7_TERM_LOAD_CRL_LOC;
sdb.primitive  = REQ;

if ((ret_val = profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                               NULL,
                               (USIGN8) PB_TRUE )) != E_OK )
{
  return(ret_val);
}

/* --- wait for confirmation ----------------------------------------------- */
wait_for_cnf_ind(WAIT);


return(E_OK);
}


FUNCTION static INT16 fms_set_phys_status(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to set the physical device status in the vfd object

possible return values:

- == E_OK    -> no error
- != E_OK    -> error

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

T_VFD_SET_PHYS_STATUS_REQ  vfd_phys_status;

FUNCTION_BODY

printf("\nFMS_SET_PHYS_STATUS_LOC.req:\n");

vfd_phys_status.vfd_number      = (USIGN32)1;
vfd_phys_status.physical_status = (USIGN8) OPERATIONAL;

/* --- set parameter block parameters -------------------------------------- */
sdb.comm_ref   = 0;
sdb.layer      = FMS;
sdb.service    = FMS_VFD_SET_PHYS_STATUS_LOC;
sdb.primitive  = REQ;

if ((ret_val = profi_snd_req_res((T_PROFI_SERVICE_DESCR FAR*) &sdb,
                               (VOID FAR *) &vfd_phys_status,
                               (USIGN8) PB_TRUE )) != E_OK )
{
  return(ret_val);
}

/* --- wait for confirmation ----------------------------------------------- */
wait_for_cnf_ind(WAIT);

return(E_OK);
}





FUNCTION static INT16 init_profi(
           IN USIGN8 ts
         )

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

this function is used to initialize the communication interface

- download firmware
- create a VFD
- load an OD
- load the busparameter
- load the crl
- set the physical status of the VFD

possible return values:

- == E_OK    -> no error

-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES


FUNCTION_BODY

#ifdef DOS16
/* --- download firmware --------------------------------------------------- */
_setbkcolor(0);
_settextposition(24,24);
_settextcolor(7+16);
_displaycursor(_GCURSOROFF);
_outtext("   INITIALIZE CONTROLLER    ");
#endif

#if defined (WIN32) || defined (_LINUX)
ret_val = profi_set_default((USIGN8) board_number,0,0,0);
#else
ret_val = init_profibus (dpr_base_address, io_address, PB_FALSE);
#endif

#ifdef DOS16
_settextposition(24,24);
_settextcolor(7);
_outtext("                            ");
_displaycursor(_GCURSORON);
#endif

if (ret_val != E_OK) return(ret_val);

/* --- print version ------------------------------------------------------- */
#ifdef DOS16
_clearscreen ( _GCLEARSCREEN );
#endif
print_version();
print_papi_firmware_version();
printf("\n\n press return");
getch();
#ifdef DOS16
_clearscreen ( _GCLEARSCREEN );
#endif
/* --- build OD ------------------------------------------------------------ */
build_od();

/* --- build CRL ----------------------------------------------------------- */
build_crl(ts);

/* --- set configuration --------------------------------------------------- */
if ((ret_val = fmb_set_configuration()) != E_OK)
{
  print_if_error(REQ,ret_val);
  return(ret_val);
}

/* --- create a VFD -------------------------------------------------------- */
if ((ret_val = fms_create_vfd()) != E_OK)
{
  print_if_error(REQ,ret_val);
  return(ret_val);
}

/* --- load the OD --------------------------------------------------------- */
if ((ret_val = fms_load_od()) != E_OK)
{
  print_if_error(REQ,ret_val);
  return(ret_val);
}

/* --- load the bus parameter ---------------------------------------------- */
if ((ret_val = fmb_load_bus_param(ts)) != E_OK)
{
  print_if_error(REQ,ret_val);
  return(ret_val);
}

/* --- load the crl -------------------------------------------------------- */
if ((ret_val = fm7_load_crl()) != E_OK)
{
  print_if_error(REQ,ret_val);
  return(ret_val);
}

/* --- set the physical status in the VFD ---------------------------------- */
if ((ret_val = fms_set_phys_status()) != E_OK)
{
  print_if_error(REQ,ret_val);
  return(ret_val);
}

return(E_OK);
}


/* ------------------------------------------------------------------------- */
/* Functions for initializing the program and program parameters             */
/* - print program version                                                   */
/* - print program usage                                                     */
/* - check command line parameters                                           */
/* - set default options                                                     */
/* ------------------------------------------------------------------------- */


FUNCTION static VOID print_version(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

prints program version on screen

possible return values:

- NONE
-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  printf ("\n             PROFIBUS FMS Demo V 5.24 (C) 1995-2003 Softing AG\n\n");
  printf ("                               Softing AG\n");
  printf ("                          Richard-Reitzner-Allee 6\n");
  printf ("                             D-85540 Haar\n\n");
  printf ("                     Phone: (++49) (0) 89 / 45 65 6 - 0\n");
  printf ("                     Fax  : (++49) (0) 89 / 45 65 6 - 399\n\n");

}

FUNCTION static VOID print_papi_firmware_version(VOID)

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

prints program version on screen

possible return values:

- NONE
-----------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  CSTRING   papi_version[VERSION_STRING_LENGTH];
  CSTRING   fw_version[VERSION_STRING_LENGTH];
  USIGN32   serial_device_number;

  FUNCTION_BODY

  profi_get_versions((CSTRING FAR*) papi_version,(CSTRING FAR*) fw_version);
  profi_get_serial_device_number((USIGN32 FAR*)&serial_device_number);

  printf("\n Serial Device Number: %lu",serial_device_number);
  printf("\n PAPI Version:         %s" ,papi_version);
  printf("\n FW Version:           %s" ,fw_version);


}


FUNCTION static VOID print_usage(
  IN    VOID
  )
/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

prints program version and usage on screen

possible return values:

- NONE
-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

#if defined (WIN16) || defined (WIN32)
  printf ("Usage  : %s [options]  [station]\n\n", exe_fname);
#else
  printf ("Usage  : fmsdemo [options]\n\n");
#endif

  printf ("options: \n");
#if defined (WIN32) || defined (_LINUX)
  printf ("-b<add>  = Board Number          (default: 0\n");
  printf ("-h | -?  = display this help information\n\n");
#else
  printf ("-a<add>  = PC DPRAM address      (default: 0x%08lX)\n", H_DPR_BASE_ADDRESS);
  printf ("-i<add>  = PC IO address         (default: 0x%04X)\n" , IO_ADDRESS);
  printf ("-h | -?  = display this help information\n\n");
#endif

  printf ("station: \n");
  printf ("5        = station address 5     (client)\n");
  printf ("10       = station address 10    (server)\n\n");

} /* print_usage */


FUNCTION static VOID check_options(
  IN    char    options []
  )
/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

check command line parameters for valid options and TS5/TS10

valid options:
    -b<BOARD_NUMBER>             only WIN32
    -a<DPR_ADDRESS>              only WIN16
    -i<IO_ADDRESS>               only WIN16
    -h | -?                      print usage
possible return values:

- NONE
-----------------------------------------------------------------------------*/
{
  LOCAL_VARIABLES

  char *token_ptr;

  FUNCTION_BODY

  token_ptr = strtok (options, " \t\r");

  if (token_ptr != NULL)
  {
#if defined (WIN32) || defined (_LINUX)
    if (strncmp (token_ptr, "-b", 2) == 0)
    {
      if (sscanf (token_ptr, "-b%hi", &board_number) == 1)
      {
        if(!((board_number > 0 ) && (board_number <= 9 )))
        {
          board_number = BOARD_NUMBER;
        }
      }
      else
      {
        board_number = BOARD_NUMBER;
      }
    }
#else
    if (strncmp (token_ptr, "-a", 2) == 0)
    {
      if (sscanf (token_ptr, "-a0x%8lx", &dpr_base_address) != 1)
      {
        if (sscanf (token_ptr, "-a%10lu", &dpr_base_address) != 1)
        {
          dpr_base_address = (USIGN32) H_DPR_BASE_ADDRESS;
        }
      }
    }
    else if (strncmp (token_ptr, "-i", 2) == 0)
    {
      if (sscanf (token_ptr, "-i0x%4x", &io_address) != 1)
      {
        if (sscanf (token_ptr, "-i%5u", &io_address) != 1)
        {
          io_address = (USIGN16) IO_ADDRESS;
        }
      }
    }
#endif
    else if ( stricmp (token_ptr,"5") == 0 )
    {
      ts = TS5;
    }
    else if ( stricmp (token_ptr,"10") == 0 )
    {
      ts = TS10;
    }
    else if ( (strcmp (token_ptr, "-?") == 0) ||
              (strcmp (token_ptr, "-h") == 0) )
    {
      print_version ();
      print_usage ();
      exit (0);
    }
    else
    {
      print_version ();
      printf ("error: unknown command line parameter \"%s\" \n\n", token_ptr);
      print_usage ();
      exit (-1);
    }

    token_ptr = strtok (NULL, " \t\r");

  } /* if token_ptr */
}


FUNCTION static VOID set_default_options(
  IN    VOID
  )
/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

set default options for command line parameters

possible return values:

- NONE
-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

#if defined (WIN32) || defined (_LINUX)
  board_number      = BOARD_NUMBER;
#else
  dpr_base_address  = (USIGN32) H_DPR_BASE_ADDRESS;
  io_address        = (USIGN16) IO_ADDRESS;
#endif

  ts                = 5;

}



/* ------------------------------------------------------------------------- */
/* main program                                                              */
/* ------------------------------------------------------------------------- */

FUNCTION extern int main(
          IN  int  argc,
          IN  char **argv
        )

/*----------------------------------------------------------------------------
FUNCTIONAL_DESCRIPTION

main procedure of the demo application

possible return values:

- NONE
-----------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  int       i;

#if defined (WIN16) || defined (WIN32)
  char      dummy [_MAX_PATH];
#endif

FUNCTION_BODY
#ifdef DOS16
  _clearscreen ( _GCLEARSCREEN );
#endif

#if defined (WIN16) || defined (WIN32)
  _splitpath (argv [0], dummy, dummy, exe_fname, dummy);
#endif

  set_default_options ();                      /* program default parameters */

  /* -- check command line ------------------------------------------------- */
  for (i = 1; i < argc; i++)
  {
    check_options (argv [i]);
  }

  print_version();

  printf("Settings:\n");
  printf("acting as %s \n", (ts==5) ? TS5_TXT : TS10_TXT);

  /* initialize the communication interface ---------------------------------- */
  if ((ret_val = init_profi(ts)) == E_OK)
  {
    printf("\nInitialization successful\n");

    /* schedule the application and communication --------------------------- */
    if (ts == 5) station5();
    else         station10();
    profi_end();
    exit(0);
  }
  else
  {
    printf("\nInitialization failed\n");
    print_if_error(0xFF,ret_val);

    profi_end();
    exit(1);
  }

  profi_end();

  exit(0);
} /* main */

